require 'thread'

class CalculationQueue
	#=======================================================================
	# Constructor
	#=======================================================================
	
	def initialize(threads)
		# Initialize empty queue
		@calculations = []
		@compression_on = false

		# Load serial or parallel queue as requested
		if threads == 1
			Cuby::log.puts_debug "Serial queue created"
			extend CalculationQueueSerial
		else
			Cuby::log.puts_debug "Parallel queue created using #{threads} threads"
			extend CalculationQueueParallel
			@threads = threads
		end
	end

	#=======================================================================
        # Queue info
	#=======================================================================

        def size
                return @calculations.size
        end

	#=======================================================================
	# Settings
	#=======================================================================
	
	def set_compression(compflag)
		@compression_on = compflag
	end

	#=======================================================================
	# Filling the queue
	#=======================================================================
	
	def add_to_queue(calculation)
		Cuby::log.puts_debug "Calculation \"#{calculation.name}\" added to queue"
		@calculations << calculation
	end

	#=======================================================================
	# Queue compression/decompression
	#=======================================================================
	
	# Removes duplicate calculation, and copies their results after
	# performing the calculations

	def compress!
		return unless @compression_on

		# Invalidate any results
		@calculations.each{|calc|
			calc.clean_results
		}

		# Save copy of the complete list of calculations
		@old_calculations = @calculations.dup

		# Save checksums of all calculations
		@old_checksums = []
		@old_calculations.each{|calc|
			@old_checksums << calc.checksum
		}

		# Store only unique calculations
		@calculations = []
		@compressed_calculations = {}
		@old_checksums.uniq.each{|ch|
			i = @old_checksums.index(ch)
			@calculations << @old_calculations[i]
			@compressed_calculations[@old_checksums[i]] = @old_calculations[i]
		}

		Cuby::log.puts_debug "Compressing the queue: kept #{@calculations.size} out of #{@old_calculations.size} calculations"
		return nil
	end

	def decompress!
		Cuby::log.puts_debug "Decompressing queue"
		return unless @old_calculations
		@old_calculations.each_index{|i|
			calc = @old_calculations[i]
			unless calc.has_results?
				ch = @old_checksums[i]
				# At this stage, the results exist, but were not processed
				# yet. They can be accessed using results_raw
				results =  @compressed_calculations[ch].results_raw
				calc.copy_results!(results)
			end
		}

	end

	#=======================================================================
	# Private methods
	#=======================================================================
	
	def sort_largest_first!
		@calculations.sort!{|a, b|
			a.priority <=> b.priority
		}
		@calculations.reverse!
		return nil
	end

end

module CalculationQueueSerial
	def execute
		if @calculations.size == 0
			Cuby::log.puts_debug "Queue executed but is empty"
			return 0 
		end

		sort_largest_first!

		compress!

		size = @calculations.size
		Cuby::log.puts_debug "Queue running - #{size} calculations"
		while calc = @calculations.shift
			Cuby::log.puts_debug "Calculating \"#{calc.name}\" (priority #{calc.priority})"
			calc.calculate(0)
		end
		Cuby::log.puts_debug "Queue execution finished"
		decompress!
		return size
	end

        def clear
                @calculations.clear
        end
end

module CalculationQueueParallel
	def execute
		if @calculations.size == 0
			Cuby::log.puts_debug "Queue executed but is empty"
			return 0 
		end

		sort_largest_first!

		compress!

		size = @calculations.size
		Cuby::log.puts_debug "Queue running - #{size} calculations, running #{@threads} in parallel "
		semaphore = Mutex.new
		index = 0
		controllers = []
		@threads.times {|i|
			controllers[i] = Thread.new {
				semaphore.lock
				while index < @calculations.size
					this_i = index
					index += 1
					Cuby::log.puts_debug "   Calculating \"#{@calculations[this_i].name}\" (priority #{@calculations[this_i].priority})"
					semaphore.unlock
					@calculations[this_i].calculate(i)
					semaphore.lock
				end
				semaphore.unlock
			}
			controllers[i].abort_on_exception = true
		}
		controllers.each_index {|i| controllers[i].join}

		@calculations = []

		Cuby::log.puts_debug "Queue execution finished"
		decompress!

		return size
	end
end
