require "classes/objects/molecular_charge.rb"

class Job
	attr_reader :name, :settings
	attr_reader :protocol_name
	attr_reader :prepared

	#=======================================================================
	# Build list of protocols available in the protocols directory
	#=======================================================================
	@@protocols = {} # List of available protocols
	@@protocols_path = Cuby.install_dir + "/protocols"
	# Fill the list
	Dir.entries(@@protocols_path).sort.each{|dir|
		next if dir == '.' || dir == '..'
		next unless FileTest.directory?(@@protocols_path + "/" + dir)
		@@protocols[dir.downcase.to_sym] = @@protocols_path + "/" + dir
	}

	#=======================================================================
	# Access to the list of protocols
	#=======================================================================
	def Job.available_protocols
		return @@protocols
	end

	#=======================================================================
	# Core functionality
	#=======================================================================
	def initialize(name, settings)
		# Create new job, loading the appropriate protocol module
		@name = name
		@settings = settings

		unless Job.available_protocols.has_key?(@settings[:job])
			Cuby::error("Job type '#{@settings[:job]}' not found.")
		end

		# Load the protocol file and extend this job with the prtotocol module
		protocol_name = @settings[:job].to_s
		protocol_path = Job.available_protocols[@settings[:job]]
		require protocol_path + "/" + protocol_name + ".rb"
		protocol_module = eval("Protocol" + protocol_name.to_s.split("_").map{|s| s.capitalize}.join(''))
		extend protocol_module

		# Save the protocol name
		@protocol_name = @settings[:job]

		# No results yet
		@results = nil

		@prepared = false
	end

	def print_header
		# Print ASCIIart header for selected calculation
		if @settings.set?(:job_title)
			Cuby.log.print_cuby_logo(:comment => false, :text => @settings[:job_title])

		else
			print_header_protocol
		end
	end

	def prepare
		#: Preparation of calculations etc
		#: Allowed options:
		#* :silent (true/false) - disable printing of the header
		#!# options might be passed through settings

		# Evaluate conditions
		job_conditions

		# Call the protocol preparation
		prepare_protocol
		@prepared = true
		if @settings[:prepare_only]
			Cuby.log.puts('Calculations have been prepared, calculation will be skipped');
		end
	end

	def run(queue = Cuby.queue)
		return if @settings[:prepare_only]
		# Invalidate old results, we want to calculate it anew
		@results = nil
		# Runs the job
		run_protocol(queue)
	end

	def print
		return if @settings[:prepare_only]
		# Print results to log
		print_protocol
		
		# Save serialized results
		if @settings.set?(:write_results_yaml)
			File.open(@settings[:write_results_yaml], "w+"){|f|
				f.puts(results.to_yaml)
			}
		end
	end

	def cleanup
		return if @settings[:prepare_only]
		cleanup_protocol
		@prepared = false
	end

	def geometry
		# If there is a single geometry the protocol operates on, return it
		if @geometry
			return @geometry
		else
			Cuby::error("Job #{name} does not yield the geometry it operates on")
		end
	end

	#=======================================================================
	# Optional functionality
	#=======================================================================
	
	# Optionally, the protocols might implement following methods:
	#
	# single_line_result
	# returns a string with the bare result - e.g. energy job would return
	# a string "Energy: 1.23456 kcal/mol" (if applicable)
	#
	# single_number_result
	# returns just one number (if applicable)

	#=======================================================================
	# Helpers
	#=======================================================================
	
	# If more code is added to this section, it should be moved to a separate
	# module (e.g. ProtocolCommon) that will be included or used as ancestor
	# of the protocols that use it 
	
	def job_conditions
		# Evaluate the conditions in conditional blocks and
		# mix the contents in if evaluated to true
		@settings.blocks.each_pair{|blockname, contents|
			if blockname.to_s.downcase =~ /^condition_/
				condition = contents[:condition]
				if eval(condition)
					Cuby::log.puts_debug "Conditional block '#{blockname}' is applied"
					@settings.copy_from!(contents, :overwrite)
				else
					Cuby::log.puts_debug "Conditional block '#{blockname}' not used"
				end
			end
		}
	end
	
	def load_geometry
		# Loads the geometry from :geometry keyword, applying selection
		# from the :selection keyword (everything by default)
	
		# Load geometry
		@geometry = Geometry.load_from_settings(@settings)

		# Track changes to the geometry, to be printed later
		changes = []

		# Set ghost atoms
		if @settings.set?(:ghost_atoms)
			 @geometry.geometry_from_selection(@settings[:ghost_atoms]).each{|atom|
				 atom.properties[:ghost] = true
			 }
			changes << "some atoms changed to ghosts"
		end

		# Load atomic charges into geometry
		if @settings.set?(:geometry_load_charges)
			charges = AtomicCharges.from_file(@settings[:geometry_load_charges])
			charges.write_to_geometry(@geometry)
			Cuby::log.puts_debug("Atomic charges from file #{@settings[:geometry_load_charges]} loaded into geometry")

		end

		# Charge and multiplicity from geometry file
		if @settings[:geometry_setup_from_file]
			info = @geometry.get_extended_info
			if info
				@settings[:charge] = info.charge || 0
				@settings[:multiplicity] = info.multiplicity || 1
			else
				@settings[:charge] = 0
				@settings[:multiplicity] = 1
				# Charge, test for integer value (in case it is a sum of atomic charges)
				if @geometry.info[:geofile_charge]
					charge = @geometry.info[:geofile_charge].round
					err = (@geometry.info[:geofile_charge] - charge).abs
					if err > 0.1
						"Charge from geometry file (#{@geometry.info[:geofile_charge]}) is far from integer value"
					end
					@settings[:charge] = charge
				end
				# Multiplicity
				if @geometry.info[:geofile_multiplicity]
					@settings[:multiplicity] = @geometry.info[:geofile_multiplicity]
				end
			end
		end

		# Charge from arbitrary file
		if @settings.set?(:charge_from_file)
			@settings[:charge] = MolecularCharge.from_file(@settings[:charge_from_file])
			Cuby::log.puts_v(:normal, "Molecular charge #{@settings[:charge]} read from file #{@settings[:charge_from_file]}")
            if @settings.set?(:charge_from_file_add)
              @settings[:charge] += @settings[:charge_from_file_add]
              Cuby::log.puts_v(:normal, "Adding #{@settings[:charge_from_file_add]} to charge read from file.")
            end
		end

		# Apply selection
		unless @settings[:selection] == "auto"
			@geometry = @geometry.geometry_from_selection(@settings[:selection])
			changes << "selection applied"
			if @geometry.size == 0
				Cuby::error("No atoms left in the geometry after application of the keyword 'selection'")
			end
		end

		# Print the geometry if requested
		if @settings[:print].include?(:input_geometry)
			if changes.size > 0
				s = " (" + changes.join(', ') + ")"
			else
				s = ""
			end
			Cuby::log.puts "Input geometry#{s}:"
			Cuby::log.puts @geometry.to_s
			Cuby::log.puts
		end
	end

	def write_geometry
		filename = "---"
		if @settings[:geometry_write] != ""
			filename = @settings[:geometry_write]
			filetype = @settings[:geometry_write_format]
			if filetype == :auto
				if @settings[:geometry_write] == "/dev/stdout"
					# Standard output - use same format for input if possible
					if @geometry.info[:filetype] && GeoFile::TYPE_MODULE_W.has_key?(@geometry.info[:filetype])
						filetype = @geometry.info[:filetype]
					else
						# otherwise use .xyz
						filetype = :xyz
					end
				else
					# Regular file - determine from file name
					filetype = GeoFile.type_from_filename(filename)
					unless filetype
						Cuby::error("Failed to determine geometry file format from file name '#{filename}'")
					end
				end
			end

			# General options
			write_options = {
				:append => @settings[:geometry_write_append],
				:extra_columns => @settings[:pdb_extra_columns]
			}

			if @settings.set?(:connectivity_tolerance)
				Connectivity.set_tolerance(@settings[:connectivity_tolerance])
			end

			# Header
			if @settings[:geometry_write_cuby_header]
				write_options[:cuby_header] = true
			end

			# File-format specific options
			case filetype
			when :xyz
				write_options[:second_line] = @geometry.info[:protocol_comment] if @geometry.info[:protocol_comment]
			when :mol2
				# explicit connectivity has to be defined
				case @settings[:geometry_write_connectivity]
				when :from_radii
					conn = Connectivity.new(@geometry, @settings[:connectivity_tolerance])
					write_options[:bondlist] = conn.bond_list
				when :from_zmat
					if @geometry.info[:z_matrix]
						write_options[:bondlist] = @geometry.info[:z_matrix].bond_list
					else
						Cuby::error("Using z-matrix connectivity was requested but the geometry does not have\na z-matrix associated.")
					end
				end
			end

			@geometry.write_file(filename, filetype, write_options)
		end
		return filename
	end

	def protocol_dir
		return Job.available_protocols[@settings[:job]]
	end

end
