
module GeometrySmiles

	def GeometrySmiles.build(smiles_string, settings)
		case settings[:smiles_engine]
		when :balloon
			return GeometrySmiles.build_balloon(smiles_string, settings)
		when :obabel
			return GeometrySmiles.build_obabel(smiles_string, settings)
		else
			Cuby::error "No smiles builder set - see keyword smiles_engine"
		end
	end

	def GeometrySmiles.build_balloon(smiles_string, settings)
		# Check for presence of the Ballon program
		if settings[:balloon_dir] == "" || !File.exist?(settings[:balloon_dir])
			Cuby::error("To build geometry from the SMILES notation, the Balloon program\n(http://users.abo.fi/mivainio/balloon) must be installed\nand pointed to by the keyword 'balloon_dir'.")
		end

		# Create a temporary filename 
		while FileTest.exist?(filename = 'temp' + sprintf("%08d",(rand*1.0e8).to_i) + ".sdf") do end

		# Run balloon
		balloon_dir = settings[:balloon_dir]
		system "#{balloon_dir}/balloon -f #{balloon_dir}/MMFF94.mff --noGA '#{smiles_string}' #{filename} 2> /dev/null > /dev/null"

		# Test result
		Cuby::error "SMILES translation by Balloon failed:\n#{smiles_string}" unless FileTest.exist?(filename)

		# Read geometry and delete the file
		g = Geometry.new
		g.read_file(filename)
		File.delete(filename)

		# Fix hydrogens
		g.each_with_index{|atom_h, i|
			# The first element is hydrogen
			next unless atom_h.element == :H
			# Find the closest atom
			dist = 9.9e99
			atom_c = nil
			g.each_with_index{|atom, j|
				next if i == j
				r = atom_h.distance(atom)
				if r < dist
					dist = r
					atom_c = atom
				end
			}
			next unless atom_c
			# Check the distance
			r = atom_h.distance(atom_c)
			rcov = PeriodicTable.covalent_radius(atom_h.element) +
				PeriodicTable.covalent_radius(atom_c.element)
			if (r > rcov * 1.1) && (r < rcov * 1.5)
				# Bond too long, set it to rcov
				vec = atom_h - atom_c
				vec = vec * (rcov/r)
				atom_h.set_coord(atom_c + vec)
			end
		}

		return g
	end

	def GeometrySmiles.build_obabel(smiles_string, settings)
		# Create a temporary filename 
		while FileTest.exist?(filename = 'temp' + sprintf("%08d",(rand*1.0e8).to_i) + ".mol2") do end

		# Write smiles file
		File.open(filename + '.smi', "w+"){|f| f.puts smiles_string}

		# Run obabel
		system "obabel -ismi #{filename}.smi -omol2 -O #{filename} -h --gen3d 2> /dev/null > /dev/null"

		# Delete input file
		File.delete(filename + '.smi')

		# Test result
		Cuby::error "SMILES translation by obabel failed:\n#{smiles_string}" unless FileTest.exist?(filename)

		# Read geometry and delete the file
		g = Geometry.new
		g.read_file(filename)
		File.delete(filename)

		return g
	end
end
