module GeometrySystemInfo
	# Information stored across file formats in the form of remarks
	
	class GeoExtendedInfo
		attr_accessor :charge
		attr_accessor :multiplicity
        attr_accessor :energy
        attr_accessor :converged # yes/no
		attr_accessor :fragment_selection
		attr_accessor :fragment_charge

		def initialize
		end

        def size
          size = 0
          size += 1 if @charge
          size += 1 if @multiplicity
          size += 1 if @energy
          size += 1 if @converged
          size += @fragment_selection.size if @fragment_selection
          size += @fragment_charge.size if @fragment_charge
          return size
        end

		def GeoExtendedInfo.from_single_line(string)
			# Read the info from a single line
			# In this case, the format is "key=value" (without spaces) records
			# separated by space
			result = GeoExtendedInfo.new
			string.strip.split.each{|entry|
				if entry =~ /[^=]+=[^=]+/
					k,v = entry.split("=")
					result.save_name_value(k,v)
				end
			}
			return result
		end

		def GeoExtendedInfo.from_lines(array)
			# Read the info from multi-line data (array of strings)
			# In this case, the format is "key value" with one record per line
			result = GeoExtendedInfo.new
			array.each{|line|
				name, value = line.strip.split
				result.save_name_value(name, value)
			}
			return result
		end

		def save_name_value(name, value)
			# save key-value pair into the respective variable
			case name.downcase
			when "charge"
				@charge = value.to_i
			when "multiplicity"
				@multiplicity = value.to_i
            when "energy"
                @energy = value.to_f
            when "converged"
                @converged = value.downcase == "yes"
			when /^selection_.*/
				fragname = name.gsub(/^selection_/,'')
				@fragment_selection = {} unless @fragment_selection
				@fragment_selection[fragname] = value
			when /^charge_.*/
				fragname = name.gsub(/^charge_/,'')
				@fragment_charge = {} unless @fragment_charge
				@fragment_charge[fragname] = value.to_i
			end
			return nil
		end

		def to_s
			s = []
			s << "Geometry system info:"
			s << "   Charge: #{@charge}" if @charge
			s << "   Multiplicity: #{@multiplicity}" if @multiplicity
			s << "   Energy: #{@energy}" if @energy
            s << "   Converged: yes" if @converged
			if @fragment_selection
				s << "   Fragment selections:"
				@fragment_selection.each_pair{|k,v|
					s << "      #{k}: #{v}"
				}
			end
			if @fragment_charge
				s << "   Fragment charges:"
				@fragment_charge.each_pair{|k,v|
					s << "      #{k}: #{v}"
				}
			end
			return s.join("\n")
		end

	    def to_line
			s = []
			s << "charge=#{@charge}" if @charge
			s << "multiplicity=#{@multiplicity}" if @multiplicity
			s << "energy=#{@energy}" if @energy
            s << "converged=yes" if @converged
			if @fragment_selection
				@fragment_selection.each_pair{|k,v|
					s << "selection_#{k}=#{v}"
				}
			end
			if @fragment_charge
				@fragment_charge.each_pair{|k,v|
					s << "charge_#{k}=#{v}"
				}
			end
			return s.join(" ")
		end

	end

	def get_extended_info
		if info[:xyz_remark]
			# parse single-line remark
			return GeoExtendedInfo.from_single_line(info[:xyz_remark])
		elsif info[:pdb_remark]
			# parse multi-line record
			return GeoExtendedInfo.from_lines(info[:pdb_remark].map{|s| s.gsub(/^REMARK */,"")})
		else
			return nil
		end
	end
end
