class PeriodicBoxRectangular
	attr_reader :dimensions
	attr_reader :origin

	#========================================================================
	# Constructors
	#========================================================================

	def initialize(a,b,c, origin = Coordinate[0,0,0])
		@dimensions = Coordinate[a,b,c]
		@origin = origin
	end

	def PeriodicBoxRectangular.new_cubic(a, origin = Coordinate[0,0,0])
		return PeriodicBoxRectangular.new(a, a, a, origin)
	end

	def PeriodicBoxRectangular.new_from_settings(settings)
		boxarray = settings[:periodic_box]
		origin = settings[:periodic_box_origin]
		# Check if all elements are numbers
		boxarray.map!{|x|
			begin
				x = Settings.to_float(x)
			rescue
				Cuby::error("Period box dimensions must be numbers.")
			end
			if x <= 0.0
				Cuby::error("Periodic box dimensions must be positive numbers.")
			end
			x
		}
		origin.map!{|x|
			begin
				x = Settings.to_float(x)
			rescue
				Cuby::error("Period box origin coordinates must be numbers.")
			end
			x
		}
		# Get origin as a coordinate
		unless origin.size == 3
			Cuby::error("Keyword 'periodic_box_origin' has to be an array of three numbers.")
		end
		origin = Coordinate.from_array(origin)

		# Build the box
		case boxarray.size
		when 1
			return PeriodicBoxRectangular.new_cubic(boxarray.first, origin)
		when 3
			a,b,c = boxarray
			return PeriodicBoxRectangular.new(a,b,c, origin)
		else
			Cuby::error("Keyword 'periodic_box' must be an array of one or three numbers.")
		end
	end

	def PeriodicBoxRectangular.new_from_geometry(geo)
		case geo.info[:filetype]
		when :pdb
			if geo.info[:geofile_pdb_cryst]
				kw, a,b,c, alpha, beta, gamma = geo.info[:geofile_pdb_cryst].strip.split.map{|x| Settings.to_float(x)}
				if alpha != 90.0 || beta != 90.0 || gamma != 90.0
					Cuby::error("The periodic box defined in the PDB file is not recatangular")
				end
				return PeriodicBoxRectangular.new(a,b,c)
			else
				Cuby::error("The PFB file does not contain periodic box information")
			end
		else
			Cuby::error("Periodic box information can be read only from PDB files")
		end
	end

	#========================================================================
	# Derived properties
	#========================================================================
	
	def lattice_vectors
		# Return an array of three lattice vectors (Coordinates)
		return [
			Coordinate[@dimensions.x, 0, 0],
			Coordinate[0, @dimensions.y, 0],
			Coordinate[0, 0, @dimensions.z]
		]
	end

	def volume
		return @dimensions.x * @dimensions.y * @dimensions.z
	end

	#========================================================================
	# Info
	#========================================================================
	
	def inspect
		return "#{self.class} ##{self.object_id}:\n   dimensions: #{@dimensions}\n   origin: #{@origin}"
	end
	
	#========================================================================
	# Access to the periodic systems
	#========================================================================
	
	def periodic_distance(coord1, coord2)
		# Difference between coordinates
		dx = coord1 - coord2
		# Fit it into the box
		shortest = Coordinate[
			dx.x % @dimensions.x,
			dx.y % @dimensions.y,
			dx.z % @dimensions.z
		]
		# Find shorter distance across box boundary
		shortest.x = @dimensions.x - shortest.x if shortest.x > @dimensions.x / 2
		shortest.y = @dimensions.y - shortest.y if shortest.y > @dimensions.y / 2
		shortest.z = @dimensions.z - shortest.z if shortest.z > @dimensions.z / 2
		# Calculate distance
		return shortest.abs
	end

	def iterate_periodic_pairs(geometry, cutoff)
		# Iterates i, j, distance
		if cutoff > @dimensions.to_a.min / 2
			raise "Cutoff must fit into the periodic box (max. cutoff is #{@dimensions.to_a.min / 2})"
		end

		geometry.each_index{|i|
			i.times{|j|
				d = periodic_distance(geometry[i], geometry[j])
				if d <= cutoff
					yield(i, j, d)
				end
			}
		}
	end

	#========================================================================
	# Coordinate transformations
	#========================================================================
	
	def reimage!(cart_coord)
		# Move a cartesian coordinate to the box
		# Origin shift
		shifted = cart_coord - @origin
		# Get modulus
		modulus = Coordinate[
			shifted.x % dimensions.x,
			shifted.y % dimensions.y,
			shifted.z % dimensions.z
		]
		# Update coordinate
		cart_coord.set_coord(origin + modulus)
		return nil
	end

	def reimage_translation_vector(cart_coord)
		# Get a vector which moves a coordinate into the box
		# Origin shift
		shifted = cart_coord - @origin
		# Get modulus
		modulus = Coordinate[
			shifted.x % dimensions.x,
			shifted.y % dimensions.y,
			shifted.z % dimensions.z
		]
		# Translation vector: new coordinate - original
		return origin + modulus - cart_coord
	end
end
