################################################################################
#
# Module GeoFile::SDF
#
# Author: Jan Rezac
# Date created: 2009-08-17
# License: Cuby license
# Description: SDF file reader
# Status: Working
#
################################################################################

#: The SDF reader is limited ro reading the xyz coordinates from the file,
#: it ignores the connectivity and other information.

# Useful info on .sdf format:
# http://www.nonlinear.com/progenesis/sdf-studio/v0.9/faq/sdf-file-format-guidance.aspx
#
# https://en.wikipedia.org/wiki/Chemical_table_file

module GeoFile
module SDF

	ATOM_CHG_TABLE = {
		7 => -3,
		6 => -2,
		5 => -1,
		0 => 0,
		3 => 1,
		2 => 2,
		1 => 3
		# 4 => 0 # Doublet radical
	}

	ATOM_CHG_TABLE_WRITE = {
		-3 => 7,
		-2 => 6,
		-1 => 5,
		 0 => 0,
		 1 => 3,
		 2 => 2,
		 3 => 1
		# 0 => 4 # Doublet radical
	}

	def SDF.read(geometry, arguments = {})
		#: Reads SDF file. Arguments:
		#* :file - could be either File (file handle) or string (file name)
		#* :read_into_geometry

		file = arguments[:file]
		read_into = arguments[:read_into_geometry]

		close = false
		filename = "IO stream"
		if file.class == String
			filename = file
			close = true
			file = File.new(file,"r")
		end

		unless file.kind_of?(IO) || file.kind_of?(StringIO)
			raise(TypeError,"File must be IO object (is #{file.class})")
		end

		if read_into
			readgeo = Geometry.new
		else
			readgeo = geometry
			geometry.clear # delete old atoms
		end

		# Read the 3 comments
        sdf_comments = []
		3.times {
			firstline = file.gets
            sdf_comments << firstline.chomp
			return false if firstline == nil
		}

		# Read the information line
		firstline = file.gets
		return false if firstline == nil
		info = firstline.chomp
		natom = info[0..2]
		nconnect = info[3..5].to_i
        sdf_counts = []
        11.times{|i|
          sdf_counts[i] = info[(0+3*i)..(2+3*i)].to_i
        }

		# Try to get number of atoms
		if natom =~ /^ *[0-9]+$/
			header_natom = natom.to_i
		else
			Cuby::error('Invalid SDF file - can''t read number of atoms record')
		end

		# Read the geometry
        sdf_atom_flags = []
		header_natom.times{|i|
			if (line = file.gets) == nil
				Cuby::error('Invalid SDF file - it contains less atoms than specified in the header')
			end
			line.chomp!
			if matchdata =  / *([-]?[0-9]+\.[0-9]+) +([-]?[0-9]+\.[0-9]+) +([-]?[0-9]+\.[0-9]+) +([a-zA-Z]+) +(.*)/.match(line)
				element = matchdata[4]
                flags = matchdata[5]
				element = "X" if element == "Lp"
				readgeo << Atom.new(element, matchdata[1].to_f, matchdata[2].to_f, matchdata[3].to_f, {:file_index => i})
                sdf_atom_flags << flags.split.map{|x| x.to_i}
			else
				Cuby::error("Invalid SDF file - wrong format of atom record:\n'#{line}'")
			end

		}

		# Connectivity
		sdf_bonds = {}
        sdf_bond_flags = {}
		nconnect.times{|i|
			if (line = file.gets) == nil
				Cuby::error('Invalid SDF file - it contains less bonds than specified in the header')
			end
			line.chomp!
			a1 = line[0..2].to_i
			a2 = line[3..5].to_i
			order = line[6..8].to_i
			sdf_bonds[[a1,a2]] = order
            flags = line[9..20].split.map{|x| x.to_i}
            sdf_bond_flags[[a1,a2]] = flags
		}

		# Get to the end of structure, read and save lines with charges
		lines_chg = [] # Charged atom records "M CHG"
		atomic_charges = [] # Atomic charge records, 2-line
        while (line = file.gets)
            if line =~ /^M  CHG/
                lines_chg << line.chomp
            end
			if line =~ /^A/
				num = line.strip.split[1].to_i
				ch = file.gets.to_f
				atomic_charges[num-1] = ch
			end
            break if line =~ /M  *END/
        end

		# Process charges
		sdf_charges = {}
		begin
			if lines_chg.size > 0
				total_charge = 0
				lines_chg.each{|lx|
                    l = lx.dup
					l.slice!(0..5) # remove "M  CHG"
					n = l.slice!(0..2).to_i # number of records
					n.times{
						index = l.slice!(0..3) # Index of atom with charge
						charge = l.slice!(0..3) # Charge size
						sdf_charges[index.to_i-1] = charge.to_i
						total_charge += charge.to_i
					}
				}
			end
		rescue
			Cuby::warning("SDF file parser: error reading atom charges, charge information may not be valid\nfile: #{filename}\nline: #{file.lineno}")
		end

		# Atomic charges
		if atomic_charges.size != 0
			if readgeo.size == atomic_charges.size
				readgeo.each_with_index{|atom, i|
					atom.properties[:charge] = atomic_charges[i]
				}
			else
				Cuby::error("Reading atomic charges from SDF file - no. of charges different from no. of atoms")
			end
		end


		# Read properties
		sdf_data = {}
		data_name = nil
		while (line = file.gets)
			begin
				line.strip!
				break if line =~ /^\$\$\$\$/
				if line =~ /^> *<.*>/
					data_name = /<(.*)>/.match(line)[1]
				elsif data_name && line != ""
					if sdf_data[data_name]
						sdf_data[data_name] += "\n" + line
					else
						sdf_data[data_name] = line
					end
				end
			rescue
				Cuby::warning("SDF file parser: error reading data fields\nfile: #{filename}\nline: #{file.lineno}")
				data_name = nil
			end
		end

		# Read into option
		if read_into
			geometry.coords_from_geometry(readgeo)
		end

        # Store misc. sdf data
        geometry.info[:sdf_comments] = sdf_comments
        geometry.info[:sdf_counts] = sdf_counts
        geometry.info[:sdf_atom_flags] = sdf_atom_flags
        geometry.info[:sdf_bond_flags] = sdf_bond_flags
        geometry.info[:sdf_chg_lines] = lines_chg
		geometry.info[:sdf_charges] = sdf_charges

		# Store charge
		if total_charge
			geometry.info[:sdf_total_charge] = total_charge
			geometry.info[:geofile_charge] = total_charge
		end

		# Store bonds
		geometry.info[:sdf_bonds] = sdf_bonds
		
		# Store SDF data
		geometry.info[:sdf_data] = sdf_data

		# Charge from data
		if geometry.info[:sdf_data]["charge"]
			geometry.info[:geofile_charge] = sdf_data["charge"].to_i
		end

		return true
	end

	def SDF.write(geometry, arguments = {})
		#: Writes geometry in xyz format. Arguments:
		#* :file - file handle or name, default is $stdout
		#* :append (true/false), default is false

		arguments[:file] = $stdout if arguments[:file].nil?
		arguments[:append] = false if arguments[:append].nil?
		file = arguments[:file]

        # Open the file
		close = false
		if file.class == String
			close = true
			mode="w+"
			mode = "a+" if arguments[:append] == true
			file = File.new(file,mode)
		end
		unless file.kind_of?(IO)
			raise(TypeError,"File must be IO object")
		end

		# Header
        if geometry.info[:sdf_comments]
          if geometry.info[:sdf_comments].class == Array && geometry.info[:sdf_comments].size == 3
            file.puts geometry.info[:sdf_comments]
          else
            raise "Wrong structure of SDF comments"
          end
        else
          file.puts("SDF file generated by Cuby")
          file.puts
          file.puts
        end

        # Collect other SDF data
        if geometry.info[:sdf_bonds]
          bonds = geometry.info[:sdf_bonds]
        else
          Cuby::warning("SDF file write - bonds will be generated by Cuby, all considered as single bonds")
          bonds = {}
          conn = Connectivity.new(geometry)
          conn.each_bond{|i,j|
            bonds[[j+1,i+1]] = 1
          }
        end

        # First line
        file.printf("%3d", geometry.size)
        file.printf("%3d", bonds.size)
        9.times{|i|
          if geometry.info[:sdf_counts] && geometry.info[:sdf_counts][i+2]
            x = geometry.info[:sdf_counts][i+2]
          else
            x = 0
          end
          file.printf("%3d", x)
        }
        file.puts " V2000"

		# Read or create empty atom flags
		if geometry.info[:sdf_atom_flags]
	        flags = geometry.info[:sdf_atom_flags]
			geometry.size.times{|i| 
				flags[i] = Array.new(12,0) unless flags[i]
			}
		else
			flags = []
			geometry.size.times{|i| flags[i] = Array.new(12,0)}
		end

		# Update atom flags from charges if available
		geometry.each_with_index { |atom, i|
			if geometry.info[:sdf_charges] && geometry.info[:sdf_charges][i]
				flags[i][1] = ATOM_CHG_TABLE_WRITE[geometry.info[:sdf_charges][i]]
			else
				flags[i][1] = 0
			end
		}

        # Coordinates
		geometry.each_with_index { |atom, i|
          file.printf("%10.4f%10.4f%10.4f", atom.x, atom.y, atom.z)
          file.printf(" %-2s", atom.element.to_s)
          12.times{|j| 
            if flags[i][j]
              file.printf("%3d",flags[i][j])
            else
              file.printf("%3d",0)
            end
          }

          file.puts
		}

        # Bonds
        bonds.each_pair{|atoms, order|
          file.printf("%3d%3d%3d", atoms[0], atoms[1], order)
          4.times{|i|
            if geometry.info[:sdf_bond_flags] && geometry.info[:sdf_bond_flags][atoms][i]
              x = geometry.info[:sdf_bond_flags][atoms][i]
            else
              x = 0
            end
            file.printf("%3d", x)
          }
          file.puts
        }

        # Charges
        if geometry.info[:sdf_chg_lines]
          geometry.info[:sdf_chg_lines].each{|l| file.puts(l)}
        end

        # END
        file.puts "M  END"

        # Data
        if geometry.info[:sdf_data]
          geometry.info[:sdf_data].each_pair{|name, value|
            file.puts ">  <#{name}>"
            file.puts value
            file.puts
          }
        end

        # File end
        file.puts "$$$$"


		file.close if close
		return nil
	end
end
end
