class Clustering
	attr_reader :items
	attr_reader :clusters 
	# Cluster linkage:
	# :max -> Complete-linkage clustering (default)
	# :min -> Single-linkage clustering
	# :avg -> Mean or average linkage clustering / UPGMA (Unweighted Pair Group Method with Arithmetic Mean)
	attr_accessor :cluster_linkage # :max (default), :avg, :min
	

	def initialize(items, mode, threshold, distance_proc)
		# Save mode and threshold
		@mode = mode
		unless [:no_clusters, :dist].include?(mode)
			raise "Mode can be only :no_clusters or :dist"
		end
		@threshold = threshold
		@cluster_linkage = :max

		# Save the proc calculating the distance
		@dist_proc = distance_proc

		# Save items
		@items = items

		# Calculate item distances
		@item_distances = []
		@items.each_index{|i|
			@item_distances[i] = []
			i.times{|j|
				# j < i
				@item_distances[i][j] =  @dist_proc.call(@items[i], @items[j])
			}
		}

		# Initialize clusters - each item has its own cluster
		@clusters = []
		@items.each_index{|i|
			@clusters[i] = [i]
		}

		# Copy item distances to cluster distances
		@cluster_distances = []
		@items.each_index{|i|
			@cluster_distances[i] = []
			i.times{|j|
				# j < i
				@cluster_distances[i][j] = @item_distances[i][j]
			}
		}
	end

	#=======================================================================
	# Module methods
	#=======================================================================

	# Pearson correlation distance between two vectors
	def Clustering.pearson_dist(a,b)
		avg_a = a.sum / a.size
		avg_b = b.sum / b.size

		sum1 = 0.0
		sum2 = 0.0
		sum3 = 0.0

		a.each_index{|i|
			sum1 += (a[i] - avg_a) * (b[i] - avg_b)
			sum2 += (a[i] - avg_a)**2
			sum3 += (b[i] - avg_b)**2
		}

		return 1.0 - sum1 / Math::sqrt(sum2 * sum3)
	end

	#=======================================================================
	# Internal methods
	#=======================================================================
	
	def cluster_distance(i,j)
		return @cluster_distances[i][j]
	end

	def cluster_distance_calc(i,j)
		case @cluster_linkage
		when :max
			return cluster_distance_calc_max(i,j)
		when :min
			return cluster_distance_calc_min(i,j)
		when :avg
			return cluster_distance_calc_avg(i,j)
		else
			raise "Unknown cluster linkage"
		end
	end

	def cluster_distance_calc_max(i,j)
		maxdist = -9.9e99
		@clusters[i].each{|index1|
			@clusters[j].each{|index2|
				# Fix order of indexes
				if index2 < index1
					i1 = index1; i2 = index2
				else
					i1 = index2; i2 = index1
				end
				dist = @item_distances[i1][i2]
				if dist > maxdist
					maxdist = dist
				end
			}
		}
		return maxdist
	end

	def cluster_distance_calc_min(i,j)
		mindist = 9.9e99
		@clusters[i].each{|index1|
			@clusters[j].each{|index2|
				# Fix order of indexes
				if index2 < index1
					i1 = index1; i2 = index2
				else
					i1 = index2; i2 = index1
				end
				dist = @item_distances[i1][i2]
				if dist < mindist
					mindist = dist
				end
			}
		}
		return mindist
	end

	def cluster_distance_calc_avg(i,j)
		sum = 0.0
		count = 0.0
		@clusters[i].each{|index1|
			@clusters[j].each{|index2|
				# Fix order of indexes
				if index2 < index1
					i1 = index1; i2 = index2
				else
					i1 = index2; i2 = index1
				end
				sum += @item_distances[i1][i2]
				count += 1
			}
		}
		return sum/count
	end

	def delete_from_distances(i)
		@cluster_distances.each{|row|
			row.delete_at(i)
		}
		@cluster_distances.delete_at(i)
	end

	#=======================================================================
	# Public methods
	#=======================================================================
	
	def cluster!
		# Stop if no. of clusters is reached
		if @mode == :no_clusters
			return if @clusters.size <= @threshold
		end
		# Find closest clusters
		min = 9.9e99
		a = nil
		b = nil
		@clusters.each_index{|i|
			@clusters.each_index{|j|
				next if j >= i
				d = cluster_distance(i,j)
				if d < min
					min = d
					a = j
					b = i
				end
			}
		}
		if @mode == :dist && min >= @threshold
			return
		end
		# Now  a < b
		# Merge clusters a and b
		@clusters[a] = @clusters[a] | @clusters[b]
		# Delete cluster b
		@clusters.delete_at(b)
		# Delete b from cluster distances
		delete_from_distances(b)
		# Recalculate distances for a
		@clusters.each_index{|j|
			next if a == j
			j,i = [a,j].sort
			dist = cluster_distance_calc(i,j)
			@cluster_distances[i][j] = dist
		}
		cluster!
	end

	def cluster_representatives_by_dist
		representatives = []

		@clusters.each_index{|i|
			cluster = @clusters[i]

			# Fore each item, calculate the sum of its distances to others
			dist_sums = []
			cluster.each_index{|j|
				dist_sums[j] = 0.0
				item = cluster[j]
				cluster.each_index{|x|
					next if j == x
					if cluster[x] < item # indexes must be ordered
						dist_sums[j] += @item_distances[item][cluster[x]]
					else
						dist_sums[j] += @item_distances[cluster[x]][item]
					end
				}
			}

			# Find the item with smallest sum of distances
			minvalue = dist_sums.min
			min_i = dist_sums.index(minvalue)
			representatives[i] = cluster[min_i]
		}

		return representatives
	end

	#=======================================================================
	# Cluster information
	#=======================================================================
	
	def get_cluster_sizes
		# Returns array of cluster sizes
		# Cluster size is measured as the largest distance between
		# two items in the cluster
		sizes = []
		@clusters.each{|cluster|
			size = 0.0
			cluster.each{|i|
				cluster.each{|j|
					next if j >= i
					if @item_distances[i][j] > size
						size = @item_distances[i][j]
					end
				}
			}
			sizes << size
		}
		return sizes
	end

	#=======================================================================
	# Printing
	#=======================================================================

	def print_cluster_info
		puts "No. of clusters: #{@clusters.size}"
		@clusters.each_index{|i|
			puts "Cluster #{i} has #{@clusters[i].size} items"
		}
	end

	def print_cluster_distances
		puts "Cluster distances (max, average, min)"
		@clusters.each_index{|i|
			i.times{|j|
				printf("%4d%4d%10.3f%10.3f%10.3f\n", i, j,
				       cluster_distance_calc_max(i,j),
				       cluster_distance_calc_avg(i,j),
				       cluster_distance_calc_min(i,j)
				)
			}
		}
	end

end


