################################################################################
#
# AMBER interface
#
# Author: Jan Rezac
# Date created: 2012-08-23
# License: Cuby4 license
# Description: Interface for external calculations in AMBER
# Status: Works
#
################################################################################

#===============================================================================
# Interface to the AMBER molecular mechanics package
# Commercial software
# http://ambermd.org/
#===============================================================================

# Possible improvements
# * Option to run the calculations using NAB
# * Automated charge generation using Antechamber (am1-bcc charges first)

require "erb" # ERB templating system, used for construction of the input
require "classes/tools/output_parser.rb"
require "interfaces/amber/classes/amber_prmtop.rb"

module InterfaceAmber
	#=======================================================================
	# Interface header
	#=======================================================================
	# Interface status
	DEVELOPMENT_FLAG = :ok
	DEVELOPMENT_STATUS = "All cuby3 functionality implemented, examples for all features"
	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy, :gradient, :atomic_charges, :point_charges, :point_charges_gradient, :native_opt]
	MODIFIER = false
	DIRECTORY = "AMBER"
	# Methods provided by the interface:
	METHODS = {
		:forcefield		=> [:energy, :gradient, :atomic_charges, :point_charges, :point_charges_gradient, :native_opt]
	}
	SOLVENTS = [:gbm, :igb2, :igb5, :igb7, :igb8, :pb]
	ATOMIC_CHARGES = [:forcefield]
	#=======================================================================

	#-----------------------------------------------------------------------
	# Solvation setup
	#-----------------------------------------------------------------------
	SOLVENT_MODELS = {
		:gbm => 1,
		:igb2 => 2,
		:igb5 => 5,
		:igb7 => 7,
		:igb8 => 8,
		:pb => 10
	}

	SOLVENT_RADII = {
		:gbm => "mbondi",
		:igb2 => "mbondi2",
		:igb5 => "mbondi2",
		:igb7 => "bondi",
		:igb8 => "mbondi3",
		:pb => "nil"
	}

	# Nonpolar (surface) energy for GB models
	SOLVENT_GBSA = {
		:none => 0,
		:lcpo => 1,
		:icosa => 2
	}

	# Nonpolar energy for PB model
	SOLVENT_PBSA = {
		:none => 0,
		:sasa => 1,
		:cavdisp => 2
	}
	#-----------------------------------------------------------------------


	def prepare_interface
		# Check config
		amber_check_config

		# Check geometry - it must contain the PDB data
		unless @geometry[0].properties[:pdb_atom_name]
			Cuby::error("Amber calculations are possible only on geometry from a PDB file")
		end

		# Check features not compatible with native_opt
		if @what.include?(:native_opt)
			unless @settings[:solvent_model] == :none ||  @settings[:amber_opt_mode] == :all
				Cuby::error("Native_opt  in AMBER does not support solvent unless all atoms are optimized") 
			end

			if ([:gradient, :atomic_charges, :point_charges, :point_charges_gradient] & @what).size > 1
				Cuby::error("Native_opt in AMBER does not support simultaneous calculation of other properties")
			end
		end

		# Prepare calculation directory
		if calc_dir_mkdir("sander.in", "sander.out") == :old
			calc_using_input
			return
		end

		# Write geometry
		GeoFile::AmberCrd.write(@geometry, :file => calc_dir+"/crd")

		# Generate or copy top file
		if @settings.set?(:amber_top_file)
			FileUtils.cp(@settings[:amber_top_file], calc_dir+"/prmtop")
		else
			amber_generate_prmtop
		end

		# Read top if needed
		# Atom types are needed when selecting hydrogens
		if @settings[:amber_opt_mode] == :hydrogens || @settings[:amber_opt_mode] == :hydrogens_annealing
			top = AmberPrmtop.new(in_calc_dir("prmtop"))
			@atomtypes = top.atomtypes
		end

		# Modify top
		if @settings[:amber_zero_dihedral_fc]
			top = AmberPrmtop.new(in_calc_dir("prmtop")) unless top
			top.dihedral_fc.each_index{|i|
				top.dihedral_fc[i] = 0.0
			}
			top.write_top(in_calc_dir("prmtop"))
		end

		# Close top
		top = nil

		# Write input
		if @settings.set?(:amber_input_file)
			FileUtils.cp(@settings[:amber_input_file], calc_dir+"/sander.in")
		else
			amber_write_input(calc_dir+"/sander.in")
		end

		# Save info on the system upon writing the input
		calc_writing_input
	end

	def calculate_interface
		amber_run_calculation unless @settings[:existing_calc_dir] == :read_results && FileTest.exist?(calc_dir+"/sander.out")
		if @what.include?(:native_opt)
			# Results of optimization
			results = amber_read_results_opt
		else
			# Energy and gradient
			results = amber_read_results
			results.atomic_charges = amber_read_charges if @what.include?(:atomic_charges)
			if @what.include?(:point_charges_gradient)
				results.point_charges_gradient = @point_charges.gradient
			end
		end
		return results
	end

	def cleanup_interface
		calc_dir_delete if @settings[:job_cleanup]
	end

	#=======================================================================
	# Extra methods - MM specific
	#=======================================================================
	
	def get_atomic_charges
		if ! @atomic_charges
			@atomic_charges = amber_read_charges
		end
		return @atomic_charges
	end

	#=======================================================================
	# Private methods
	#=======================================================================

	def amber_write_input(filename)

		# Load the template and set required variables
		if @what.include?(:native_opt)
			# Geometry optimization
			case @settings[:solvent_model]
			when :none
				case @settings[:amber_opt_mode]
				when :all
					template = ERB.new(IO.read(interface_dir + "/templates/sander.in_vac_opt_all.erb"), 0, '>')
				when :hydrogens
					opt_h_selections = belly_hydrogen_selection(@atomtypes)
					template = ERB.new(IO.read(interface_dir + "/templates/sander.in_vac_opt_h.erb"), 0, '>')
				when :hydrogens_annealing
					opt_h_selections = belly_hydrogen_selection(@atomtypes)
					template = ERB.new(IO.read(interface_dir + "/templates/sander.in_vac_anneal_h.erb"), 0, '>')
				when :custom_selection
					atomlist = @geometry.atomlist_from_selection(@settings[:optimize_region])
					Cuby::log.puts "Optimization of #{atomlist.size} atoms"
					opt_h_selections = belly_atom_selection(atomlist)
					template = ERB.new(IO.read(interface_dir + "/templates/sander.in_vac_opt_h.erb"), 0, '>')
				when :custom_selection_annealing
					atomlist = @geometry.atomlist_from_selection(@settings[:optimize_region])
					opt_h_selections = belly_atom_selection(atomlist)
					template = ERB.new(IO.read(interface_dir + "/templates/sander.in_vac_anneal_h.erb"), 0, '>')
				end
			else
				case @settings[:amber_opt_mode]
				when :all
					template = ERB.new(IO.read(interface_dir + "/templates/sander.in_GBM_opt_all.erb"), 0, '>')
				else
					Cuby::error "Solvent model not supported by AMBER for native_opt with amber_opt_mode != all"
				end
			end
		else
			# Single-point and gradient calculation
			if @settings[:solvent_model] == :none
				template = ERB.new(IO.read(interface_dir + "/templates/sander.in_vacuum.erb"), 0, '>')
			else
				unless SOLVENT_MODELS.has_key?(@settings[:solvent_model])
					Cuby::error "Solvent model not supported by AMBER"
				end
				template = ERB.new(IO.read(interface_dir + "/templates/sander.in_GBM.erb"), 0, '>')
			end
		end

		# Random seed for md
		if @settings.set?(:random_seed)
			random_seed = @settings[:random_seed]
		else
			random_seed = rand(9999999)
		end

		file = File.open(filename,"w+")
		file.print(template.result(binding))
		file.close
	end

	def amber_run_calculation
		# If point charges are used, and atomic charges are not available, get atomic charges
		if @point_charges && ! @atomic_charges
			@atomic_charges = amber_read_charges
		end

		# Write current geometry
		GeoFile::AmberCrd.write(@geometry, :file => calc_dir+"/crd")

		# Forcedump.dat must be deleted, otherwise is not updated
		FileUtils.rm(calc_dir + "/forcedump.dat") if FileTest.exist?(calc_dir + "/forcedump.dat")

		# Run sander
		ld_path = ""
		ld_path = " export LD_LIBRARY_PATH=#{@settings[:amber_ld_path]}:$LD_LIBRARY_PATH;" if @settings[:amber_ld_path] != ""
		if @settings[:parallel] == 1
			system("cd #{calc_dir}; export AMBERHOME=#{@settings[:amber_amberhome]};#{ld_path} #{@settings[:amber_sander_exe]} -O -i sander.in -p prmtop -o sander.out -c crd -r restrt > /dev/null 2> sander.err")
		else
			system("cd #{calc_dir}; export AMBERHOME=#{@settings[:amber_amberhome]};#{ld_path} mpirun -np #{@settings[:parallel]} #{@settings[:amber_sander_exe_mpi]} -O -i sander.in -p prmtop -o sander.out -c crd -r restrt > /dev/null 2> sander.err")
		end

		# Error check - for single-point only
		unless @what.include?(:native_opt)
			Cuby::error("Calculation failed, sander did not produce the file sander.out",self) unless FileTest.exist?(calc_dir + "/sander.out")
			unless FileTest.exist?(calc_dir + "/forcedump.dat")
				amber_check_sander_out_errors
				Cuby::error("Calculation failed, sander did not produce the file forcedump.dat",self) 
			end
		end
	end


	ENERGY_COMPONENTS = [
		:total_energy,
		:van_der_Waals,
		:electrostatics,
		:GBM,
		:bonds,
		:angles,
		:dihedrals,
		:"1-4_van_der_Waals",
		:"1-4_electrostatics"
	]

	def amber_read_results
		results = Results.new
		results.gradient = Gradient.new if @what.include?(:gradient)

		energyOK = false
		gradOK = false
		readgrad = false
		readenergy = false
		energycount = 0
		gradlinecount = 0

		case @settings[:amber_grad_component]
		when :vdw
			force_block = /Van der Waals Force/
		else
			force_block = /Total Force/
		end

		IO.foreach(calc_dir + "/forcedump.dat") { |frcline|
			frcline.chomp!
			if frcline =~ /START of Energies/
				readenergy = true
				next
			end
			if readenergy
				if frcline =~ /[-]*[0-9]+[.][0-9]+E[+-][0-9]+/
					if energycount == 0
						results.energy = ($&).to_f
					else
						results.energy_components[ENERGY_COMPONENTS[energycount]] = ($&).to_f
						if energycount == 8
							energyOK = true
							readenergy = false
						end
					end
				end
				if  frcline =~ /NaN/
					@geometry.write_pdb(:file => calc_dir + "/failed_geo.pdb")
					Cuby::error "AMBER calculation failed, found NaN as a result (calculation #{@name})"
				end
				energycount += 1
			end
			if frcline =~ force_block && @what.include?(:gradient)
				readgrad = true
				next
			end
			if readgrad 
				gx, gy, gz = frcline.strip.split(/\s+/)  # split line, whitespace separated
				results.gradient[gradlinecount] = Coordinate.new(
					-1* gx.to_f,
					-1* gy.to_f,
					-1* gz.to_f
				)
				gradlinecount += 1
				if gradlinecount == @geometry.size
					gradOK = true
					break
				else
					gradOK = false
				end
			end
		}

		# Check if the results were read successfuly
		Cuby::error("Energy not found in file \"forcedump.dat\"", self) if ! energyOK
		Cuby::error("Valid gradient not found in file \"forcedump.dat\"", self) if ! gradOK && @what.include?(:gradient)

		# Calculate the interaction with point charges
		if @point_charges
			pch_energy = @point_charges.interaction_with_molecule(@geometry, @atomic_charges, results.gradient, @what.include?(:point_charges_gradient))
			results.energy_components[:point_charges_interaction] = pch_energy
			results.energy += pch_energy
		end

		return results
	end

	def amber_check_sander_out_errors
		parser = OutputParser.new(in_calc_dir("sander.out"), @name)
		parser.add_error(/Regenerate prmtop file with Bondi radii/, "The calculation requires Bondi radii in the prmtop file.")
		parser.execute
	end

	def amber_check_leap_out_errors
		parser = OutputParser.new(in_calc_dir("tleap.out"), @name)
		parser.add_error(/Could not open file/, "Error in parmtop creation, leap could not find some file\n(check file #{in_calc_dir('tleap.out')} for details)")
		parser.add_error(/Leap added 1 missing atom/, "Error in parmtop creation, leap added an atom\n(this means the geometry does not match the residue template in the ff.,\ncheck file #{in_calc_dir('tleap.out')} for details))")
		parser.add_error(/Leap added .* missing atoms/, "Error in parmtop creation, leap added multiple atoms\n(this means the geometry does not match the residue template in the ff.)")
		parser.execute
	end

	def amber_check_config
		@settings.check_dir(:amber_amberhome)
	end

	def amber_generate_prmtop
		# Write a PDB file 
		pdb_file_name = "tmp_in.pdb"
		@geometry.write_pdb(:file => in_calc_dir(pdb_file_name))

		# Check all leaprc files
		nondefault = @settings.set?(:amber_leaprc)
		leaprc_files = []
		@settings[:amber_leaprc].each{|filename|
			filename = filename.gsub("$AMBERHOME", @settings[:amber_amberhome])
			filename.gsub!("%cuby",Cuby.install_dir)
			filename.gsub!("%interface",interface_dir)
			filename.gsub!("%amberhome",@settings[:amber_amberhome])
			filename.gsub!("%pwd",Dir.pwd)
			filename.gsub!("%local",Dir.pwd)
			filename.gsub!("%home",File.expand_path("~"))
			filename = File.expand_path(filename)
			unless FileTest.exist?(filename)
				Cuby::error("Leaprc file not found:\n#{filename}") if nondefault
			else
				leaprc_files << filename
			end
		}

		# Create a tleap input file containing the commands needed to build the prmtop
		File.open(in_calc_dir('tleap.in'),'w+') {|f| 
			leaprc_files.each{|filename|
				f.puts "#==============================================================================="
				f.puts "# From #{filename}"
				File.open(filename, "r"){|orig|
					while line = orig.gets do
						line.chomp!
						line.gsub!("%cuby",Cuby.install_dir)
						line.gsub!("%interface",interface_dir)
						line.gsub!("%amberhome",@settings[:amber_amberhome])
						line.gsub!("$AMBERHOME",@settings[:amber_amberhome])
						line.gsub!("%pwd",Dir.pwd)
						line.gsub!("%local",Dir.pwd)
						line.gsub!("%home",File.expand_path("~"))
						f.puts line
					end
				}
				f.puts
			}
			f.puts "#==============================================================================="
			f.puts "# Prepare calculation"
			f.puts "x = loadpdb #{pdb_file_name}"
			#f.puts "set default OldPrmtopFormat on" if @settings [:amber_old_top]

			# Radii specific for each solvent model
			if SOLVENT_RADII[@settings[:solvent_model]]
				f.puts "set default PBradii #{SOLVENT_RADII[@settings[:solvent_model]]}"
			end

            # Disulfide bridges
            if @settings[:amber_leap_disulfide_bridges]
              bridges = @geometry.pdb_disulfide_bridges(["CYX"])
              bridges.each{|b|
                atom1 = @geometry[b[0]]
                atom2 = @geometry[b[1]]
                f.puts "bond x.#{atom1.properties[:pdb_res_no]}.SG x.#{atom2.properties[:pdb_res_no]}.SG"
              }
            end

			# Custom leap commands
			f.puts @settings[:amber_leap_commands] if @settings[:amber_leap_commands] != ""

			# Save the topology
			f.puts "saveamberparm x prmtop restrt.tmp"

			# Optional write of processed pdb
			if Cuby::log.debug?
				Cuby::log.puts_debug("Amber calculation preparation: saving leap putput as .pdb")
				f.puts "savepdb x tmp_out.pdb"
			end

			f.puts "quit"
		}

		# Run tleap
		ld_path = ""
		ld_path = " export LD_LIBRARY_PATH=#{@settings[:amber_ld_path]}:$LD_LIBRARY_PATH;" if @settings[:amber_ld_path] != ""
		system("cd #{calc_dir}; export AMBERHOME=#{@settings[:amber_amberhome]};#{ld_path} #{@settings[:amber_amberhome]}/bin/tleap -f tleap.in > tleap.out 2> tleap.err")

		# Check if prmtop was crated
		if FileTest.exist?(in_calc_dir("prmtop"))
			if File.size(in_calc_dir("prmtop")) == 0
				amber_check_leap_out_errors
				Cuby::error("Amber prmtop is empty")
			end
		else
			Cuby::error("Amber prmtop file was not created.")
		end
		amber_check_leap_out_errors

		# Delete unnecessary files
		File.delete(in_calc_dir("restrt.tmp"))

		return nil
	end

	def amber_read_charges
		# Reads the atomic charges from the prmtop file
		charges = AtomicCharges.new(:forcefield, @geometry.size)

		top = AmberPrmtop.new(in_calc_dir("prmtop"))
		Cuby::error "Can't load MM charges from prmtop - different number of atoms (system #{name})" if @geometry.size != top.natom

		@geometry.each_index { |i|
			charges[i] = top.charges[i]
		}

		return charges
	end

	def amber_read_results_opt
		results = Results.new

		if @settings[:amber_opt_mode] == :hydrogens_annealing || @settings[:amber_opt_mode] == :custom_selection_annealing
			# MD run - no energy to read
			results.energy = 0.0

			# Read restrt to get geometry
			GeoFile::AmberCrd.read_into_geo(@geometry, :file => in_calc_dir("restrt"))
		else
			# Optimization - read energy and check convergence,
			# read geometry

			# Read mdinfo to get energy and number of cycles
			mdinfo = IO.readlines(in_calc_dir("mdinfo"))
			unless mdinfo[2] =~ /NSTEP/
				Cuby::error("Error reading file 'mdinfo' produced by AMBER")
			end
			nstep, energy, rms = mdinfo[3].strip.split.map{|x| x.to_f}
			nstep = nstep.to_i
			Cuby::log.puts_debug("AMBER optimization finished in #{nstep} cycles")

			# Check convergence
			if nstep == @settings[:maxcycles] - 1
				Cuby::error("Native optimization in AMBER did not converge")
			end

			# Energy as a sum of components - to prevent rounding
			energy_sum = 0.0
			mdinfo[5..7].each{|line|
				line.strip.split.each{|item|
					if item =~ /\.[0-9][0-9][0-9][0-9]/
						energy_sum += item.to_f
					end
				}
			}

			# Save energy
			results.energy = energy_sum

			# Read restrt to get geometry
			GeoFile::AmberCrd.read_into_geo(@geometry, :file => in_calc_dir("restrt"))
		end

		return results
	end

	#=======================================================================
	# Tools for native_opt
	#=======================================================================
	
	def belly_hydrogen_selection(atomtypes)
		# Build lists of hydrogen for belly constraints

		# List of hydrogen atom types
		h_atomtypes = []
		atomtypes.each{|t|
			h_atomtypes |= [t] if t =~ /^[0-9]*[Hh]/
		}
		# Make the hydrogen selections, lists must be broken into batches with up to 10 records
		opt_h_selections = []
		h_atomtypes.each_index{|i|
			# Batch header
			if i % 10 == 0
				opt_h_selections << "Hydrogen selection #{i/10}"
				opt_h_selections << "FIND"
			end
			# Selections
			opt_h_selections << "* #{h_atomtypes[i]} * *"
			# Batch end
			if (i % 10 == 9) || (i == h_atomtypes.size - 1)
				opt_h_selections << "SEARCH"
				opt_h_selections << "RES 1 99999"
				opt_h_selections << "END"
			end

		}
		opt_h_selections << "END"
		return opt_h_selections.join("\n")
	end

	def belly_atom_selection(atomlist)
		# Build list of atoms for belly constraints from cuby's atomlist

		# Make the hydrogen selections, lists must be broken into batches with up to 10 records
		opt_h_selections = []
		atomlist.each_index{|i|
			atnum = atomlist[i] + 1
			opt_h_selections << "Atom selection #{i}"
			opt_h_selections << "FIND"
			opt_h_selections << "* * * *"
			opt_h_selections << "SEARCH"
			opt_h_selections << "ATOM #{atnum} #{atnum}"
			opt_h_selections << "END"
		}
		opt_h_selections << "END"
		return opt_h_selections.join("\n")
	end

end
