################################################################################
#
# Template interface
#
################################################################################


# To read the output of the calculation easily, we will use the parser provided
# by the framework
require "classes/tools/output_parser.rb"
require "interfaces/dispersion3/classes/d3_parameters"

module InterfaceDftd3
	#=======================================================================
	# Interface header
	#=======================================================================
	# Development status
	DEVELOPMENT_FLAG = :ok
	DEVELOPMENT_STATUS = "Under development"

	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy, :gradient, :ghost_atoms]
	MODIFIER = true
	IGNORE_AS_MODIFIER = [:atomic_charges, :point_charges, :point_charges_gradient, :mos]

	# Suffix used for the calculation directory
	DIRECTORY = "DFTD3"

	#=======================================================================
	# Local constants
	#=======================================================================
	INPUT_FILE = ".dftd3par.local"
	OUTPUT_FILE = "dftd3.out"
	ERROR_FILE = "dftd3.err"

	#=======================================================================
	# Interface data - shared by all instances
	#=======================================================================
	@@parameters = nil

	#=======================================================================
	# Interface methods
	#=======================================================================
	
	def prepare_interface
		# Check for ghost atoms
		@geometry.each{|atom|
			if atom.properties[:ghost]
				Cuby::error "The DFTD3 interface does not support ghost atoms - try dispersion3 instead"
			end
		}

		# Load parameters
		@@parameters = D3Parameters.load(@settings, interface_dir + "/../dispersion3") unless @@parameters
		# Determine parameters automatically from parent calculation setup
		@@parameters.defaults_to_settings(@settings)

		# Check for limited features compared to Dispersion3
		# Must be done after loading default parameters!
		unless [:zero, :bj].include?(@settings[:d3_damping])
			Cuby::error "The DFTD3 interface supports only zero and bj damping"
		end

		if calc_dir_mkdir(INPUT_FILE, OUTPUT_FILE) == :old
			# When the old input is used, the following method
			# is called to check if it is applicable to the current
			# calculation.
			calc_using_input
			return
		end

		# Write input and geometry
		dftd3_write_input
		dftd3_write_geo

		# Save info on the system upon writing the input
		calc_writing_input

		return nil
	end

	def calculate_interface
		# The calculation is run unless reading existing results is
		# requested in the input
		dftd3_run unless @settings[:existing_calc_dir] == :read_results && FileTest.exist?(calc_dir+"/dftd3.in.out")

		# The dftd3_read method reads the output file and builds the
		# Results object
		results = dftd3_read

		# H-H repulsion for SQM
		if @settings[:d3_hh_fix]
			require "interfaces/dispersion3/classes/d3_hh_fix"
			@hh_fix = D3HHFix.new(@settings)
			@hh_fix.calc_add_to_results(geometry, results)
		end

		return results
	end

	def cleanup_interface
		calc_dir_delete if @settings[:job_cleanup]
	end

	#=======================================================================
	# Private methods
	#=======================================================================
	
	def dftd3_write_input
		filename = in_calc_dir(INPUT_FILE)
		f = File.open(filename,"w+")

		inpvalues = []
		case @settings[:d3_damping]
		when :zero
			version = 3
			# s6 sr6 s8 sr8 alpha6 version
			inpvalues << @settings[:d3_s6]
			inpvalues << @settings[:d3_sr6]
			inpvalues << @settings[:d3_s8]
			inpvalues << @settings[:d3_sr8]
			inpvalues << @settings[:d3_alpha6]
			inpvalues << version
		when :bj
			version = 4
			# s6 a1 s8 a2 alpha6 version
			inpvalues << @settings[:d3_s6]
			inpvalues << @settings[:d3_a1]
			inpvalues << @settings[:d3_s8]
			inpvalues << @settings[:d3_a2]
			inpvalues << 1.0 # placeholder
			inpvalues << version
		else
			Cuby::erro "D3 interface supports only BJ and Zero damping"
		end

		f.puts inpvalues.map{|x| x.to_s }.join(" ")

		f.close
	end

	def dftd3_write_geo
		# Open the geometry file
		@geometry.write_xyz(:file => in_calc_dir("geo.xyz"))
	end

	def dftd3_run
		# Update the geometry - it may have changed since the calculation
		# was prepared
		dftd3_write_geo

		# Command-line options for DFTD3
		options = "geo.xyz"
		options += " -grad" if @what.include?(:gradient) 
		options += " -abc" if @settings[:d3_3body]

		# Prepare the command for running dftd3
		# First, go to the calculation directory
		command = "cd #{calc_dir};"
		# And finally, the command that runs dftd3 on the input file
		command << "#{@settings[:dftd3_exe]} #{options} > #{OUTPUT_FILE}  2> #{ERROR_FILE};"

		# Run dftd3, exit if it returned bad exit code
		unless system(command)
			Cuby::error "DFTD3 returned nonzero exit code (calculation #{@name})"
		end
	end

	def dftd3_read 
		# Read the output produced by successful MOPAC run
		# Returns new instance of the Results class
	
		# Create empty Results object
		results = Results.new

		# Create parser acting on the ouput file
		parser = OutputParser.new(in_calc_dir(OUTPUT_FILE), @name)

		parser.add_pattern(:energy, /^ +Edisp.*: * +([-]*[0-9]*\.[0-9]+)/, :get => :submatch)

		# Run parser
		parser.execute

		# Save the value extracted by the parser to the Results object
		# The result is in kcal/mol, no conversion is needed
		results.energy = parser[:energy]

		# Read gradient from a separate file
		if @what.include?(:gradient)
			results.gradient = Gradient.new
			gradfile = File.open(in_calc_dir("dftd3_gradient"))
			while line = gradfile.gets
				x,y,z = line.strip.split.map{|n| n.to_f * HARTREE2KCAL / BOHR2ANGSTROM} 
				results.gradient << Coordinate[x,y,z]
			end
			gradfile.close
		end

		return results
	end

end
