require "classes/math/factorial.rb"

module DampingFunction

	#=======================================================================
	# Zero damping
	#=======================================================================

	# The same function is used for both C6 and C8 term,
	# just the parameters differ
	
	def DampingFunction.zero(r, r0, sr, alpha)
		return 1.0 / (1.0 + 6.0 * (r/(sr*r0))**(-1.0 * alpha))
	end

	def DampingFunction.zero_deriv(r, r0, sr, alpha)
		z = (r/(sr*r0))**alpha
		return alpha * z / (r *(z+6.0)) - alpha * z**2/ (r*(z+6.0)**2)
	end

	#=======================================================================
	# Becke-Johnson damping
	#=======================================================================
	
	def DampingFunction.bj(r, r0, a1, a2, order)
		fd = a1*r0 + a2
		return r**order / (r**order + fd**order)
	end

	def DampingFunction.bj_deriv(r, r0, a1, a2, order)
		fd = a1*r0 + a2
		return (order * r**(order-1)) /
		       (r**order + fd**order) - 
		       (order * r**(2*order-1)) /
		       (r**order + fd**order)**2
	end
	
	#=======================================================================
	# Error function damping
	#=======================================================================
	def DampingFunction.erf(r, r0, a1, alpha)
		return Math::erf(a1 * r / r0)**alpha
	end

	def DampingFunction.erf_deriv(r, r0, a1, alpha)
		# Temporary solution: numerical gradient
		step = 1e-4
		deriv = (DampingFunction.erf(r+step, r0, a1, alpha) - DampingFunction.erf(r-step, r0, a1, alpha)) / (step * 2)
		return deriv
	end


	#=======================================================================
	# Tang-Toennies damping
	#=======================================================================
	
	def DampingFunction.tt(r, r0, a1, a2, order)
		b = r0 * a1 + a2
		sum = 0.0
		(0..order).each{|k|
			sum += (b * r)**k / Factorial.n(k)
		}

		return 1.0 - Math::exp(-b * r) * sum
	end

	def DampingFunction.tt_deriv(r, r0, a1, a2, order)
		# Temporary solution: numerical gradient
		step = 1e-4
		deriv = (DampingFunction.tt(r+step, r0, a1, a2, order) - DampingFunction.tt(r-step, r0, a1, a2, order)) / (step * 2)
		return deriv
	end

	#=======================================================================
	# Optimized power damping
	#=======================================================================
	
	def DampingFunction.op(r, r0, a1, a2, beta)
		fd = a1*r0 + a2
		return r**beta / (r**beta + fd**beta)
	end

	def DampingFunction.op_deriv(r, r0, a1, a2, beta)
		fd = a1*r0 + a2
		return (beta*r**(beta-1))/(r**beta+fd**beta)-(beta*r**(2*beta-1))/(r**beta+fd**beta)**2
	end

	#=======================================================================
	# CSO damping
	#=======================================================================
	
	def DampingFunction.cso(r, r0, a1, s6)
		a2 = 2.5
		a3 = 0.0
		a4 = 6.25
		return r**6 / (r**6 + (a3 * r0 + a4)**6) * 
			(1.0 + (a1 / (s6 * (1.0 + Math::exp(r - a2*r0)))))
	end

	def DampingFunction.cso_deriv(r, r0, a1, s6)
		a2 = 2.5
		a3 = 0.0
		a4 = 6.25
		r0a = a3*r0+a4
		ex = Math::exp(r-a2*r0)
		r6 = r**6
		return (6*r**5*(a1/((ex+1.0)*s6)+1.0)) / ((r0a)**6+r6) -
			(6*r**11*(a1/((ex+1.0)*s6)+1.0))/((r0a)**6+r6)**2 -
			(a1*r6*ex)/(((r0a)**6+r6)*(ex+1.0)**2*s6)
	end

end
