################################################################################
#
# Interface Dispersion3
#
# Author: Jan Rezac
# Date created: 2012-12-14
# License: Cuby4 license
# Description: Implementation of Grimme's DFT-D3
# Status: Works
#
################################################################################

#===============================================================================
# Implementation of Grimme's DFT-D3
#===============================================================================

require "classes/geometry/hybridization"
require "classes/math/func"
require "interfaces/dispersion3/classes/d3_parameters"
require "interfaces/dispersion3/classes/damping_function.rb"
require "classes/compatibility/yaml_loader.rb"

# References
# OP damping: DOI: 10.1021/acs.jctc.7b00176


module InterfaceDispersion3
	#=======================================================================
	# Interface header
	#=======================================================================
	# Interface status
	DEVELOPMENT_FLAG = :ok
	DEVELOPMENT_STATUS = "OK"
	# Interface information
	INTERFACE = :calculation
	CAPABILITIES = [:energy, :gradient, :ghost_atoms]
	IGNORE_AS_MODIFIER = [:atomic_charges, :point_charges, :point_charges_gradient, :mos]
	MODIFIER = true
	#=======================================================================

	#=======================================================================
	# C6 record class used in the YAML data file
	#=======================================================================
	class C6Record
		attr_accessor :c6
		attr_accessor :hyb_a
		attr_accessor :hyb_b
	end

	#=======================================================================
	# Interface data - shared by all instances
	#=======================================================================
	@@data = nil
	@@parameters = nil

	#=======================================================================
	# Interface methods
	#=======================================================================

	def prepare_interface
		# Load D3 data
		d3_data_load
		# Patch data
		d3_data_patch

		# Load parameters
		@@parameters = D3Parameters.load(@settings, interface_dir) unless @@parameters

		# Determine parameters automatically from parent calculation setup
		@@parameters.defaults_to_settings(@settings)
	end

	def calculate_interface
		# Calculate
		results = d3_dispersion_e_g(@what.include?(:gradient))
		return results
	end

	#=======================================================================
	# Private methods: Parameter setup
	#=======================================================================
	
	def d3_data_load
		# Load the data file if not available already
		unless @@data
			# Psych YAML parser is very slow and the data file is large
			# Therefore, we cache the serialized data separately
			if FileTest.exist?(Cuby::local_data_dir + "/dispersion3_data.dat")
				# Load the raw data
				Cuby::log.puts_debug "Dispersion3: cached data used"
				df = File.open(Cuby::local_data_dir + "/dispersion3_data.dat")
				@@data = Marshal.load(df)
				df.close
			else
				# load the database
				Cuby::log.puts_debug "Dispersion3: rebuilding data cache"
				f = File.open(interface_dir + "/dispersion3_data.yaml")
				@@data = YAMLLoader.load(f, [C6Record])
				f.close
				# dump the data to cache
				df = File.open(Cuby::local_data_dir + "/dispersion3_data.dat","w+")
				Marshal.dump(@@data, df)
				df.close
			end
		end
		
		# Make the data available locally
		@covalent_radii = @@data["covalent_radii"]
		@c6 = @@data["c6"]
		@r2r4 = @@data["r2r4"]
		@r0 = @@data["r0ab"]

		# Experimental: make all hybridization numbers integer
		if @settings[:development][:d3_integer_hybridizations]
			@c6.each_pair{|e1,h1|
				h1.each_pair{|e2,arr|
					arr.each_index{|i|
						arr[i].hyb_a = arr[i].hyb_a.round.to_f
						arr[i].hyb_b = arr[i].hyb_b.round.to_f
					}
				}
			}
		end
	end

	def d3_data_patch
		#: Load a patch which could override c6 coefficients
		if @settings.set?(:d3_data_patch)
			# File name - first local, then interface dir
			if FileTest.exist?(@settings[:d3_data_patch])
				fn = @settings[:d3_data_patch]
			elsif FileTest.exist?(interface_dir + "/data/" + @settings[:d3_data_patch])
				fn = interface_dir + "/data/" + @settings[:d3_data_patch]
			else
				Cuby::error "D3 data patch file not found neither locally nor in interface directory."
			end

			f = File.open(fn)
			data_patch = YAMLLoader.load(f, [C6Record])
			f.close
			
			@c6_patch = data_patch["c6"]
		else
			@c6_patch = nil
		end
	end

	def d3_parameters_from_input
		#: Read parameters from input if present, override what has been
		#: set up previously

		@s6 = @settings[:d3_s6]

		case @settings[:d3_damping]
		when :zero
			@s8 = @settings[:d3_s8]
			@sr6 = @settings[:d3_sr6]
			@sr8 = @settings[:d3_sr8]
			@alpha = @settings[:d3_alpha6]
			if @settings.set?(:d3_alpha8)
				@alpha8 = @settings[:d3_alpha8]
			else
				@alpha8 = @alpha + 2.0
			end
		when :bj, :tt, :bjr0, :bjvdw
			@s8 = @settings[:d3_s8]
			@a1 = @settings[:d3_a1]
			@a2 = @settings[:d3_a2]
		when :erf
			@s8 = @settings[:d3_s8]
			@a1 = @settings[:d3_a1]
			@alpha = @settings[:d3_alpha6]
			if @settings.set?(:d3_alpha8)
				@alpha8 = @settings[:d3_alpha8]
			else
				@alpha8 = @alpha
			end
		when :op
			@s8 = @settings[:d3_s8]
			@a1 = @settings[:d3_a1]
			@a2 = @settings[:d3_a2]
			@beta = @settings[:d3_op_beta]
		when :cso
			@a1 = @settings[:d3_a1]
			@s8 = 0.0
		when :none
			@s8 = @settings[:d3_s8]
		end
	end


	#=======================================================================
	# Private methods: Calculation
	#=======================================================================

	def d3_dispersion_e_g(grad = false)
	#: Dispersion energy calculation, complete -D3 form with continuous coordination
	
		# re-read parameters from input
		# (needed for parameterization where the input changes)
		d3_parameters_from_input 
		
		# Set up empty results
		results = Results.new
		results.gradient = Gradient.zero(@geometry.size) if grad

		# Hybridization
		hybridization = Hybridization.new(@geometry)
		case @settings[:d3_hybridization]
		when :zero
			# Do nothing - hybridization is set to 0, atomic C6s are used
		when :custom
			# Read from the custom values keyword
			@geometry.each_index{|i|
				hybridization[i] = @settings[:d3_hybridization_custom][i].to_f
			}
		when :fixed
			hybridization.evaluate_from_connectivity
		when :grimme
			if grad
				Cuby::recommendation("Gradient of the D3 correction with continuous hybridization not implemented.\nPlease set keyword 'd3_hybridization' to 'fixed'.")
			end
			# Using the radii from the data file
			hybridization.evaluate_grimme(@covalent_radii, false, grad)
			hybridization_grad = hybridization.gradient / ANGSTROM2BOHR if grad
			#hybridization_grad = false
		when :grimme_mod
			if grad
				Cuby::recommendation("Gradient of the D3 correction with continuous hybridization not implemented.\nPlease set keyword 'd3_hybridization' to 'fixed'.")
			end
			# Using the radii from the data file
			hybridization.evaluate_grimme_mod(@covalent_radii, false, grad)
			hybridization_grad = hybridization.gradient / ANGSTROM2BOHR if grad
			#hybridization_grad = false
		end

		# Cutoff
		if @settings[:d3_cutoff] != 0.0
			cutoff = true
			cutoff_r = @settings[:d3_cutoff]
			cutoff_roff = @settings[:d3_cutoff] + @settings[:d3_cutoff_buffer]
			cutoff_f = Func::PolySwitching.new(cutoff_r, cutoff_roff, :sw_1_to_0)
		else
			cutoff = false
		end

		# Distance matrix
		matrix_r = Matrix.zero(@geometry.size)
		@geometry.each_index {|i|
			(i).times { |j|
				r = @geometry.at(i).distance(@geometry.at(j)) * ANGSTROM2BOHR
				matrix_r[i,j] = r
				matrix_r[j,i] = r
			}
		}

		# Store distances and c6 coefficients
		if @settings[:d3_3body]
			save_matrices = true
			matrix_r0 = Matrix.zero(@geometry.size)
			matrix_c6 = Matrix.zero(@geometry.size)
		else
			save_matrices = false
		end

		# What to calculate?
		do_6 = @s6 != 0.0
		do_8 = @s8 != 0.0

		# What to print?
		print_pairwise = @settings[:d3_print].include?(:pairwise_energies)
		if print_pairwise
			Cuby::log.puts "Pairwise D3 dispersion energies (kcal/mol) and c6:"
			Cuby::log.puts(sprintf("%8s%8s%5s%5s%12s%12s", "i", "j", "e_i", "e_j", "c6", "Edisp"))
		end

		# Initialize sums
		edisp6 = 0.0
		edisp8 = 0.0

		# C6 scaling
		if @settings.set?(:d3_c6_scaling)
			@c6_scaling = @settings.element_hash(:d3_c6_scaling)
			@c6_scaling.default = 1.0
		end

		# R0 scaling
		if @settings.set?(:d3_r0_scaling)
			@r0_scaling = @settings.element_hash(:d3_r0_scaling)
			@r0_scaling.default = 1.0
		end

		if hybridization_grad
			matrix_c6, matrix_c6_deriv = precalculate_c6_matrix_with_deriv(
				@geometry, 
				matrix_r, 
				hybridization,
				hybridization_grad
			)
		else
			matrix_c6 = precalculate_c6_matrix(@geometry, matrix_r, hybridization)
		end
		# Iterate over pairs
		@geometry.each_index { |i|
			atom_a = @geometry[i]
			(i).times { |j|
				atom_b = @geometry[j]
				next if (atom_a.properties[:ghost] || atom_b.properties[:ghost] || 
					 atom_a.properties[:dummy] || atom_b.properties[:dummy])

				# Pairwise data
				r = matrix_r[i,j]
				r0 = get_r0(atom_a.element, atom_b.element) * ANGSTROM2BOHR
				c6 = matrix_c6[i,j]

				# C6 scaling
				if @c6_scaling
					cs1 = @c6_scaling[atom_a.element]
					cs2 = @c6_scaling[atom_b.element]
					# Combination rule for c6 coefficients from D2 is used
					if cs1 == 0.0 || cs2 == 0.0
						# Avoid dividing by zero
						c6 = 0.0
					else	       
						c6 *= 2.0 * cs1 * cs2 / (cs1 + cs2)
					end
				end

				# R0 scaling
				if @r0_scaling
					r0_scale = (@r0_scaling[atom_a.element] + @r0_scaling[atom_b.element]) / 2.0
					r0 *= r0_scale
				else
					r0_scale = 1.0
				end
	

				# Save matrices for 3-body part
				if save_matrices
					matrix_r0[i,j] = r0
				end

				# Default: zero
				e6 = 0.0
				e8 = 0.0
		
				# Cutoff		
				next if cutoff && r > cutoff_roff

				c8 = c8_from_c6(atom_a.element, atom_b.element, c6) 

				# Double damping - modify r
				if @settings[:development][:d3_double_damping4]
					rcut = @settings[:development][:d3_double_damping_r] * r0
					w =  @settings[:development][:d3_double_damping_w] * r0
					oldr = r
					r = Func::HockeyStick.new(rcut, w).calc(r)
				end

				# Calculate terms without damping
				if do_6
					e6 = -@s6 * c6 / r**6
					e6_deriv = -@s6 * c6 * -6.0/r**7  if grad
				end

				if do_8
					e8 = -@s8 * c8 / r**8
					e8_deriv = -@s8 * c8 * -8.0/r**9 if grad
				end

				# Calculate damping function
				case @settings[:d3_damping]
				when :bj
					r0ab = (c8 / c6)**0.5 * r0_scale
					if do_6
						damping6 = DampingFunction.bj(r, r0ab, @a1, @a2, 6.0)
						damping6_deriv = DampingFunction.bj_deriv(r, r0ab, @a1, @a2, 6.0) if grad
					end
					if do_8
						damping8 = DampingFunction.bj(r, r0ab, @a1, @a2, 8.0)
						damping8_deriv = DampingFunction.bj_deriv(r, r0ab, @a1, @a2, 8.0) if grad
					end
				when :bjr0
					if do_6
						damping6 = DampingFunction.bj(r, r0, @a1, @a2, 6.0)
						damping6_deriv = DampingFunction.bj_deriv(r, r0, @a1, @a2, 6.0) if grad
					end
					if do_8
						damping8 = DampingFunction.bj(r, r0, @a1, @a2, 8.0)
						damping8_deriv = DampingFunction.bj_deriv(r, r0, @a1, @a2, 8.0) if grad
					end
				when :bjvdw
					r0vdw = (PeriodicTable.vdw_radius(atom_a.element) + PeriodicTable.vdw_radius(atom_b.element)) * ANGSTROM2BOHR
					#r0vdw = (PeriodicTable.vdw_radius(atom_a.element, set=:Truhlar2009) + PeriodicTable.vdw_radius(atom_b.element, set=:Truhlar2009)) * ANGSTROM2BOHR
					if do_6
						damping6 = DampingFunction.bj(r, r0vdw, @a1, @a2, 6.0)
						damping6_deriv = DampingFunction.bj_deriv(r, r0vdw, @a1, @a2, 6.0) if grad
					end
					if do_8
						damping8 = DampingFunction.bj(r, r0vdw, @a1, @a2, 8.0)
						damping8_deriv = DampingFunction.bj_deriv(r, r0vdw, @a1, @a2, 8.0) if grad
					end
				when :zero
					if do_6
						damping6 = DampingFunction.zero(r, r0, @sr6, @alpha)
						damping6_deriv = DampingFunction.zero_deriv(r, r0, @sr6, @alpha) if grad
					end
					if do_8
						damping8 = DampingFunction.zero(r, r0, @sr8, @alpha8)
						damping8_deriv = DampingFunction.zero_deriv(r, r0, @sr8, @alpha8) if grad
					end

				when :erf
					if do_6
						damping6 = DampingFunction.erf(r, r0, @a1, @alpha)
						damping6_deriv = DampingFunction.erf_deriv(r, r0, @a1, @alpha)
					end
					if do_8
						damping8 = DampingFunction.erf(r, r0, @a1, @alpha8)
						damping8_deriv = DampingFunction.erf_deriv(r, r0, @a1, @alpha8)
					end
				when :tt
					# Here, the original parameters were in 1/A**2 and 1/A
					a1 = @a1 / ANGSTROM2BOHR**2
					a2 = @a2 / ANGSTROM2BOHR
					if do_6
						damping6 = DampingFunction.tt(r, r0, a1, a2, 6)
						damping6_deriv = DampingFunction.tt_deriv(r, r0, a1, a2, 6) if grad
					end
					if do_8
						damping8 = DampingFunction.tt(r, r0, a1, a2, 8)
						damping8_deriv = DampingFunction.tt_deriv(r, r0, a1, a2, 8) if grad
					end
				when :op
					r0ab = (c8 / c6)**0.5 * r0_scale
					if do_6
						damping6 = DampingFunction.op(r, r0ab, @a1, @a2, @beta)
						damping6_deriv = DampingFunction.op_deriv(r, r0ab, @a1, @a2, @beta) if grad
					end
					if do_8
						damping8 = DampingFunction.op(r, r0ab, @a1, @a2, @beta+2)
						damping8_deriv = DampingFunction.op_deriv(r, r0ab, @a1, @a2, @beta+2) if grad
					end
				when :cso
					r0ab = (c8 / c6)**0.5 * r0_scale
					if do_6
						damping6 = DampingFunction.cso(r, r0ab, @a1, @s6)
						damping6_deriv = DampingFunction.cso_deriv(r, r0ab, @a1, @s6)
					end
				when :none
					if do_6
						damping6 = 1.0
						damping6_deriv = 0.0
					end
					if do_8
						damping8 = 1.0
						damping8_deriv = 0.0
					end
				end

				# Apply damping
				if grad
					if hybridization_grad
						c6_deriv = matrix_c6_deriv[i,j]
						puts "C6: #{c6}"
						puts "C6_deriv: #{c6_deriv}"

						# Numerical, calculated by changing geometry
						#c6_deriv = 2.4589591977377894
						#puts "C6_deriv: #{c6_deriv} (numerical)"
						e6_deriv = e6_deriv + e6 * c6_deriv
					end

					deriv = 0.0
					deriv += e6 * damping6_deriv + e6_deriv * damping6 if do_6
					deriv += e8 * damping8_deriv + e8_deriv * damping8 if do_8

					# Apply double damping
					if @settings[:development][:d3_double_damping4]
						d2damp = Func::HockeyStick.new(rcut, w).deriv(oldr)
						deriv *= d2damp
					end
				end

				e6 *= damping6 if do_6
				e8 *= damping8 if do_8

				# apply cutoff
				if cutoff &&  r > cutoff_r
					cut = cutoff_f.calc(r)
					if grad
						cut_d = cutoff_f.deriv(r) 
						deriv = deriv * cut + cut_d * (e6+e8)
					end
					e6 *= cut
					e8 *= cut
				end

				# Print pairwise data
				if print_pairwise
					Cuby::log.puts(sprintf("%8d%8d%5s%5s%12.6f%12.6f",
						i+1,
						j+1,
						atom_a.element,
						atom_b.element,
						c6,
						(e6+e8) * HARTREE2KCAL
						))
				end

				# Add current pair to the total energy
				edisp6 += e6 * HARTREE2KCAL
				edisp8 += e8 * HARTREE2KCAL


				if grad
					# Convert units
					deriv = deriv * HARTREE2KCAL / BOHR2ANGSTROM
					# Project gradient size to coordinate
					dcrd= @geometry[i] - @geometry[j]
					dgr = deriv/dcrd.abs * dcrd 

					# and apply to gradient
					results.gradient[i] += dgr 
					results.gradient[j] -= dgr
				end
			}
		}

		results.energy_components[:dispersion_c6] = edisp6
		results.energy_components[:dispersion_c8] = edisp8

		edisp =  edisp6 + edisp8

		# 3-body term, calculated separately
		if @settings[:d3_3body]
			if grad
				Cuby::error("Gradient of the D3 correction with 3-body term not implemented.")
			end
			if cutoff
				Cuby::error("Cutoff in D3 correction with 3-body term not implemented.")
			end
			edisp3b = d3_3body_e(hybridization, matrix_r, matrix_r0, matrix_c6) * @settings[:d3_3body_scaling]
			results.energy_components[:dispersion_3body] = edisp3b
			edisp += edisp3b
		end
		
		# Scaling
		results.energy = edisp * @settings[:d3_scaling]

		# H-H repulsion for SQM
		if @settings[:d3_hh_fix]
			require "interfaces/dispersion3/classes/d3_hh_fix"
			@hh_fix = D3HHFix.new(@settings)
			@hh_fix.calc_add_to_results(geometry, results)
		end

		return results
	end

	def d3_3body_e(hybridization, matrix_r, matrix_r0, matrix_c6)
		# Settings
		@sr3b = @settings[:d3_3body_sr]

		edisp3b = 0.0
		@geometry.each_index { |i| 
			next if (@geometry[i].properties[:ghost] || @geometry[i].properties[:dummy])
			atom_a = @geometry[i]
			(i).times { |j| 
				next if (@geometry[j].properties[:ghost] || @geometry[j].properties[:dummy])
				atom_b = @geometry[j]
				(j).times { |k|
					next if (@geometry[k].properties[:ghost] || @geometry[k].properties[:dummy])
					atom_c = @geometry[k]

					# Distances
					rab = matrix_r[i,j]
					rbc = matrix_r[j,k]
					rca = matrix_r[i,k]

					# Angles
					aa = Coordinate.angle(atom_c, atom_a, atom_b)
					ab = Coordinate.angle(atom_a, atom_b, atom_c)
					ac = Coordinate.angle(atom_b, atom_c, atom_a)

					# C9 coef
					c9 = -1.0 * (matrix_c6[i,j] * matrix_c6[j,k] * matrix_c6[i,k])**0.5

					# r0
					r0ab = matrix_r0[i,j]
					r0bc = matrix_r0[j,k]
					r0ca = matrix_r0[i,k]

					# Damping function
					r_avg = (rab * rbc * rca)**(1.0/3.0)
					r0_avg = (r0ab * r0bc * r0ca)**(1.0/3.0)
				
					if @sr3b == 0
						# no damping
						fd = 1.0
					else	
						fd = DampingFunction.zero(r_avg, r0_avg, @sr3b, 16.0)
					end

					# Dispersion energy
					edisp3b -= fd * c9 * (3.0 * Math::cos(aa) * Math::cos(ab) * Math::cos(ac) + 1.0) / (rab*rbc*rca)**3 
				}
			}
		}
		return edisp3b * HARTREE2KCAL
	end

	#=======================================================================
	# Private methods: misc
	#=======================================================================
	
	def precalculate_c6_matrix(geometry, matrix_r, hybridization)
		c6mat = Matrix.zero(geometry.size)
		geometry.each_index{|i|
			i.times{|j|
				c6mat[i,j] = interpolate_c6(
					matrix_r,
					geometry.at(i).element, 
					geometry.at(j).element, 
					hybridization[i], 
					hybridization[j]
				)
			}
		}
		return c6mat
	end

	def precalculate_c6_matrix_with_deriv(geometry, matrix_r, hybridization, hybridization_gradient)
		c6mat = Matrix.zero(geometry.size)
		c6mat_deriv = Matrix.zero(geometry.size)
		geometry.each_index{|i|
			i.times{|j|
				c6mat[i,j], c6mat_deriv[i,j] = interpolate_c6_with_deriv(
					matrix_r,
					geometry.at(i).element, 
					geometry.at(j).element, 
					hybridization[i], 
					hybridization[j],
					hybridization_gradient[i,j]
				)
			}
		}
		return c6mat, c6mat_deriv
	end
	
	def interpolate_c6(matrix_r, ele_a, ele_b, cn_a, cn_b)
		z = 0.0
		w = 0.0

		# Handle dummy atoms
		return 0.0 if ele_a == :X
		return 0.0 if ele_b == :X

		if @c6_patch && @c6_patch[ele_a] && @c6_patch[ele_a][ele_b]
			Cuby::log.puts_debug("c6 for #{ele_a} - #{ele_b} taken from patch file")
                if @settings[:d3_disable_c6_extrapolation]
					# If the valence number is outside of the range for which c6 are defined,
					# limit the value to the end of the interval
					# Max hyb_a and hyb_b
					hamax = 0.0
					hbmax = 0.0
					@c6_patch[ele_a][ele_b].each{|c6rec|
							hamax = c6rec.hyb_a if c6rec.hyb_a > hamax
							hbmax = c6rec.hyb_b if c6rec.hyb_b > hbmax
					}
					# Limit the values
					cn_a = [cn_a, hamax].min
					cn_b = [cn_b, hbmax].min
                end
			@c6_patch[ele_a][ele_b].each{|c6rec|
				l = Math::exp(-4.0 * ((cn_a - c6rec.hyb_a)**2 + (cn_b - c6rec.hyb_b)**2))
				z += c6rec.c6 * l
				w += l
				# Same element - symmetric points are missing in the table, must be constructed here
				if ele_a == ele_b && c6rec.hyb_a != c6rec.hyb_b
					l = Math::exp(-4.0 * ((cn_a - c6rec.hyb_b)**2 + (cn_b - c6rec.hyb_a)**2))
					z += c6rec.c6 * l
					w += l
				end
			}
		elsif @c6_patch && @c6_patch[ele_b] && @c6_patch[ele_b][ele_a]
			return interpolate_c6(matrix_r, ele_b, ele_a, cn_b, cn_a)
		elsif @c6[ele_a][ele_b]
			if @settings[:d3_disable_c6_extrapolation]
					# If the valence number is outside of the range for which c6 are defined,
					# limit the value to the end of the interval
					# Max hyb_a and hyb_b
					hamax = 0.0
					hbmax = 0.0
					@c6[ele_a][ele_b].each{|c6rec|
							hamax = c6rec.hyb_a if c6rec.hyb_a > hamax
							hbmax = c6rec.hyb_b if c6rec.hyb_b > hbmax
					}
					# Limit the values
					cn_a = [cn_a, hamax].min
					cn_b = [cn_b, hbmax].min
			end
			@c6[ele_a][ele_b].each{|c6rec|
				l = Math::exp(-4.0 * ((cn_a - c6rec.hyb_a)**2 + (cn_b - c6rec.hyb_b)**2))
				z += c6rec.c6 * l
				w += l
				# Same element - symmetric points are missing in the table, must be constructed here
				if ele_a == ele_b && c6rec.hyb_a != c6rec.hyb_b
					l = Math::exp(-4.0 * ((cn_a - c6rec.hyb_b)**2 + (cn_b - c6rec.hyb_a)**2))
					z += c6rec.c6 * l
					w += l
				end
			}
		elsif @c6[ele_b][ele_a]
			return interpolate_c6(matrix_r, ele_b, ele_a, cn_b, cn_a)
		else
			raise "No C6 parameters found for pair #{ele_a} ... #{ele_b}"
		end

		return z/w
	end

	def interpolate_c6_with_deriv(matrix_r, ele_a, ele_b, cn_a, cn_b, dcn_ab)
		z = 0.0
		w = 0.0
		d1 = 0.0
		d2 = 0.0

		if @c6[ele_a][ele_b]
                        if @settings[:d3_disable_c6_extrapolation]
                                # Max hyb_a and hyb_b
                                hamax = 0.0
                                hbmax = 0.0
                                @c6[ele_a][ele_b].each{|c6rec|
                                        hamax = c6rec.hyb_a if c6rec.hyb_a > hamax
                                        hbmax = c6rec.hyb_b if c6rec.hyb_b > hbmax
                                }
                                # Limit the values
                                cn_a = [cn_a, hamax].min
                                cn_b = [cn_b, hbmax].min
                        end
			@c6[ele_a][ele_b].each{|c6rec|
				l = Math::exp(-1.0 * 4.0 * ((cn_a - c6rec.hyb_a)**2 + (cn_b - c6rec.hyb_b)**2))
				z += c6rec.c6 * l
				w += l
				d = 2.0*(cn_b - c6rec.hyb_b)*dcn_ab + 2.0*(cn_a - c6rec.hyb_a)*dcn_ab
				d1 += l * d
				d2 += c6rec.c6 * l * d
				# Same element - symmetric points are missing in the table, must be constructed here
				if ele_a == ele_b && c6rec.hyb_a != c6rec.hyb_b
					l = Math::exp(-1.0 * 4.0 * ((cn_a - c6rec.hyb_b)**2 + (cn_b - c6rec.hyb_a)**2))
					z += c6rec.c6 * l
					w += l
					d = 2.0*(cn_a - c6rec.hyb_b)*dcn_ab + 2.0*(cn_b - c6rec.hyb_a)*dcn_ab
					d1 += l * d
					d2 += c6rec.c6 * l * d
				end
			}
		elsif @c6[ele_b][ele_a]
			c6, deriv = interpolate_c6_with_deriv(matrix_r, ele_b, ele_a, cn_b, cn_a, dcn_ab)
			return [c6, deriv]
		else
			raise "No C6 parameters found for pair #{ele_a} ... #{ele_b}"
		end

		derivative = 3.0*z*d1/w**2 - 3.0*d2/w

		return [z/w, derivative]
	end

	def c8_from_c6(ele_a, ele_b, c6)
		# Handle dummy atoms
		return 0.0 if ele_a == :X
		return 0.0 if ele_b == :X

		s42 = 0.5 # Constant

		q_a = s42 * PeriodicTable.proton_number(ele_a).to_f**0.5 * @r2r4[ele_a]
		q_b = s42 * PeriodicTable.proton_number(ele_b).to_f**0.5 * @r2r4[ele_b]

		c8 = 3.0 * c6 * (q_a * q_b)**0.5
		return c8
	end

	def get_r0(ele_a, ele_b)
		# Handle dummy atoms
		return 1.0 if ele_a == :X
		return 1.0 if ele_b == :X

		if @r0[ele_a][ele_b]
			return @r0[ele_a][ele_b]
		else
			return @r0[ele_b][ele_a]
		end
	end

end
