class FMO
	PDB_AA = [ "ALA", "ARG", "ASN", "ASP", "CYS", "GLN", "GLU", "GLY", "HIS", "HIE", "HID", "HIP", "ILE", "LEU", "LYS", "MET", "PHE", "PRO", "SER", "THR", "TRP", "TYR", "VAL" ]
	PDB_AA_PLUS = [ "ARG", "LYS", "HIP" ]
	PDB_AA_MINUS = [ "ASP", "GLU" ]
	PDB_CUBY_CAPS = ["BCC", "BNC", "BCB", "BCG"]

	def initialize(geometry, settings)
		@geometry = geometry
		@settings = settings


		if @settings[:gamess_fmo_cut_bonds].size == 0
			# Non-covalent fragmentation
			conn = Connectivity.new(@geometry)
			@fragments = conn.molecules
			@cut_bonds = []
		else
			# Cut bonds
			connectivity = Connectivity.new(@geometry)

			# Expand the selections
			# The selections can be generic, make list of real bonds matching all combinations
			expanded_bond_list = []
			settings[:gamess_fmo_cut_bonds].each{|line|
				Cuby::error("The entries of the list of bonds to be cut must be in a form \"selection;selection\"") unless line =~ /[^;]+;[^;]+/
				sel1, sel2 = line.strip.split(/\s*;\s*/)
				list1 = @geometry.atomlist_from_selection(sel1)
				list2 = @geometry.atomlist_from_selection(sel2)
				list1.each{|at1|
					list2.each{|at2|
						if connectivity.is_bond?(at1, at2)
							expanded_bond_list << [at1, at2]
							expanded_bond_list << [at2, at1]
						end
					}
				}
			}
			expanded_bond_list.uniq!
			# Print some info
			fragments = connectivity.molecules
			Cuby::log.puts_v(:normal, "Number of non-covalent fragments: #{fragments.size}")
			Cuby::log.puts_v(:normal, "Number of bonds cut: #{expanded_bond_list.size.to_f/2}")
			#puts expanded_bond_list.map{|x| x.join(" - ")}

			# Make cuts in connectivity, save list of bonds
			@cut_bonds = []
			expanded_bond_list.each{|bond|
				at1, at2, ratio = bond
				# Erase from connectivity
				connectivity[at1,at2] = 0
				sorted = [at1,at2].sort
				@cut_bonds << sorted unless @cut_bonds.include?(sorted)
			}
			#puts @cut_bonds.map{|x| x.join(" - ")}

			# Get final fragments from the modified connectivity
			@fragments = connectivity.molecules # Yields atomlist
			Cuby::log.puts_v(:normal, "Number of resulting fragments: #{@fragments.size}\n\n")
		end

		if @fragments.size < 2
			Cuby::error "FMO calculation requires at least two fragments"
		end

		# PDB analysis, set charge for AA residues
		if @geometry.first.properties[:pdb_res_name]
			total_charge = 0
			Cuby::log.puts_v(:normal, "Fragmentation of PDB file:")
			@fragments.each_index{|i|
				# Collect atom names, original residue names
				# and residue name of atom named CA
				atomnames = []
				orig_res = []
				ca = nil
				@geometry.geometry_from_list(@fragments[i]).each{|atom|
					atomnames << atom.properties[:pdb_atom_name]
					orig_res << atom.properties[:pdb_res_name]+":"+atom.properties[:pdb_res_no].to_s
					if atom.properties[:pdb_atom_name] == "CA"
						ca = atom.properties[:pdb_res_name]
					end
				}
				orig_res.uniq!

				# Initialize fragment information
				s = sprintf("%10d", i+1)
				charge = 0

				# Try to identify fragments
				if ca && PDB_AA.include?(ca) && orig_res.size <= 2
					# If CA atom is present, it may be an amino acid...
					s += " is amino acid #{ca}"

					charge += 1 if PDB_AA_PLUS.include?(ca)
					charge -= 1 if PDB_AA_MINUS.include?(ca)
					if atomnames.include?("OXT")
						charge -= 1 
						s += " (O terminus)"
					end
					if (atomnames & ["H1","H2","H3"]).size == 3
						charge += 1 
						s += " (N terminus)"
					end
					
					
				elsif ca && PDB_CUBY_CAPS.include?(ca) && orig_res.size <= 2
					# Cuby QM/MM caps
					s += " is QM/MM cap #{ca}"
				elsif orig_res.size == 1 && orig_res.first =~ /^WAT/
					# Water
					s += " is water"
				else
					# Everything else
					s += " is built from residues #{orig_res.join(', ')}"
					# Sum charge
					if @settings[:gamess_fmo_auto_charge]
						# Sum charge of contained residues
						orig_res.each{|rname_no|
							name, no = rname_no.split(":")
							if @settings[:gamess_fmo_res_charge][name]
								charge += @settings[:gamess_fmo_res_charge][name]
							end
						}

					end
				end

				# Print charge if it is nonzero
				s += ", charge #{charge}" if charge != 0

				# Set charge if requested
				if @settings[:gamess_fmo_auto_charge]
					if @settings[:gamess_fmo_frag_charge][i+1] && @settings[:gamess_fmo_frag_charge][i+1] != charge
						# User-defined charge different from computed one
						s += ", WARNING - supplied fragment charge was #{@settings[:gamess_fmo_frag_charge][i+1]}, charge #{charge} not set"
					else
						# No user charge, or OK
						@settings[:gamess_fmo_frag_charge][i+1] = charge
					end
				end

				total_charge += @settings[:gamess_fmo_frag_charge][i+1] if @settings[:gamess_fmo_frag_charge][i+1]

				# print fragment info
				Cuby::log.puts_v(:normal, s)
			}
			Cuby::log.puts_v(:normal, "")
			Cuby::log.puts_v(:normal, "Total charge: #{total_charge}")
			Cuby::log.puts_v(:normal, "")
		end

		# Save fragments as PDB
		frag_geo = Geometry.new
		@fragments.each_index{|i|
			fgeo = @geometry.geometry_from_list(@fragments[i]).deep_copy
			fgeo.each{|atom|
				atom.properties[:pdb_res_no] = i+1
				atom.properties[:terafter] = false
				frag_geo << atom
			}
			frag_geo.last.properties[:terafter] = true
			
		}
		frag_geo.write_pdb(:file => "fragments.pdb")
		Cuby::log.puts_v(:normal, "Fragments writtent to file 'fragments.pdb'")
		Cuby::log.puts_v(:normal, "")
	end

	def write_input(f)
		# $FMO
		f.puts " $FMO"
		f.puts "   NFRAG=#{@fragments.size}"
		# Atom assignment to fragments
		frag_i = []
		@geometry.each_index{|i|
			found = nil
			@fragments.each_index{|fi|
				found = fi + 1 if @fragments[fi].include?(i)
			}
			frag_i << found
		}
		f.puts "   INDAT(1)=" + frag_i.join(",\n")
		# Fragment charges
		charges = []
		@fragments.each_index{|fi|
			charges[fi] = 0
			if @settings[:gamess_fmo_frag_charge][fi+1]
				charges[fi] = @settings[:gamess_fmo_frag_charge][fi+1]
			end
		}
		f.puts "   ICHARG(1)=" + charges.join(",\n")
		f.puts " $END"

		# Cut bonds input
		if @cut_bonds.size > 0
			# List of bonds
			f.puts " $FMOBND"
			@cut_bonds.each{|bond|
				i1, i2 = bond
				f.puts "   -#{i1+1} #{i2+1} DFTBC"
			}
			f.puts " $END"

			# FMO terminal orbital definition
			f.puts " $FMOLMO
 DFTBC 4 4
1 0  0.562060  0.000000  0.000000  0.827096
0 1  0.562060  0.779794  0.000000 -0.275699
0 1  0.562060 -0.389897  0.675322 -0.275698
0 1  0.562060 -0.389897 -0.675322 -0.275698
 $END "
		end

		# $FMOPRP
		f.puts " $FMOPRP"
		f.puts "   NPRINT=8"
		f.puts "   MAXIT=#{@settings[:dftb_scc_cycles]}"
		f.puts "   CONV=#{@settings[:dftb_scc_convergence]}"
		f.puts " $END"
	end

	def write_geo(f)
		# For FMO, there is no geometry in the $DATA, only atom types
		f.puts " $DATA"
		f.puts "  cuby"
		f.puts "  C1"
		@geometry.elements_in_system.each{|ele|
			f.puts "  #{ele.to_s} #{PeriodicTable.proton_number(ele)}"
		}
		f.puts " $END"

		# And now the geometry
		f.puts " $FMOXYZ"
		@geometry.each_index{|i|
			atom = @geometry[i]
			f.printf(" %3s %5.1f %15.10f %15.10f %15.10f\n",atom.element, atom.proton_number.to_f, atom.x, atom.y, atom.z)
		}
		f.puts " $END"
	end

	def read_results(output_file) # => Results
		# Initialize results
		results = Results.new
		parser = OutputParser.new(output_file)

		# Errors
		parser.add_error(/SCF IS UNCONVERGED/, "SCF not converged")

		# Energy
		case @settings[:solvent_model]
		when :none
			parser.add_pattern(:fmo_energy, /Euncorr\(2\)= *(.*)/, {:get => :submatch})
		when :cpcm
			parser.add_pattern(:fmo_energy, /Eunc\+so\(2\)= *(.*)/, {:get => :submatch})
		when :smd
			parser.add_pattern(:fmo_energy, /Free uncorr energy in solvent= *(.*)/, {:get => :submatch})
		end

		# Run the parser
		parser.execute 

		# Save energy
		results.energy = parser[:fmo_energy] * HARTREE2KCAL

		return results
	end
end
