################################################################################
#
# Gaussian interface
#
# Author: Jan Rezac
# Date created: 2015-02-02
# License: Cuby4 license
# Description: Interface for external calculations in Gaussian
# Status: Works
#
################################################################################

#===============================================================================
# Interface to the Gaussian package ver. 03 and above
# Commercial software
# http://gaussian.com
#===============================================================================

require "classes/calculation/dft_setup.rb"
require "classes/tools/output_parser.rb"

module InterfaceGaussian
	#=======================================================================
	# Interface header
	#=======================================================================
	DEVELOPMENT_FLAG = :ok
	DEVELOPMENT_STATUS = "Only simple implementation for now"
	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy, :gradient, :point_charges, :ghost_atoms]
	MODIFIER = false
	DIRECTORY = "Gauss"
	# Methods provided by the interface:
	METHODS = {
		:hf		=> [:energy, :point_charges],
		:dft		=> [:energy, :point_charges],
		:am1		=> [:energy],
		:pm3		=> [:energy],
		:mndo		=> [:energy],
		:pm6		=> [:energy, :gradient]
	}
	#SOLVENTS = [:cosmo]
	#ATOMIC_CHARGES = [:mulliken]
	#=======================================================================
	
	INPUT_FILE = "gaussian.com"
	OUTPUT_FILE = "gaussian.log"

	def prepare_interface
		# Things that have to be done even when old input is reused:

		# Input checks - DFT with custom functional
		if @settings[:method] == :dft && @settings[:functional].downcase == "custom" 
			unless @settings.set?(:functional_custom_type)
				Cuby::error("When using custom functional, the keyword 'functional_custom_type' must be set as well")
			end
		end

		# Prepare calculation directory
		if calc_dir_mkdir("gaussian.header", OUTPUT_FILE) == :old
			calc_using_input # Old input should be reused
			return
		end

		# GDMA setup
		if @settings[:gaussian_run_gdma]
			# Keep checkpoint file
			@settings[:gaussian_checkpoint] = :text
			# Write GDMA input
			gdma_write_input
		end

		# Write the keyword part of the input
		gaussian_write_keywords # This writes files mopac.header and mopac.footer

		# Create complete input by combining header, footer and the current geometry
		gaussian_write_input_geo

		# Save info on the system upon writing the input
		calc_writing_input
	end

	def calculate_interface
		gaussian_run unless @settings[:existing_calc_dir] == :read_results && FileTest.exist?(calc_dir+"/"+OUTPUT_FILE)
		return gaussian_read
	end

	def cleanup_interface
		calc_dir_delete if @settings[:job_cleanup]
	end

	#=======================================================================
	# Private methods
	#=======================================================================

	def gaussian_write_keywords
		#: Write the header of the input where the calculation is set up
		#: geometry is added later, so that the header can be modified
		#: by the user after the job is prepared

		# HEADER FILE
		filename = in_calc_dir("gaussian.header")
		f = File.open(filename,"w+")

		# Setup
		f.puts "%mem=#{@settings[:mem]}MB"
		f.puts "%NProcShared=#{@settings[:parallel]}"
		f.puts "%Chk=checkpoint.chk" if @settings[:gaussian_checkpoint] != :none


		# Method
		options = ""
		case @settings[:method]
		when :hf
			method = "HF"
			if [:scf, :both].include?(@settings[:density_fitting])
				Cuby::warning "Gaussian does not support density fitting in HF"
			end
		when :dft
			dft_setup = DftSetup.new(interface_dir + "/dft_functionals.yaml", interface_dir + "/dft_grids.yaml")
			method = dft_setup.functional(@settings)
			options += " Integral(Grid=#{dft_setup.grid(@settings)})"
			options += " DensityFit" if [:scf, :both].include?(@settings[:density_fitting])
		when :am1
			method = "AM1"
		when :pm3
			method = "PM3"
		when :pm6
			method = "PM6"
		when :mndo
			method = "MNDO"
		end

		# Basis setup - not applied to some methods
		unless [:am1, :pm3, :pm6, :mndo].include?(@settings[:method])
			# Bais set harmonics form
			case @settings[:basisset_harmonics]
			when :spherical
				options += " 5D 7F"
			when :cartesian
				options += " 6D 10F"
			end

			# Basis set
			basisset = @settings[:basisset]
			basisset = basisset.gsub(/^def2-/, "def2") # Remove dash from def2-XXX basis
			if @settings.set?(:basisset_elements)
				basisset = "gen"
			end
			method += "/" + basisset
		end

		# Properties
		if @what.include? :gradient
			method += " force"
		end

		# Point charges
		if @point_charges
			options += " Charge"
		end
		
		f.puts "#P #{method}#{options} #{@settings[:gaussian_keywords]}"
		f.puts
		f.puts "Gaussian calculation for Cuby"
		f.puts
		f.puts "#{@settings[:charge]} #{@settings[:multiplicity]}"

		f.close

		# FOOTER FILE
		filename = in_calc_dir("gaussian.footer")
		f = File.open(filename,"w+")

		if @settings.set?(:basisset_elements)
			f.puts
			basisset_elements = @settings.elements_hash(:basisset_elements)
			# Custom basis block
			basis_by_ele = {}
			@geometry.elements_in_system.each{|element|
				if basisset_elements.has_key?(element)
					# Custom basis
					basis = basisset_elements[element]
					basis_by_ele[basis] = [] unless basis_by_ele[basis]
					basis_by_ele[basis] << element
				else
					# Default basis
					basis_by_ele[@settings[:basisset]] = [] unless basis_by_ele[@settings[:basisset]]
					basis_by_ele[@settings[:basisset]] << element
				end
			}
			basis_by_ele.each_pair{|bas, elements|
				f.puts (elements + ['0']).join(" ")
				f.puts bas
				f.puts "****"
			}
		end

		if @settings[:gaussian_extra_block] != ""
			f.puts
			f.puts @settings[:gaussian_extra_block]
			f.puts
		end
		f.puts
		f.puts
		f.close
	end

	def gaussian_write_input_geo
		#: Write final input for mopac, using the previously prepared header and footer
		#: and a current geometry

		# Read header and footer
		header = IO.readlines(in_calc_dir("gaussian.header"))
		footer = IO.readlines(in_calc_dir("gaussian.footer"))

		# Complete input:
		filename = in_calc_dir(INPUT_FILE)
		f = File.open(filename,"w+")

		# Print header from the header file
		f.puts header

		# Write geometry
		@geometry.each_index{|i|
			atom = @geometry[i]

			ele = atom.element.to_s
			if atom.properties[:ghost]
				ele += "-Bq"
			end
			
			f.printf("%-5s  %15.10f %15.10f %15.10f\n",ele,atom.x,atom.y,atom.z)
		}

		# Point charges
		if @point_charges
			f.puts
			@point_charges.each{|pch|
				f.printf("%15.10f %15.10f %15.10f % 15.10f\n", pch.x, pch.y, pch.z, pch.charge)
			}
		end

		# Print footer
		f.puts footer

		f.close
	end

	def gaussian_run
		#: Set up environment for gaussian and run it

		# Build input with the current geometry
		gaussian_write_input_geo

		# Run gaussian
		command =  "export GAUSS_EXEDIR=\"#{@settings[:gaussian_bin_dir]}\";"
		command << "export LD_LIBRARY_PATH=\"#{@settings[:gaussian_bin_dir]}\":$LD_LIBRARY_PATH;"
		command << "export GAUSS_SCRDIR=\"#{@settings[:gaussian_scratch_dir]}\";"
		command << "cd #{calc_dir};"
		command << "mkdir -p $GAUSS_SCRDIR;"
		command << "$GAUSS_EXEDIR/#{@settings[:gaussian_version]} #{INPUT_FILE} 2> gaussian.err > gaussian.stdout;"
		command << "rm -rf $GAUSS_SCRDIR;"
		unless system(command)
			Cuby::error "Gaussian returned nonzero exit code (calculation #{@name})"
		end

		# Error check
		Cuby::error "Gaussian output file not found" unless FileTest.exist?(in_calc_dir(OUTPUT_FILE))

		# Error check
		f = File.open(in_calc_dir(OUTPUT_FILE)) 	# Get last line of the file:
		f.seek(-100, IO::SEEK_END)			# Set position 100 bytes before EOF
		f.gets						
		lastline = f.gets
		f.close
		Cuby::error "Gaussian run did not terminate normally" unless lastline =~ /Normal termination of Gaussian/

		# Convert checkpoint to text
		if @settings[:gaussian_checkpoint] == :text
			system("cd #{calc_dir}; #{@settings[:gaussian_bin_dir]}/formchk checkpoint.chk 2>&1 > formchk.out")
		end

		# Run GDMA
		if @settings[:gaussian_run_gdma]
			system("cd #{calc_dir}; #{@settings[:gdma_exe]} < gdma.in 2>&1 > gdma.out")
		end
	end

	def gaussian_read # => Results
		# Initialize results
		results = Results.new

		# Try reading the results block first
		results_block = gaussian_get_results_block

		# Debugging: print the results block
		if Cuby::log.logs[0].verbosity == :debug
			Cuby::log.puts_debug "Gaussian results archive:"
			results_block.each_index{|i|
				Cuby::log.puts_debug "block #{i}:" 
				Cuby::log.puts_debug results_block[i]
			}
			Cuby::log.puts_debug ""
		end

		# Read the energy
		if @settings[:method] == :dft && @settings[:functional].downcase == "custom" && @settings[:functional_custom_type] == :double_hybrid
			# Double hybrid functionals - read from the "MP2" record
			results.energy = find_in_block(results_block[4], "MP2").to_f * HARTREE2KCAL
		elsif [:hf, :dft, :am1, :pm3, :pm6, :mndo].include?(@settings[:method])
			# Energy from the results block
			results.energy = find_in_block(results_block[4], "HF").to_f * HARTREE2KCAL
		else
			raise "Reading energy not implemented"
		end

		# Gradient
		if @what.include?(:gradient)
			results.gradient = gaussian_gradient
		end

		return results
	end

    def gaussian_gradient
        #: Get gradients from the ouput file (Gaussian does not print gradient to the archive block).
        gradient = Gradient.new
        f = File.open(in_calc_dir("gaussian.log"), "r")
        read = false
        found = false
        f.each_line {|line|
            read = true if line =~ /Forces \(Hartrees\/Bohr\)/
            if read && m = / *([0-9]+) +([0-9]+) +(\S+) +(\S+) +(\S+)/.match(line)
                found = true
                gradient << Coordinate[m[3].to_f, m[4].to_f, m[5].to_f] * HARTREE2KCAL / BOHR2ANGSTROM * -1.0
            else
                break if found
            end
        }
        f.close

        Cuby::error "Gradients record not found in gaussian output" unless read
        Cuby::error "Gradients not found in gaussian output" unless found
        Cuby::error "Gradients read does not match number of atoms in geometry" if @geometry.size != gradient.size
        
        return gradient
    end

	def gaussian_get_results_block
		#: Find the machine-readable "archive" block and return it as an array
		f = File.open(in_calc_dir(OUTPUT_FILE), "r")
		results_s = ""
		save_this = false
		f.each_line {|line|
			if line =~ /^ *1\\1/
				save_this = true
			end
			results_s << line.chomp.strip if save_this
			if line =~ /\\\\@$/
				save_this = false
			end
		}
		f.close
		a = results_s.split("\\\\").collect{|b| b.split("\\")}

		return a
	end

	def find_in_block(block, var)
		block.each{|item|
			if item =~ /^#{var}=/
				return item.gsub(/^#{var}=/,'')
			end
		}
		return nil
	end

	#========================================================================
	# GDMA
	#========================================================================
	
	def gdma_write_input
		f = File.open(in_calc_dir("gdma.in"), "w+")
		f.puts "File checkpoint.fchk"
		f.puts "Angstrom"
		f.puts "Multipoles"
		f.puts "  Switch 4"
		f.puts "  Limit #{@settings[:gdma_rank]}"
		#f.puts "  Limit 1 H"
		f.puts "  Punch gdma.punch"
		f.puts "Start"
		f.puts "Finish"
		f.close
	end
	

end
