################################################################################
#
# MOPAC interface
#
# Author: Jan Rezac
# Date created: 2013-04-09
# License: Cuby4 license
# Description: Interface for external calculations in MOPAC
# Status: Works
#
################################################################################

#===============================================================================
# Interface to the MOPAC 2009-2016
# Free for academic use, available after registration
# http://www.openmopac.net
#===============================================================================

# EXTOPT TBD
# * optimization setup

require "classes/tools/output_parser.rb"

module InterfaceMopac
	#=======================================================================
	# Interface header
	#=======================================================================
	DEVELOPMENT_FLAG = :ok
	DEVELOPMENT_STATUS = "All cuby3 functionality implemented, examples for all features"
	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy, :gradient, :atomic_charges, :native_opt, :electric_field, :point_charges]
	MODIFIER = false
	DIRECTORY = "MOPAC"
	# Methods provided by the interface:
	METHODS = {
		:"mndo"		=> [:energy, :gradient, :atomic_charges, :native_opt, :point_charges],
		:"mndo-d"	=> [:energy, :gradient, :atomic_charges, :native_opt, :point_charges],
		:"am1" 		=> [:energy, :gradient, :atomic_charges, :native_opt, :point_charges],
		:"rm1" 		=> [:energy, :gradient, :atomic_charges, :native_opt, :point_charges],
		:"pm3" 		=> [:energy, :gradient, :atomic_charges, :native_opt, :point_charges],
		:"pm6" 		=> [:energy, :gradient, :atomic_charges, :native_opt, :point_charges],
		:"pm7" 		=> [:energy, :gradient, :atomic_charges, :native_opt, :point_charges]
	}
	SOLVENTS = [:cosmo, :cosmo2]
	ATOMIC_CHARGES = [:mulliken]
	#=======================================================================


	# Interface constants (lookup tables for construction of the input)
	MOPAC_MULTIPLICITY = {
		1 => "SINGLET",
		2 => "DOUBLET",
		3 => "TRIPLET",
		4 => "QUARTET",
		5 => "QUINTET",
		6 => "SEXTET",
		7 => "SEPTET",
		8 => "OCTET",
		9 => "NONET"
	}
	MOPAC_METHODS = {
		:"mndo" 	=> "MNDO",
		:"mndo-d" 	=> "MNDOD",
		:"am1" 		=> "AM1",
		:"rm1" 		=> "RM1",
		:"pm3" 		=> "PM3",
		:"pm6" 		=> "PM6",
		:"pm7" 		=> "PM7"
	}
	MOPAC_CORRECTIONS = {
		:none =>	"",
		:"d2" =>	"-D2",
		:"h2" =>	"-H2",
		:"dh2" =>	"-DH2",
		:"dh+" =>	"-DH+",
		:"d3" =>	"-D3",
		:"h4" =>	"-H4",
		:"d3h4" =>	"-D3H4",
		:"d3(h4)" =>	"-D3(H4)",
		:"d3h4x" =>	"-D3H4X"
	}


	def prepare_interface
		# Compatibility checks
		if @settings[:mopac_version] == :mopac7
			if @geometry.size <= 3
				# For small molecules, mopac7 can use only internal coordinates
				Cuby::error("Mopac7 interface supports only molecules with 4 or more atoms (calculation #{@name})")
			end
		end

		if @settings[:solvent_model] == :cosmo2
			unless [:pm6, :pm7].include?(@settings[:method])
				Cuby::error("COSMO2 parameters are available only for PM6 and PM7")
			end
			if @what.include?(:gradient)
				Cuby::error("COSMO2 gradient is not implemented")
			end
		end

		# Things that have to be done even when old input is reused:

		# Setpi - read from file
		if @settings.set?(:mopac_setpi_from_file)
			geo_sp = Geometry.from_file(@settings[:mopac_setpi_from_file])
			unless geo_sp.info[:filetype] == :sdf
				Cuby::error("Keyword 'mopac_setpi_from_file' supports only .SDF files")
			end
		
			# get connectivity and atom types
			conn = Connectivity.new(geo_sp)
			atypes = conn.atom_types_simple(geo_sp)


			bonds = geo_sp.info[:sdf_bonds]
			count = 0
			
			bonds.each_pair{|idx, order|
				if order == 2
					i1, i2 = idx
					# Skip double bonds in -NO2 group
					next if atypes[i1-1] == "N:CO2"
					next if atypes[i2-1] == "N:CO2"
					# Add to the list
					@settings[:mopac_setpi] << "#{i1};#{i2}"
					count += 1
				end
			}

			unless @settings[:mopac_setpi_silent]
				Cuby::log.puts "Mopac_setpi_from_file read #{count} double bonds"
			end
		end

		# puts @settings[:mopac_setpi]

		# Setpi - manual definition of double bonds for mozyme
		@setpi_list = [] # Array of the lines added after geometry
		if @settings[:mopac_setpi].size > 0
			conn = Connectivity.new(@geometry)
			@settings[:mopac_setpi].each{|rec|
				unless rec.scan(/;/).count == 1
					Cuby::error "SETPI requires two selections separated by a semicolon"
				end
				# Build two atom lists
				s1, s2 = rec.split(";")
				list1 = @geometry.atomlist_from_selection(s1)
				list2 = @geometry.atomlist_from_selection(s2)
				# Find any bonds between these lists
				list1.each{|i|
					list2.each{|j|
						if conn.is_bond?(i,j)
							@setpi_list << [i+1, j+1].sort
						end

					}
				}
			}

			# Remove multiple entries
			@setpi_list.uniq!

			# Convert setpi list to strings
			@setpi_list.map!{|ab| a,b = ab; "#{a} #{b}"}

			# Truncate to max. no used in MOPAC
			setpi_max = 49
			if @setpi_list.size > setpi_max
				Cuby::warning("SETPI can work with max. #{setpi_max} records. Current list of #{@setpi_list.size} bonds will be truncated.")
				@setpi_list.slice!(0,setpi_max)
			end


			unless @settings[:mopac_setpi_silent]
				if @setpi_list.size == 0
					Cuby::log.puts "Mopac_setpi keyword was evaluated but no matching bonds were found"
				else
					Cuby::log.puts "No. of bonds matching the mopac_setpi patterns: #{@setpi_list.size}"
				end
			end
		end

		# Manual assignment of atomic charges
		@charge_selections = {}
		@settings[:mopac_setcharge].each_pair{|sel, value|
			list = @geometry.atomlist_from_selection(sel.to_s)
			list.each{|i| @charge_selections[i] = value}
		}

		# COSMO2 solvation - automatic setup
		if @settings[:solvent_model] == :cosmo2
			# Load data file
			f = File.open(interface_dir + "/cosmo2_parameters.yaml")
			cosmo2_data = YAML.unsafe_load(f)[@settings[:method]]
			f.close
			# Modify input
			# Radii scaling
			if ! @settings.set?(:mopac_cosmo2_scale_missing_radii) && cosmo2_data[:mopac_cosmo2_scale_missing_radii] # not used now
				@settings[:mopac_cosmo2_scale_missing_radii] = cosmo2_data[:mopac_cosmo2_scale_missing_radii]
			end
			# Radii
			unless @settings.set?(:mopac_cosmo_radii)
				@settings[:mopac_cosmo_radii] = {}
				@geometry.elements_in_system.each{|ele|
					# Do not override user input
					next if @settings[:mopac_cosmo_radii][ele]
					if cosmo2_data[:mopac_cosmo_radii][ele]
						# If there is parameter, use it
						@settings[:mopac_cosmo_radii][ele] = cosmo2_data[:mopac_cosmo_radii][ele]
					elsif @settings[:mopac_cosmo2_scale_missing_radii] != 0.0
						# Otherwise, apply the scaling to a vdw radius
						@settings[:mopac_cosmo_radii][ele] = PeriodicTable.vdw_radius(ele) * @settings[:mopac_cosmo2_scale_missing_radii]
					end
				}
			end
			# SASA term
			unless @settings.set?(:mopac_cosmo2_surf_tension)
				@settings[:mopac_cosmo2_surf_tension] = cosmo2_data[:mopac_cosmo2_surf_tension]
			end

		end

		# Prepare calculation directory
		if calc_dir_mkdir("mopac.header", "mopac.in.out") == :old
			calc_using_input # Old input should be reused
			return
		end

		# Write the keyword part of the input
		mopac_write_keywords # This writes file mopac.header

		# Create complete input by combining mopac.header and the current geometry
		mopac_write_input_geo

		# Save info on the system upon writing the input
		calc_writing_input
	end

	def calculate_interface
		mopac_run unless @settings[:existing_calc_dir] == :read_results && FileTest.exist?(calc_dir+"/mopac.in.out")
		return mopac_read
	end

	def cleanup_interface
		calc_dir_delete if @settings[:job_cleanup]
	end

	#=======================================================================
	# Private methods
	#=======================================================================

	def mopac_write_keywords
		#: Write the header of the input where the calculation is set up
		#: geometry is added later, so that the header can be modified
		#: by the user after the job is prepared

		filename = in_calc_dir("mopac.header")
		f = File.open(filename,"w+")

		# Method
		if @settings[:development][:mopac_custom_method]
			# Development - allow aribtrary method name
			f.print @settings[:development][:mopac_custom_method]
		else
			# Standard methods
			f.print MOPAC_METHODS[@settings[:method]]
		end

		# Corrections
		f.print MOPAC_CORRECTIONS[@settings[:mopac_corrections]]

		# Single point / optimization	
		if @what.include?(:native_opt)
			# Optimization - save aux file
			save_aux = true
			f.print " CYCLES=#{@settings[:maxcycles]}" 
		else
			# Single point
			f.print " 1SCF" 
		end

		# SCF
		f.print " DENOUT"
		f.print " CHARGE=#{@settings[:charge]}"

		# Spin
		unless MOPAC_MULTIPLICITY[@settings[:multiplicity]]
			Cuby::error("MOPAC can work only with spin multiplicity in the range 1-9 (calculation #{@name})")
		end
		f.printf " #{MOPAC_MULTIPLICITY[@settings[:multiplicity]]}"
		if @settings[:spin_restricted] == :rhf || (@settings[:spin_restricted] == :auto_uhf && @settings[:multiplicity] == 1)
			f.printf " RHF" if @settings[:mopac_version] == :mopac2012
		else
			f.printf " UHF"
		end

		# Mozyme
		f.print " MOZYME" if @settings[:mopac_mozyme]

		# Setpi - manual definition of double bonds for mozyme
		if @setpi_list.size > 0 
			# Write to input
			f.print " SETPI"
		end

		# SCF convergence
		# Precise - accurate gradient
		f.print " PRECISE" if @settings[:mopac_precise]
		f.print " RELSCF=#{@settings[:mopac_relscf]}" if @settings.set?(:mopac_relscf)
		f.print " DAMP=#{@settings[:mopac_damp_scf]}" if @settings.set?(:mopac_damp_scf)

		# COSMO
		if @settings[:solvent_model] != :none
			f.print " EPS=#{@settings[:solvent_epsilon]}"

			if @settings.set?(:mopac_cosmo_radii)
				rads = []
				@settings[:mopac_cosmo_radii].each_pair{|ele, rad|
					rads << "#{ele.to_s}=#{rad}"
				}
				f.print " VDW(#{rads.join(';')})"
			end

			f.print " NSPA=#{@settings[:mopac_cosmo_nspa]}" if @settings.set?(:mopac_cosmo_nspa)
		end

		# Calculate gradient or charges if needed
		#f.print " GRADIENTS DCART" if @what.include?(:gradient)
		if @what.include?(:gradient)
			f.print " GRADIENTS"
			save_aux = true
		end
		f.print " MULLIK" if @what.include?(:atomic_charges)

		# Homogenous electric field
		if @settings.set?(:electric_field)
			# A 3D vector of numbers
			field = @settings[:electric_field].map{|x| x.to_f}
			unless field.size == 3
				Cuby::error("Electric field must be specified as 3D vector in cartesian coordinates")
			end
			# MOPAC's units are Volts/Angstrom
			field.map!{|x| x / 51.4220652}
			# Round numbers
			field.map!{|x| sprintf("%.4f", x)}
			# Print it
			f.print " FIELD=(#{field.join(',')})"
		end

		# Parallelization
		if @settings[:mopac_parallel] && @settings[:mopac_version] == :mopac2012
			f.print " THREADS=#{@settings[:parallel]}"
		end

		# Correction for peptide bonds (addition for RM1, PM6, and PM7)
		if @settings[:mopac_peptide_bond_fix] && [:pm6, :pm7, :rm1].include?(@settings[:method])
			f.print " MMOK"
		else
			f.print " NOMM"
		end

		# Print the corrections included in PM7
		if @settings[:method] == :pm7
			f.print " DISP"
		end

		# CPE coorection (development)
		if @settings.set?(:mopac_cpe)
			f.print " mullik ef external=cpe.dat"
		end

		# Save auxiliary file if needed
		if save_aux
			f.print " AUX(MOS=-99999,XW,XS,XP,PRECISION=6)"
		end

		# Custom parameters from external file
		if @settings.set?(:mopac_external_params)
			f.print " EXTERNAL=external.rp"
			mopac_write_external("external.rp")
		end

		# Options available only in custom version of MOPAC
		if @settings.set?(:mopac_cosmo_limit)
			if @settings[:mopac_custom_exe]
				f.print " COSLIMIT=#{@settings[:mopac_cosmo_limit]}"
			else
				Cuby::warning("Keyword 'mopac_cosmo_limit' not active, it requires custom version of MOPAC (calculation #{@name})")
			end
		end

		# Extra keywords
		f.print " " + @settings[:mopac_keywords] if @settings[:mopac_keywords] != ""

		f.close
	end

	def mopac_write_input_geo
		#: Write final input for mopac, using the previously prepared header
		#: and a current geometry

		# Read mopac.header
		header = IO.readlines(in_calc_dir("mopac.header"))

		# Complete input:
		filename = in_calc_dir("mopac.in")
		f = File.open(filename,"w+")

		# Print header from the header file, adding option to restart from previous density
		restart = ""
		if FileTest.exist?(in_calc_dir("mopac.in.den")) && @settings[:start_from_previous]
			restart =  " OLDENS" 
		end
		header[0] = header[0].strip + restart


		# Point charges
		if @point_charges
			header[0] = header[0].strip + " QMMM"
			mopac_write_pch
		end

		# Write the header line
		f.puts header[0].strip

		# Separator
		f.print "\n\n"
		case @settings[:mopac_version]
		when :mopac2012
			@geometry.each_index{|i|
				atom = @geometry[i]
				ch = ""
				if @charge_selections[i]
					ch = "(#{@charge_selections[i]})"
				end
				f.printf("  %2s%-4s    %15.10f %15.10f %15.10f\n",atom.element,ch,atom.x,atom.y,atom.z)
			}
		when :mopac7
			@geometry.each_index{|i|
				atom = @geometry[i]
				f.printf("  %2s   %15.10f 0 %15.10f 0 %15.10f 0\n",atom.element,atom.x,atom.y,atom.z)
			}
		end
		f.puts

		# Setpi - manual definition of double bonds for mozyme
		# Write the list to the input
		if @setpi_list.size > 0
			@setpi_list.each{|line| f.puts line}
			f.puts
		end

		f.close

		# Write also external params
		if @settings.set?(:mopac_external_params)
			mopac_write_external("external.rp")
		end
	end

	def mopac_write_external(fn)
		f = File.open(in_calc_dir(fn),"w+")
		@settings[:mopac_external_params].each_pair{|para, value|
			p, ele = para.split(",")
			ele = "" unless ele
			f.printf("    %-10s%-4s%20.10f\n", p, ele, value)
		}
		f.close
	end

	def mopac_run
		#: Set up environment for MOPAC and run it

		# CPE coorection (development)
		if @settings.set?(:mopac_cpe)
			mopac_write_cpe
		end

		# Write external parameters before every run
		if @settings.set?(:mopac_external_params)
			mopac_write_external("external.rp")
		end

		# Build input with the current geometry
		mopac_write_input_geo

		# Prepare the command for running mopac
		command = "cd #{calc_dir}; "
		case @settings[:mopac_version]
		when :mopac2012
			command << "export OMP_NUM_THREADS=#{@settings[:parallel]};"
			command << "export MKL_NUM_THREADS=#{@settings[:parallel]};"
			command << "export MOPAC_LICENSE=#{File.dirname(@settings[:mopac_exe])};"
			command << "export LD_LIBRARY_PATH=#{File.expand_path(File.dirname(@settings[:mopac_exe]))}:$LD_LIBRARY_PATH;"
			command << "#{@settings[:mopac_exe]} mopac.in > mopac.stdout 2> mopac.err;"
		when :mopac7
			command << "export FOR005=mopac.in; "
			command << "export FOR006=mopac.out2; "
			command << "export FOR009=mopac.res; "
			command << "export FOR010=mopac.den; "
			command << "export FOR011=mopac.log; "
			command << "export FOR012=mopac.arc; "
			command << "#{@settings[:mopac_exe]} 2> mopac.err > mopac.in.out"
		end

		# Run mopac
		unless system(command)
			# Catch specific errors
			parser = OutputParser.new(in_calc_dir("mopac.in.out"), @name)
			parser.add_error(/SINGLET SPECIFIED WITH ODD NUMBER OF ELECTRONS/, "Mopac interface: Charge specified on input not possible for closed-shell system")
			parser.add_error(/FAILED TO ACHIEVE SCF/, "Mopac interface: SCF failed")
			parser.add_error(/COSMO failed, COSLIMIT reached/, "COSMO calculation failed, COSLIMIT reached")
			parser.execute

			Cuby::error "Mopac returned nonzero exit code (calculation #{@name})"
		end
	end

	def mopac_read #=> Results
		#: Reads energy, gradient and dipole from MOPAC output
	
		results = Results.new
		parser = OutputParser.new(in_calc_dir("mopac.in.out"), @name)

		# Energy
		parser.add_pattern(:energy, /^ +FINAL HEAT OF FORMATION = +([-]*[0-9]*\.[0-9]+)/, :get => :submatch, :required => false)
		parser.add_pattern(:energy_el, /^ +ELECTRONIC ENERGY += +([-]*[0-9]*\.[0-9]+)/, :get => :submatch, :required => false)

		if @settings[:method] == :pm7
			parser.add_pattern(:energy_d,  /^ +DISPERSION ENERGY += +([-]*[0-9]*\.[0-9]+)/, :get => :submatch, :required => false)
			parser.add_pattern(:energy_hb, /^ +HYDROGEN-BOND ENERGY += +([-]*[0-9]*\.[0-9]+)/, :get => :submatch, :required => false)
		end

		# Dipole
		parser.add_pattern(:dipole,
			/^ SUM\s+([-]*[0-9]*\.[0-9]+)\s+([-]*[0-9]*\.[0-9]+)\s+([-]*[0-9]*\.[0-9]+)\s+([-]*[0-9]*\.[0-9]+)/,
			:get => :match_array, :type => :float, :required => false
		)

		# Errors
		if @settings[:mopac_mozyme]
			parser.add_error(/CHARGE SPECIFIED.* IS INCORRECT./, "Mopac thinks the charge is wrong.\nIt may be an issue in MOZYME, please check the mopac output file\n#{in_calc_dir("mopac.in.out")}\nand correct the charge or fix the Lewis structure guess for MOZYME.")
		else
			parser.add_error(/CHARGE SPECIFIED.* IS INCORRECT./, "Mopac thinks the charge is wrong")
		end
		parser.add_error(/EXCESS NUMBER OF OPTIMIZATION CYCLES/, "Geometry optimization in MOPAC did not converge")
		parser.add_error(/SPECIFIED SPIN COMPONENT NOT SPANNED BY ACTIVE SPACE/, "Mopac open-shell calculation requires a definition of the active space")
		parser.add_error(/Impossible user-supplied pi bond/, "Error in SETPI: Impossible user-supplied pi bond")
		parser.add_error(/SINGLET SPECIFIED WITH ODD NUMBER OF ELECTRONS/, "Mopac interface: Charge specified on input not possible for closed-shell system")
		parser.add_error(/FAILED TO ACHIEVE SCF/, "Mopac interface: SCF failed")
		parser.add_error(/COSMO failed, COSLIMIT reached/, "COSMO calculation failed, COSLIMIT reached")


		# Add charges reader if needed
		if what.include?(:atomic_charges)
			results.atomic_charges = AtomicCharges.new(:mulliken)
			# The default charges printed by MOPAC
			#--------------------------------------------------
			# parser.add_block(
			# 	:charges, 
			# 	/NET ATOMIC CHARGES/, 
			# 	/^\s+[0-9]+\s+[A-Z,a-z]+\s+([-]*[0-9]*\.[0-9]+)\s+/,
			# 	/^ DIPOLE/,
			# 	{:get => :submatch, :type => :float}
			# )
			#-------------------------------------------------- 

			# Mulliken charges
			parser.add_block(
				:charges, 
				/MULLIKEN POPULATIONS AND CHARGES/, 
				/^\s+[0-9]+\s+[A-Z,a-z]+\s+[^ ]+\s+([-]*[0-9]*\.[0-9]+)\s+/,
				/^ TOTAL CPU TIME/,
				{:get => :submatch, :type => :float}
			)
		end

		# CPE coorection (development)
		if @settings.set?(:mopac_cpe)
			parser.add_pattern(:energy_cpe, /^ +Energy += +([-]*[0-9]*\.[0-9]+) +kcal/, :get => :submatch)
		end

		# Optimized geometry
		if @what.include?(:native_opt)
			# Convergence check
			parser.add_pattern(:geo_conv_grad,  /^ +GRADIENT = +([-]*[0-9]*\.[0-9]+) +IS LESS THAN CUTOFF/, :get => :submatch, :required => false)
		end

		# COSMO details
		if @settings[:solvent_model] == :cosmo || @settings[:solvent_model] == :cosmo2
			parser.add_pattern(:cosmo_area, /^ +COSMO AREA += +([-]*[0-9]*\.[0-9]+) +/, :get => :submatch, :required => false)
			parser.add_pattern(:cosmo_volume, /^ +COSMO VOLUME += +([-]*[0-9]*\.[0-9]+) +/, :get => :submatch, :required => false)

			parser.add_pattern(:cosmo_energy, /^ +DIELECTRIC ENERGY += +([-]*[0-9]*\.[0-9]+) +/, :get => :submatch, :required => false)
		end

		# Run parser
		parser.execute{|name, result, count|
			case name
			when :charges
				results.atomic_charges << result
			end
		}

		# Energy
		Cuby::error "Energy not found in MOPAC output (calculation #{@name})" unless parser[:energy]
		results.energy = parser[:energy]

		results.energy_components[:electronic_energy] = parser[:energy_el].to_f * EV2KCAL

		# CPE coorection (development)
		if @settings.set?(:mopac_cpe)
			results.energy += parser[:energy_cpe]
		end

		# PM7 corrections
		if @settings[:method] == :pm7
			results.energy_components[:dispersion] = parser[:energy_d] if parser[:energy_d]
			if parser[:energy_hb]
				results.energy_components[:h_bonds] = parser[:energy_hb] 
			else
				results.energy_components[:h_bonds] = 0.0
			end

		end

		# Dipole
		if parser[:dipole]
			results.multipoles[:dipole] = Multipoles::Dipole.new(
				parser[:dipole][0] * DEBYE2CUBY,
				parser[:dipole][1] * DEBYE2CUBY,
				parser[:dipole][2] * DEBYE2CUBY
			)
		end

		# Gradient is read from the .aux file
		if what.include?(:gradient)
			mopac_aux = MopacAux.new(in_calc_dir("mopac.in.aux"))
			results.gradient = mopac_aux.gradient
		end

		# Geometry optimization
		if @what.include?(:native_opt)
			# parser[:geo_conv_grad]
			mopac_aux = MopacAux.new(in_calc_dir("mopac.in.aux")) unless mopac_aux
			@geometry.update_from_vector!(mopac_aux.optimized_geometry_vec)
		end

		# COSMO details
		if @settings[:solvent_model] == :cosmo || @settings[:solvent_model] == :cosmo2
			Cuby::error "COSMO output not found in MOPAC output (calculation #{@name})" unless parser[:cosmo_area]
			Cuby::error "COSMO output not found in MOPAC output (calculation #{@name})" unless parser[:cosmo_volume]
			Cuby::error "COSMO output not found in MOPAC output (calculation #{@name})" unless parser[:cosmo_energy]
			cosmo_area = parser[:cosmo_area] # Area in A^2
			cosmo_volume =  parser[:cosmo_volume] # Volume in A^3
			cosmo_energy = parser[:cosmo_energy] * EV2KCAL # Dielectric energy
			# Pass it for processing
			mopac_cosmo2_nonpolar(results, cosmo_area, cosmo_volume, cosmo_energy)
		end

		# External correction read from files
		if fn = @settings[:development][:mopac_external_correction]
			if File.exist?(in_calc_dir(fn))
				corr = IO.readlines(in_calc_dir(fn))[0].to_f
				results.energy += corr
			else
				Cuby::warning("Development correction expected in file #{fn} (calculation #{@name})")
			end
		end

		return results
	end

	def mopac_write_cpe
		filename = in_calc_dir("cpe.dat")
		f = File.open(filename,"w+")
		@settings[:mopac_cpe].each_pair{|key, value|
			f.puts "#{key.gsub('__','   ')}    #{value}"
		}
		f.close
	end

	def mopac_cosmo2_nonpolar(results, cosmo_area, cosmo_volume, cosmo_energy)
		# SASA term
		if @settings.set?(:mopac_cosmo2_surf_tension)
			sasa_energy = cosmo_area * @settings[:mopac_cosmo2_surf_tension]
			results.energy_components[:cosmo2_nonpolar_e] = sasa_energy
			results.energy += sasa_energy
		end
	end

	def mopac_write_pch
		# File format described in http://openmopac.net/manual/QMMM.html
		filename = in_calc_dir("mol.in")
		f = File.open(filename,"w+")
		f.puts
		f.puts "#{@geometry.size} 0" # No of atoms and link atoms
		# Now geometry with potentials on atoms
		@geometry.each{|atom|
			phi = 0.0
			@point_charges.each{|pch|
				phi += 332.0 * pch.charge / atom.distance(pch)
			}
			f.printf("%4s%10.3f%10.3f%10.3f%12.5f\n", atom.element.to_s, atom.x, atom.y, atom.z, phi)
		}
		f.close
	end

	class MopacAux
		attr_reader :gradient
		attr_reader :optimized_geometry_vec

		def initialize(filename)
			# Read and parse the aux file

			f = File.open(filename, "r")

			# Block readers
			read_gradient = false
			read_opt_geo = false
			gradient = []
			opt_geo = []

			# Iterate through file
			while line = f.gets
				# Remove spaces around string
				line.strip!
				# Find blocks
				if line =~ /^GRADIENTS:KCAL\/MOL\/ANGSTROM/
					read_gradient = true
					next
				elsif line =~ /^ATOM_X_OPT:ANGSTROMS/
					read_opt_geo = true
					next
				end
				# Terminate blocks
				if line[0..0] =~ /[A-Z,a-z]/
					read_gradient = false
					read_opt_geo = false
					next
				end
				# Gradient block
				if read_gradient
					line.gsub("-"," -").split.each{|x| gradient << x.to_f}
				end
				# Geometry block
				if read_opt_geo
					line.gsub("-"," -").split.each{|x| opt_geo << x.to_f}
				end
			end
			f.close

			# Build the results
			# Gradient
			@gradient = Gradient.from_vector(gradient) if gradient.size > 0
			# geometry
			if opt_geo.size > 0
				@optimized_geometry_vec = opt_geo
			end
		end
	end
end
