################################################################################
#
# ORCA interface
#
# Author: Jan Rezac
# Date created: 2013-04-29
# License: Cuby4 license
# Description: Interface for external calculations using ORCA
# Status: Works
#
################################################################################

#===============================================================================
# Interface to ORCA
# http://www.cec.mpg.de/forum/portal.php
#===============================================================================

# Important notes:
#   * Geometry can not be readed from an external file, ghost atoms can not be
#     defined in such case.

require "classes/tools/output_parser.rb"
require "classes/calculation/dft_setup.rb"

module InterfaceOrca
	#=======================================================================
	# Interface header
	#=======================================================================
	DEVELOPMENT_FLAG = :ok
	DEVELOPMENT_STATUS = "Works"
	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy, :gradient, :ghost_atoms, :point_charges, :point_charges_gradient, :native_opt, :hessian]
	MODIFIER = false
	DIRECTORY = "ORCA"
	# Methods provided by the interface:
	METHODS = {
		:"hf"		=> [:energy, :gradient, :point_charges, :point_charges_gradient],
		:"dft"		=> [:energy, :gradient, :point_charges, :point_charges_gradient, :native_opt, :hessian],
		:"mp2"		=> [:energy, :gradient, :native_opt],
		:"mp3"		=> [:energy],
		:"ccsd"		=> [:energy],
		:"ccsd(t)"	=> [:energy],
		:"qcisd"	=> [:energy],
		:"qcisd(t)"	=> [:energy],
		:"hf-3c"	=> [:energy, :gradient],
		:"pbeh-3c"	=> [:energy, :gradient],
		:"b97-3c"	=> [:energy, :gradient],
		:"r2scan-3c"	=> [:energy, :gradient],
		# SQM methods working but yield different energy than mopac 
		#:"am1"		=> [:energy] # :point_charges do not work here
		#:"pm3"		=> [:energy] # :point_charges do not work here
	}
	SOLVENTS = [:cosmo, :smd, :cpcm]
	# Unsupported keywords: options common in similar interfaces that can't
	# be used
	UNSUPPORTED_KEYWORDS = [
		:use_symmetry # Symmetry is not supported meaningfully in orca
	]
	#=======================================================================

	@@orca_basis_sets = nil

	def prepare_interface
		# Version-dependent checks
		if @settings[:orca_version] >= "4"
			Cuby::error "COSMO solvent not available in Orca 4, use CPCM" if @settings[:solvent_model] == :cosmo
        end
		if @settings[:orca_version] < "5"
			Cuby::error "Method R2SCAN-3c not implemented in orca versions older than 5.0" if @settings[:method] == :"r2scan-3c"
        end
		if @settings[:orca_version] < "4"
			# 3.x
			Cuby::error "Method PBEh-3c not implemented in orca versions older than 4.0" if @settings[:method] == :"pbeh-3c"
			Cuby::error "Method B97-3c not implemented in orca versions older than 4.0" if @settings[:method] == :"b97-3c"
			Cuby::error "Solvent SMD not implemented in orca versions older than 4.0" if @settings[:solvent_model] == :smd
		end

		# Load basis set list on the first use
		unless @@orca_basis_sets
			if @settings[:orca_version] > "4"
				basfile = interface_dir + "/basissets_4.yaml"
			else
				basfile = interface_dir + "/basissets.yaml"
			end
			File.open(basfile){|f|
				@@orca_basis_sets = YAML.unsafe_load(f)
			}
		end

		# DFT setup
		if @settings[:method] == :dft
			# Load DFT setup
			if @settings[:orca_version] >= "5"
				func_list = "/dft_functionals_5.yaml"
			elsif @settings[:orca_version] >= "4.2"
				func_list = "/dft_functionals_4.2.yaml"
			elsif @settings[:orca_version] > "4"
				func_list = "/dft_functionals_4.yaml"
			else
				func_list = "/dft_functionals.yaml"
			end
			@dft_setup = DftSetup.new(interface_dir + func_list, interface_dir + "/dft_grids.yaml")
		end

		# Prepare calculation directory
		if calc_dir_mkdir("orca.header", "orca.out") == :old
			calc_using_input # Old input should be reused
			return
		end

		# Write the keyword part of the input
		orca_write_keywords # This writes file orca.header

		# Create complete input by combining orca.header and the current geometry
		orca_write_input_geo

		# Save info on the system upon writing the input
		calc_writing_input
	end

	def calculate_interface
		orca_run unless @settings[:existing_calc_dir] == :read_results && FileTest.exist?(calc_dir+"/orca.out")
		return orca_read
	end

	def cleanup_interface
		if @settings[:job_cleanup]
			# Delete everything
			calc_dir_delete 
		elsif @settings[:delete_large_files]
			# Delete only large files (and some unnecessary files as well)
			system("cd #{calc_dir}; rm -f orca.gbw orca.loc orca.VAUX_SYM.tmp")
		end
	end

	#=======================================================================
	# Private methods
	#=======================================================================

	def orca_write_keywords
		# Write header of the input file
		filename = in_calc_dir("orca.header")
		f = File.open(filename,"w+")

		# Density fitting
		scf_ri = @settings[:density_fitting] == :scf || @settings[:density_fitting] == :both
		correlation_ri = @settings[:density_fitting] == :correlation || @settings[:density_fitting] == :both

		# Memory
		mem = @settings[:mem] * @settings[:mem_core] / 100 / @settings[:parallel]
		f.puts "%MaxCore #{mem}"

		# Paralellization
		if @settings[:parallel] > 1
			f.puts "%pal"
		        f.puts " nprocs #{@settings[:parallel]}"
			f.puts "end"
		end

		# Method
		method = ""

		# Method prefix
		if @settings[:orca_dlpno] # DLPNO
			# Checks
			Cuby::error "DLPNO available only with CCSD and CCSD(T)" unless [:ccsd, :"ccsd(t)"].include?(@settings[:method])
			Cuby::error "DLPNO: keyword auxiliary_basis_mp2 must be set" unless @settings.set?(:auxiliary_basis_mp2)


			method += "DLPNO-"
			rimp2_basis = "/C"
		end

		# Method selection
		case @settings[:method]
		when :hf
			method += "HF"
		when :dft
			method += "DFT"

			# Determine RI mode
			riscf_basis = nil
			if scf_ri
				case @dft_setup.functional_type(@settings)
				when :gga, :meta_gga
					method += " RI"
					riscf_basis = "/J"
				when :hybrid, :meta_hybrid, :double_hybrid
					if @settings[:orca_rijcosx]
						method += " RIJCOSX"
						riscf_basis = "/J"
					else
						method += " RIJK"
						riscf_basis = "/JK"
					end
				when :range_separated, :double_hybrid_rs
					if @settings[:orca_rijcosx]
						method += " RIJCOSX"
						riscf_basis = "/J"
					else
						Cuby::error "Orca does not implement RI for range-separated functionals without RIJCOSX"
					end
				else
					Cuby::error "Orca interface does not implement RI settings for #{@dft_setup.functional_type(@settings)} functionals"
				end
			end

			# Dispersion correction
			case @settings[:orca_dispersion_correction]
			when :d3bj
				method += " d3bj"
			when :d3zero
				method += " d3zero"
			when :nl
				method += " NL"
			when :scnl
				method += " SCNL"
			when :d4
				method += " D4"
			end
			
			# Functional
			if @dft_setup.functional_type(@settings) == :double_hybrid
				if correlation_ri
					# RI for the MP2 part
					method += " RI-#{@dft_setup.functional(@settings)}"
					rimp2_basis = "/C"
				else
					method += " #{@dft_setup.functional(@settings)}"
				end
			else
				method += " #{@dft_setup.functional(@settings)}"
			end
		when :mp2
			method += "MP2"
			rimp2_basis = "/C" if correlation_ri
		when :mp3
			method += "MP3"
		when :ccsd
			method += "CCSD"
		when :"ccsd(t)"
			method += "CCSD(T)"
		when :qcisd
			method += "QCISD"
		when :"qcisd(t)"
			method += "QCISD(T)"
		when :"hf-3c"
			method += "HF-3c"
		when :"pbeh-3c"
			method += "PBEh-3c"
		when :"b97-3c"
			method += "B97-3c"
		when :"r2scan-3c"
			method += "R2SCAN-3c"
		when :"am1"
			method += "AM1"
		when :"pm3"
			method += "PM3"
		end

		# SCF/RI for all but DFT (which is handled separately above)
		unless [:dft, :am1, :pm3].include?(@settings[:method])
			if scf_ri
				method += " RI-JK"
				riscf_basis = "/JK"
			end
		end

		# RHF/UHF
		if @settings[:spin_restricted] == :rhf || (@settings[:spin_restricted] == :auto_uhf && @settings[:multiplicity] == 1)
			method += " RHF"
		else
			method += " UHF"
		end

		# SCF convergence
		method += " SCFCONV#{@settings[:scf_convergence]}"

		# DLPNO setup
		if @settings[:orca_dlpno]
			case @settings[:orca_dlpno_accuracy]
			when :loose
				method += " LoosePNO"
			when :normal
				method += " NormalPNO"
			when :tight
				method += " TightPNO"
			end
		end

		# Restart
		method += " NoAutoStart" if !@settings[:start_from_previous]

        # Frequencies
        if @what.include?(:hessian)
          method += " NumFreq"
        end

		f.puts "# Orca calculation for cuby"
		f.puts "! #{method}"

		# Extra keywords, "!" added if needed
		if @settings.set?(:orca_keywords)
			line = @settings[:orca_keywords]
			if line =~ /^!/
				f.puts line
			else
				f.puts "! " + line
			end
		end

		# Point charges
		if @point_charges
			f.puts '% pointcharges "pointcharges.pc"'
		end

		# Basis set
		basis_used_in_block = []
		unless [:"hf-3c", :"pbeh-3c", :"b97-3c", :"r2scan-3c", :am1, :pm3].include?(@settings[:method])
			check_basis(@settings[:basisset])
			basis = @settings[:basisset]

			# Auxiliary basis
			if riscf_basis
				# When no aux. basis selected, use name of AO basis
				unless @settings.set?(:auxiliary_basis_scf)
					if riscf_basis == "/J"
						Cuby::warning "Auxiliary basis for SCF not set, using default def2/J"
						@settings[:auxiliary_basis_scf] = "def2"
					else
						Cuby::warning "Auxiliary basis for SCF not set, will try one matching AO basis"
						@settings[:auxiliary_basis_scf] = @settings[:basisset]
					end
				end
				check_auxiliary_basis(@settings[:auxiliary_basis_scf] + riscf_basis)
				basis += " #{@settings[:auxiliary_basis_scf]}#{riscf_basis}"
				if riscf_basis == "/J"
					basis_used_in_block << "AuxJ \"#{@settings[:auxiliary_basis_scf]}#{riscf_basis}\""
				elsif riscf_basis == "/JK"
					basis_used_in_block << "AuxJK \"#{@settings[:auxiliary_basis_scf]}#{riscf_basis}\""
				end
			end
			if rimp2_basis
				# When no aux. basis selected, use name of AO basis
				unless @settings.set?(:auxiliary_basis_mp2)
					Cuby::warning "Auxiliary basis for MP2 not set, will try one matching AO basis"
					@settings[:auxiliary_basis_mp2] = @settings[:basisset]
				end
				check_auxiliary_basis(@settings[:auxiliary_basis_mp2] + rimp2_basis)
				basis += " #{@settings[:auxiliary_basis_mp2]}#{rimp2_basis}"
				basis_used_in_block << "AuxC \"#{@settings[:auxiliary_basis_mp2]}#{rimp2_basis}\""
			end

			# Basis set setup one-liner
			# Applicable only when built-in basis sets are used
			unless @settings.set?(:basisset_file)
				case @settings[:orca_ecp]
				when :auto
					f.puts "! ECP{#{basis.gsub(' ',',')}} #{basis}"
				when :none
					f.puts "! #{basis}"
				end
			end

			# Basis block
			f.puts "%basis"
			# Basis from file
			if @settings.set?(:basisset_file)
				# Basis set
				bfn = @settings[:basisset_file]
				if FileTest.exist?(bfn)
					bfn = File.expand_path(bfn)
				elsif FileTest.exist?(interface_dir + "/basissets/" + bfn)
					bfn = interface_dir + "/basissets/" + bfn
				else
					Cuby::error "basisset_file: file not found"
				end
				f.puts "   GTOName = \"#{bfn}\""
				# Auxiliary basis sets - stored previously
				basis_used_in_block.each{|l|
					f.puts "   #{l}"
				}
				# Pseudopotentials
				if @settings.set?(:orca_ecp_file)
					orca_write_ecp_input(f, @settings[:orca_ecp_file])
				end
			end
			# Per-element setup
			@settings.elements_hash(:basisset_elements).each_pair{|element, basis|
				check_basis(basis)
				f.puts "   NewGTO #{PeriodicTable.proton_number(element)}"
				f.puts "     \"#{basis}\""
				f.puts "   end"
				if riscf_basis
					unless @settings.elements_hash(:auxiliary_basis_scf_elements)[element]
						Cuby::error "Non-default basis is set for element #{element}, but SCF auxiliary basis is not defined for it"
					end
				end
				if rimp2_basis
					unless @settings.elements_hash(:auxiliary_basis_mp2_elements)[element]
						Cuby::error "Non-default basis is set for element #{element}, but MP2 auxiliary basis is not defined for it"
					end
				end

			}
			# Auxiliary basis per element
			if scf_ri
				@settings.elements_hash(:auxiliary_basis_scf_elements).each_pair{|element, basis|
					check_auxiliary_basis("#{basis}#{riscf_basis}")
					f.puts "   NewAuxGTO #{PeriodicTable.proton_number(element)}"
					f.puts "     \"#{basis}#{riscf_basis}\""
					f.puts "   end"
				}
			end
			if rimp2_basis
				@settings.elements_hash(:auxiliary_basis_mp2_elements).each_pair{|element, basis|
					check_auxiliary_basis("#{basis}#{rimp2_basis}")
					f.puts "   NewAuxGTO #{PeriodicTable.proton_number(element)}"
					f.puts "     \"#{basis}#{rimp2_basis}\""
					f.puts "   end"
				}
			end
			f.puts "end"
		end

		# SCF setup
		f.puts "%scf"
		case @settings[:scf_integral_storage]
		when :direct
			f.puts "   SCFMode Direct"
		when :mixed
			f.puts "   SCFMode SemiDirect"
		when :disc
			f.puts "   SCFMode Conventional"
		end
		# No. of SCF cycles
		f.puts "   MaxIter #{@settings[:scf_cycles]}"
		if @settings.set?(:orca_guess)
			f.puts "   Guess #{@settings[:orca_guess].to_s}"
		end
		f.puts "end"

		# Method setup % method
		f.puts "%method"
        if @what.include?(:gradient) && ! @what.include?(:hessian)
		  f.puts "   RunTyp Gradient"
        end
		f.puts "   RunTyp Opt" if @what.include?(:native_opt) #!# not tested

		# DFT
		if @settings[:method] == :dft
			# Functional setup (the following line) was moved to the !-line at the beginning of the input
			# f.puts "   Functional #{@dft_setup.functional(@settings)}"

			if @settings[:dft_grid] == :custom
				f.puts "   Grid #{@settings[:dft_grid_custom]}"
			else
				f.puts "   Grid #{@dft_setup.grid(@settings)[0]}"
				f.puts "   FinalGrid #{@dft_setup.grid(@settings)[1]}"
			end
            if @settings[:development][:b97md4]
              f.puts "   D3s6 1.0000"
              f.puts "   D3s8 0.6633"
              f.puts "   D3a1 0.4288"
              f.puts "   D3a2 3.9935"
            end
		end

		# MP2
		if @settings[:method] == :mp2
			f.puts @settings[:correlation_frozen_core] ? "   FrozenCore FC_ELECTRONS" : "   FrozenCore FC_NONE"
		end

		# End of %method
		f.puts "end"

		# MP2 setup
		if @settings[:method] == :mp2
			# MP2 block
			f.puts "%mp2"
			# RI on/off
			f.puts correlation_ri ? "   RI on" : "   RI off"
			# Print more results
			f.puts "   PrintLevel 3"
			# SCS-MP2
			if @settings[:spin_component_scaling]
				f.puts "   DoSCS true"
				f.puts "   Ps #{@settings[:scs_os]}"
				f.puts "   Pt #{@settings[:scs_ss]}"
			end
			# End of MP2 block
			f.puts "end"
		end

		# COSMO / CPCM
		case @settings[:solvent_model]
		when :cosmo
			f.puts "%cosmo"
			f.puts "   epsilon #{@settings[:solvent_epsilon]}"
			f.puts "end"
		when :cpcm
			f.puts "!cpcm(#{@settings[:solvent_name]})"
		when :smd
			f.puts "%cpcm"
			f.puts "   smd true"
			#f.puts "   smdsolvent \"#{@settings[:solvent_name]}\""
			f.puts "   smdsolvent \"#{@settings[:solvent_name]}\""
			f.puts "end"
		end

		# Native_opt
		if @what.include?(:native_opt)
			f.puts "%geom"
			f.puts "   coordsys #{@settings[:orca_opt_coords].to_s}"
			f.puts "   MaxIter #{@settings[:maxcycles]}"
			f.puts "   TolE #{@settings[:orca_optlimit_e]}"
			f.puts "   TolRMSG #{@settings[:orca_optlimit_g]}"
			f.puts "   TolMaxG #{@settings[:orca_optlimit_g]}"
			f.puts "   TolRMSD #{@settings[:orca_optlimit_d]}"
			f.puts "   TolMaxD #{@settings[:orca_optlimit_d]}"
			f.puts "end"
		end

		# Extra input - without any processing
		if @settings.set?(:orca_extra_input)
			f.puts @settings[:orca_extra_input]
		end

		# Coordinates start
		f.puts "* xyz #{@settings[:charge]} #{@settings[:multiplicity]}"
		f.close
	end

	def orca_write_input_geo
		# Write input file using the existing header, only geometry is added
		filename = in_calc_dir("orca.in")
		f = File.open(filename,"w+")

		# Header from file
		header = IO.readlines(in_calc_dir("orca.header"))
		f.puts header

		# Geometry
		@geometry.each_index{|i|
			atom = @geometry[i]
		 	ghost = atom.properties[:ghost] ? ":" : " "
		 	f.printf(" %3s %s %15.10f %15.10f %15.10f\n", atom.element.to_s, ghost, atom.x, atom.y, atom.z)
		}
		f.puts "*"

		f.close
	end

	def orca_run
		# Build input with the current geometry
		orca_write_input_geo
		# Point charges
		orca_write_point_charges if @point_charges

		# Run orca
		command = "cd #{calc_dir};"
		command << "export PATH=#{@settings[:orca_bin_dir]}:$PATH;"
		command << "export LD_LIBRARY_PATH=#{@settings[:orca_bin_dir]}:$LD_LIBRARY_PATH;"
		command << @settings[:orca_mpi_setup]
		command << "#{@settings[:orca_bin_dir]}/orca orca.in > orca.out 2> orca.err;"

		unless system(command)
			# Catch specific errors
			orca_parse_errors
			# Or default to general error message
			Cuby::error "ORCA returned nonzero exit code (calculation #{@name})"
		else
			# Errors may be there even if exit code is OK
			orca_parse_errors
		end
	end

	def orca_parse_errors
		results = Results.new
		parser = OutputParser.new(in_calc_dir("orca.out"), @name)

		parser.add_error(/not enough memory/, "ORCA failed, not enough memory (calculation #{@name})")
		parser.add_error(/This wavefunction IS NOT CONVERGED/, "ORCA failed, SCF not converged (calculation #{@name})")
		parser.add_error(/There are no AuxJK basis functions on/, "ORCA failed, JK auxliary basis missing for some atoms")

		parser.execute
	end

	def orca_read # => Results
		results = Results.new
		parser = OutputParser.new(in_calc_dir("orca.out"), @name)

		# Dispersion correction - either requested, or included in functional
		dispersion = @settings[:orca_dispersion_correction]
		# Read built-in dispersion for some functionals
		if @dft_setup
			case @dft_setup.functional(@settings)
			when 'B97-D3'
				dispersion = :d3bj
			when 'B97-D'
				dispersion = :d
			when 'wB97X-D3'
				dispersion = :d3zero
			when 'wB97X-D3BJ'
				dispersion = :d3bj
			when 'wB97M-D3BJ'
				dispersion = :d3bj
			when 'B97M-D3BJ'
				dispersion = :d3bj
                dispersion = :d4 if @settings[:orca_dispersion_correction] == :d4
			end
		end

		# SCF energy
		parser.add_pattern(:e_scf, /^Total Energy       : *([^ ]*)/, :get => :submatch, :multi => :last)
		# Total energy - final check
		parser.add_pattern(:e_total, /^FINAL SINGLE POINT ENERGY *([^ ]*)/, :get => :submatch, :multi => :last)

		# Method-specific
		case @settings[:method]
		when :dft
			if @settings[:density_fitting] == :correlation || @settings[:density_fitting] == :both
				parser.add_pattern(:corr_mp2, /^ RI-MP2 CORRELATION ENERGY/, :required => false, :multi => :last)
			else
				parser.add_pattern(:corr_mp2, /^ MP2 CORRELATION ENERGY/, :required => false, :multi => :last)
			end
		when :mp2
			if @settings[:density_fitting] == :correlation || @settings[:density_fitting] == :both
				parser.add_pattern(:corr_mp2, /^ RI-MP2 CORRELATION ENERGY/, :multi => :last)
			else
				parser.add_pattern(:corr_mp2, /^ MP2 CORRELATION ENERGY/, :multi => :last)
			end
		when :mp3
			parser.add_pattern(:corr_mp2, /EC\(MP2\)= *([^ ]*)/, :get => :submatch)
			parser.add_pattern(:corr_mp3, /EC\(MP3\)= *([^ ]*)/, :get => :submatch)
		when :"ccsd"
			parser.add_pattern(:"e_ccsd", /^FINAL SINGLE POINT ENERGY/)
		when :"ccsd(t)"
			parser.add_pattern(:"e_ccsd", /^E\(CCSD\)/)
			parser.add_pattern(:"e_ccsd(t)", /^E\(CCSD\(T\)\)/)
		when :qcisd
			parser.add_pattern(:"e_qcisd", /^FINAL SINGLE POINT ENERGY/)
		when :"qcisd(t)"
			parser.add_pattern(:"e_qcisd", /^E\(QCISD\)/)
			parser.add_pattern(:"e_qcisd(t)", /^E\(QCISD\(T\)\)/)
		when :"hf-3c"
			parser.add_pattern(:hf3c_correction, /^Total correction to HF\/MINIX *([^ ]*)/, :get => :submatch)
		when :"pbeh-3c"
			parser.add_pattern(:d3_correction, /^Dispersion correction *(-?[0-9]+\.[0-9]+)/, :get => :submatch)
			parser.add_pattern(:gcp_correction, /^gCP correction *(-?[0-9]+\.[0-9]+)/, :get => :submatch)
		when :"b97-3c"
			parser.add_pattern(:d3_correction, /^Dispersion correction *(-?[0-9]+\.[0-9]+)/, :get => :submatch)
			parser.add_pattern(:srb_correction, /^SRB correction *(-?[0-9]+\.[0-9]+)/, :get => :submatch)
		when :"r2scan-3c"
			parser.add_pattern(:d3_correction, /^Dispersion correction *(-?[0-9]+\.[0-9]+)/, :get => :submatch)
			parser.add_pattern(:gcp_correction, /^gCP correction *(-?[0-9]+\.[0-9]+)/, :get => :submatch)
		end

		# Dispersion correction
		case dispersion
		when :d3bj, :d3zero, :d
			parser.add_pattern(:d3_correction, /^Dispersion correction *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
		when :d4
			parser.add_pattern(:d4_correction, /^Dispersion correction *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
		when :nl
			if @settings[:orca_version] >= "5"
				parser.add_pattern(:nl_correction, /^NL *Energy: *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
				parser.add_pattern(:scf_without_nl, /^SCF Energy: *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
			else
				parser.add_pattern(:nl_correction, /^NL *Energy *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
				parser.add_pattern(:scf_without_nl, /^SC Energy *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
			end
		when :scnl
			parser.add_pattern(:nl_correction, /^NL *Energy,.*: *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
		end

		# Solvation
		case @settings[:solvent_model]
		when :cpcm
			parser.add_pattern(:g_solv, /^Corrected G\(solv\) *: *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
			parser.add_pattern(:cpcm_charge_corr, /^Charge-correction *: *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
			parser.add_pattern(:cpcm_cavdisp, /^Free-energy \(cav\+disp\) *: *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
		when :smd
			parser.add_pattern(:cpcm_dielectric, /^CPCM Dielectric *: *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
			parser.add_pattern(:smd_cds, /^SMD CDS \(Gcds\) *: *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
			parser.add_pattern(:smd_total, /^Total Energy after SMD CDS correction *= *(-?[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
		end


		# Gradient
		if @what.include?(:gradient)
			results.gradient = Gradient.new
			if @settings[:method] == :mp2
				parser.add_block(
					:gradient,
					/The final MP2 gradient/, 
					/^[^:]+: +([-]*[0-9]*\.[0-9]+) +([-]*[0-9]*\.[0-9]+) +([-]*[0-9]*\.[0-9]+)/, 
					/NORM|Difference/,
					{:get => :match_array, :type => :float}
				)
			else
				parser.add_block(
					:gradient,
					/CARTESIAN GRADIENT/, 
					/^[^:]+: +([-]*[0-9]*\.[0-9]+) +([-]*[0-9]*\.[0-9]+) +([-]*[0-9]*\.[0-9]+)/, 
					/Norm|Difference/,
					{:get => :match_array, :type => :float}
				)
			end
		end

		# DLPNO
		if @settings[:orca_dlpno]
			#parser.add_pattern(:corr_sl_pno_mp2, /SL-PNO-MP2 correlation energy \(all kept pairs\)[. ]*([-]*[0-9]*\.[0-9]+)/, :get => :submatch)
			#parser.add_pattern(:corr_dlpno_init, /Initial total correlation energy[. ]*([-]*[0-9]*\.[0-9]+)/, :get => :submatch)
		end

		# Native_opt
		if @what.include?(:native_opt)
			parser.add_pattern(
				:opt_maxcycles_exceeded,
				/The optimization did not converge but reached the maximum number/,
				:required => false, :get => :line, :type => :as_is
			)
			parser.add_pattern(
				:opt_converged,
				/THE OPTIMIZATION HAS CONVERGED/,
				:required => false, :get => :line, :type => :as_is
			)
		end

		# Run parser
		parser.execute{|name, result, count|
			case name
			when :gradient
				x,y,z = result
				results.gradient << Coordinate[x,y,z] * HARTREE2KCAL / BOHR2ANGSTROM
			end
		}

		if @what.include?(:gradient)
			unless results.gradient.size == @geometry.size
				Cuby::error "Gradient not read properly from ORCA output\n(found #{results.gradient.size} records instead of #{@geometry.size})"
			end
		end

		# SCF energy
		results.energy = parser[:e_scf] * HARTREE2KCAL
		results.energy_components[:scf_energy] = results.energy

		# Correlation
		case @settings[:method]
		when :dft
			# Add MP2 in double-hybrids
			if parser[:corr_mp2]
				results.energy_components[:correlation_mp2] = parser[:corr_mp2] * HARTREE2KCAL
				results.energy += results.energy_components[:correlation_mp2]
			end
		when :mp2
			if @settings[:spin_component_scaling]
				results.energy_components[:correlation_scs_mp2] = parser[:corr_mp2] * HARTREE2KCAL
				results.energy += results.energy_components[:correlation_scs_mp2]
			else
				results.energy_components[:correlation_mp2] = parser[:corr_mp2] * HARTREE2KCAL
				results.energy += results.energy_components[:correlation_mp2]
			end
		when :mp3
			results.energy_components[:correlation_mp2] = parser[:corr_mp2] * HARTREE2KCAL
			results.energy_components[:correlation_mp3] = parser[:corr_mp3] * HARTREE2KCAL
			results.energy += results.energy_components[:correlation_mp3]
		when :"ccsd"
			results.energy_components[:"correlation_ccsd"] = parser[:"e_ccsd"] * HARTREE2KCAL - results.energy_components[:scf_energy]
			results.energy = parser[:"e_ccsd"] * HARTREE2KCAL
		when :"ccsd(t)"
			results.energy_components[:"correlation_ccsd"] = parser[:"e_ccsd"] * HARTREE2KCAL - results.energy_components[:scf_energy]
			results.energy_components[:"correlation_ccsd(t)"] = parser[:"e_ccsd(t)"] * HARTREE2KCAL - results.energy_components[:scf_energy]
			results.energy = parser[:"e_ccsd(t)"] * HARTREE2KCAL
		when :"qcisd"
			results.energy_components[:"correlation_qcisd"] = parser[:"e_qcisd"] * HARTREE2KCAL - results.energy_components[:scf_energy]
			results.energy = parser[:"e_qcisd"] * HARTREE2KCAL
		when :"qcisd(t)"
			results.energy_components[:"correlation_qcisd"] = parser[:"e_qcisd"] * HARTREE2KCAL - results.energy_components[:scf_energy]
			results.energy_components[:"correlation_qcisd(t)"] = parser[:"e_qcisd(t)"] * HARTREE2KCAL - results.energy_components[:scf_energy]
			results.energy = parser[:"e_qcisd(t)"] * HARTREE2KCAL
		when :"hf-3c"
			results.energy_components[:hf3c_correction] = parser[:hf3c_correction] * HARTREE2KCAL
			results.energy += results.energy_components[:hf3c_correction]
		when :"pbeh-3c"
			results.energy_components[:d3_correction] = parser[:d3_correction] * HARTREE2KCAL
			results.energy_components[:gcp_correction] = parser[:gcp_correction] * HARTREE2KCAL
			results.energy += results.energy_components[:d3_correction] 
			results.energy += results.energy_components[:gcp_correction]
		when :"b97-3c"
			results.energy_components[:d3_correction] = parser[:d3_correction] * HARTREE2KCAL
			results.energy_components[:srb_correction] = parser[:srb_correction] * HARTREE2KCAL
			results.energy += results.energy_components[:d3_correction] 
			results.energy += results.energy_components[:srb_correction]
		when :"r2scan-3c"
			results.energy_components[:d3_correction] = parser[:d3_correction] * HARTREE2KCAL
			results.energy_components[:gcp_correction] = parser[:gcp_correction] * HARTREE2KCAL
			results.energy += results.energy_components[:d3_correction] 
			results.energy += results.energy_components[:gcp_correction]
		end

		# Solvation
		case @settings[:solvent_model]
		when :cpcm
			results.energy_components[:cpcm_charge_corr] = parser[:cpcm_charge_corr] * HARTREE2KCAL
			results.energy_components[:cpcm_cavdisp] = parser[:cpcm_cavdisp] * HARTREE2KCAL
			results.energy_components[:e_solv_only] = results.energy
			results.energy_components[:g_solv] = parser[:g_solv] * HARTREE2KCAL
			results.energy += results.energy_components[:cpcm_charge_corr]
			results.energy += results.energy_components[:cpcm_cavdisp]
		when :smd
			# Just to be sure, the total energy is replaced with another value from the output
			# which explictly states it contains the SMD correction
			results.energy = parser[:smd_total] * HARTREE2KCAL
			results.energy_components[:cpcm_dielectric] = parser[:cpcm_dielectric] * HARTREE2KCAL
			results.energy_components[:smd_cds] = parser[:smd_cds] * HARTREE2KCAL
		end

		# Dispersion correction
		case dispersion
		when :d3bj, :d3zero
			results.energy_components[:dispersion_correction_d3] = parser[:d3_correction] * HARTREE2KCAL
			results.energy += results.energy_components[:dispersion_correction_d3]
		when :d
			results.energy_components[:dispersion_correction] = parser[:d3_correction] * HARTREE2KCAL
			results.energy += results.energy_components[:dispersion_correction]
		when :nl
			results.energy_components[:dispersion_correction_nl] = parser[:nl_correction] * HARTREE2KCAL
			results.energy_components[:scf_energy] = parser[:scf_without_nl] * HARTREE2KCAL
		when :scnl
			results.energy_components[:dispersion_correction_nl] = parser[:nl_correction] * HARTREE2KCAL
		when :d4
			results.energy_components[:dispersion_correction_d4] = parser[:d4_correction] * HARTREE2KCAL
			results.energy += results.energy_components[:dispersion_correction_d4]
		end


		if @settings[:orca_dlpno]
			#results.energy_components[:correlation_pl_pno_mp2] = parser[:corr_sl_pno_mp2] * HARTREE2KCAL
			#results.energy_components[:correlation_dlpno_init] = parser[:corr_dlpno_init] * HARTREE2KCAL
		end

		# Gradient on point charges
		if @what.include?(:gradient) && @point_charges
		        @point_charges.gradient = orca_read_point_charges_gradient
			results.point_charges_gradient = @point_charges.gradient
		end

		if parser[:e_total]
			tot = parser[:e_total] * HARTREE2KCAL
			unless (results.energy - tot).abs < 1.0e-4
				Cuby::error("The energy read by cuby does not match the 'FINAL SINGLE POINT ENERGY' reported by Orca")
			else
				Cuby::log.puts_debug("Orca: 'FINAL SINGLE POINT ENERGY' check OK")
			end
		else
			Cuby::log.puts_debug("Orca: no 'FINAL SINGLE POINT ENERGY' record found")
		end

		# SMD deltaG correction - manual says:
		# one should take into account that the SMD model considers the same concentration (1 mol/L)
		# in both the gaseous and solution phases. Then, to compare the calculated dGS with its experimental
		# counterpart (where a gas-phase standard state of 1 atm is considered), a concentration term equal 
		# to 1.89 kcal/mol has to be added to the calculated dGS.
		if @settings[:solvent_model] == :smd && @settings[:orca_smd_free_energy]
			results.energy += 1.89
		end

		# Native opt
		if @what.include?(:native_opt)
			if parser[:opt_maxcycles_exceeded]
				Cuby::warning("Optimization not converged - maxcycles exceeded")
				results.status = :failed
			else
				unless parser[:opt_converged]
					Cuby::warning("Optimization not converged")
					results.status = :failed
				else
					results.status = :converged
				end
			end
			# Read optimized geometry
			g = Geometry.from_file(in_calc_dir("orca.xyz"))
			@geometry.copy_coordinates_from(g)
		end

        # Hessian
        if @what.include?(:hessian)
          results.hessian = orca_read_hessian
        end



		return results
	end

	def check_auxiliary_basis(basis_name)
		unless @@orca_basis_sets["fitting"].has_key?(basis_name.downcase)
			Cuby::recommendation "The selected auxiliay basis is not on the list of basis sets induded in ORCA.\nThis list might be incomplete but check the input before continuing.\n(Basis #{basis_name})"
		end
	end

	def check_basis(basis_name)
		return if basis_name.downcase == "custom"
		unless @@orca_basis_sets["ao"].has_key?(basis_name.downcase)
			Cuby::recommendation "The selected basis set is not on the list of basis sets induded in ORCA.\nThis list might be incomplete but check the input before continuing.\n(Basis #{basis_name})"
		end
	end

	def orca_write_point_charges
		# Write file with point charges in the calculation directory
		filename = in_calc_dir("pointcharges.pc")
		f = File.open(filename,"w+")
		f.puts @point_charges.size
		@point_charges.each { |pcharge|
			f.printf("%15.8f%15.8f%15.8f%15.8f\n", pcharge.charge, pcharge.x, pcharge.y, pcharge.z)
		}
		f.close
	end

	def orca_read_point_charges_gradient
		# Read gradient on point charges
		gradient = Gradient.new
		filename = in_calc_dir("orca.pcgrad")
		f = File.open(filename,"r")
		ncharges = f.gets.to_i
		unless ncharges == @point_charges.size
			Cuby::error "Number of point charge gradients read does not match input"
		end
		ncharges.times{|i|
			g = f.gets.strip.gsub("D","e").split.map{|x| x.to_f}
			gradient << Coordinate[g[0],g[1],g[2]] * HARTREE2KCAL / BOHR2ANGSTROM
		}
		f.close

		return gradient
	end

    def orca_read_hessian
      f = File.open(in_calc_dir("orca.hess"))
      n = nil
      while line = f.gets do
        line.strip!
        if line == "$hessian"
          n = f.gets.to_i
          hessian = Hessian.zero(n,n)
          cols = f.gets.split.map{|x| x.to_i }
          nc = cols.last
          while nc < n do
            n.times{
              l = f.gets.strip.split
              r = l.shift.to_i
              cols.each_index{|i|
                c = cols[i]
                hessian[r,c] = l[i].to_f * HARTREE2KCAL / BOHR2ANGSTROM / BOHR2ANGSTROM
              }
            }
            l = f.gets.strip
            break unless l =~ /^[0-9]/
            cols = l.split.map{|x| x.to_i }
            nc = cols.last
          end
          break
        end
      end
      hessian = (hessian + hessian.transpose) * 0.5
      return hessian
    end

	def orca_write_ecp_input(f, filename)
		# Write ECPs from an external file (filename) into the input file (f)
		efn = filename
		if FileTest.exist?(efn)
			efn = File.expand_path(efn)
		elsif FileTest.exist?(interface_dir + "/basissets/" + efn)
			efn = interface_dir + "/basissets/" + efn
		else
			Cuby::error "orca_ecp_file: file not found"
		end
		ecpf = File.open(efn)
		while l = ecpf.gets
			f.puts l 
		end
		ecpf.close
	end
end
