#===============================================================================
# Class for reading / assembling turbomole basis set files
#===============================================================================

class TurbomoleBasis

	def initialize(geometry, settings)
		@settings = settings
		@elements = geometry.elements_in_system
	end

	def build_basis_ao(basfile_name)
		# Where to read basis from
		basdir = @settings[:turbomole_turbodir] + "/" + "basen"

		# Open empty file
		f = File.open(basfile_name, "w+")

		# Header
		f.puts "$basis"
		f.puts "*"

		# Basisset_elements to hash
		basisset_elements = @settings.elements_hash(:basisset_elements)

		@elements.each{|element|
			name = @settings[:basisset]
			if basisset_elements[element]
				name = basisset_elements[element]
			end
			basfile = basdir + "/" + element.to_s.downcase
			basis = get_tm_basis(name, element, basfile)
			unless basis
				Cuby::error("Turbomole: basis sets '#{name}' not found in the library")
			end
			f.puts basis.to_s
		}

		f.puts "$end"
	end

	#=======================================================================
	# Private
	#=======================================================================
	
	class TmBasisRecord
		attr_accessor :name
		attr_accessor :header_comments
		attr_accessor :body

		def initialize
			@name = nil
			@header_comments = []
			@body = []
		end

		def to_s
			s =  @name + "\n"
			s += @header_comments.join("\n") + "\n"
			s += "*\n"
			s += @body.join("\n") + "\n"
			s += "*"
		end
	end
	
	def get_tm_basis(name, element, basfile)
		ele = element.to_s.downcase

		# Open basfile
		unless File.exist?(basfile)
			Cuby::error("Turbomole basis file not found: #{basfile}")
		end
		f = File.open(basfile)

		# Output: array of strings
		output = TmBasisRecord.new

		# Make basis set name compatible with regexp
		basname_safe = Regexp.escape(name)

		# Read basis
		status = :search
		while line = f.gets
			line.gsub!("\n","")
			if line =~ /^#{ele} #{basname_safe}$/ && status == :search
				# First line with the name of basis set
				output.name = line
				status = :in_bas_header
				next
			elsif status == :in_bas_header && line =~ /^#/
				# Comments in basis set header
				output.header_comments << line
				next
			elsif status == :in_bas_header && line =~ /^\*/
				# Start of the basis
				status = :in_bas
				next
			elsif status == :in_bas && line =~ /^\*/
				# End of basis
				status = :finished
				next
			elsif status == :in_bas && line =~ /^->/
				# Linked basis
				bas = line.gsub(/^->  *[a-z][a-z]*  */,"")
				basdata = get_tm_basis(bas, element, basfile)
				output.body << basdata.body
			elsif status == :in_bas
				# Basis itself
				output.body << line
				next
			end

		end

		f.close

		# If not found, return nil
		return nil if status == :search
		# Otherwise return the result
		return output
	end
end
