require "protocols/dataset/classes/plot.rb"

#=======================================================================
# Dataset definition classes, used in the yaml representation of the set
#=======================================================================

module ProtocolDataset
	class DataSetDescription
		attr_accessor :name
		attr_accessor :dataset_setup # hash, to be copied into the settings
		attr_accessor :global_setup # hash, to be copied into calculation settings
		attr_accessor :groups
		attr_accessor :multi_group_averages
		attr_accessor :references # hash DOI -> reference
		attr_accessor :text
		attr_accessor :unit
		# Additional metadata
		attr_accessor :notes
		attr_accessor :method_geometry
		attr_accessor :method_energy
		attr_accessor :groups_by
        attr_accessor :geoms_with_metadata


		def initialize
			# Sample only, real data are read from yaml
			@name = "Data set example"
			@dataset_setup = {}
			@global_setup = {}
			@groups = ["group"] # is nil where there are no groups
			@references = {}
			@text = ""
		end
	end

	class DataSetItem
		attr_accessor :name
		attr_accessor :shortname
		attr_accessor :geometry # file or yaml
		attr_accessor :reference_value
		attr_accessor :setup # hash, to be copied into calculation settings
		attr_accessor :group
		attr_accessor :tags # string
		attr_accessor :add_to_result
		# Special features for data set of curves
		attr_accessor :curve # identifier of the curve (name or number)
		attr_accessor :curve_x # value of x axis
		# Weight
		attr_accessor :error_weight
		# Internal
		attr_accessor :full_set_index # Index in full set(before filtering)

		def initialize
			# Sample only, real data are read from yaml
			@name = "System example"
			@shortname = "dataset_example"
			@geometry = "geofile.xyz"
			@reference_value = 0.0
			@setup = {}
			@group = "group"
			@tags = "example, item"
			@add_to_result = nil

			@curve = nil
			@curve_x = nil

			@error_weight = 1.0
		end
	end

	class DataSetPlot
		# A plot (e.g. dissociation curve) in dataset
		attr_accessor :name # String, printed in the plot
		attr_accessor :filename # String, filename-friendly
		attr_accessor :first # Index of first item in the series
		attr_accessor :last # Index of last item in the series
		attr_accessor :x_values # Array of the x values

		def collect_results(dataset, jobs, job_results)
			@x = []
			@y_ref = []
			@y = []
			@x_values.each_index{|i|
				item_i = @first + i
				@x << @x_values[i]
				@y_ref << dataset.original_items[item_i].reference_value
				@y << job_results[dataset.items.index(dataset.original_items[item_i])]
			}
		end

		def save(format, one_file, extra_cols, methodname, force_zero = false, xlabel="", ylabel="")
			case format
			when :text
				f = File.open(@filename+".txt", "w+")
				@x_values.each_index{|i|
					f.printf("%12.6f%12.6f%12.6f\n", @x[i], @y_ref[i], @y[i])
				}
				f.close
			when :gnuplot
				y_cols = [@y_ref, @y]
				y_min = [@y_ref.min, @y.min].min
				y_max = [@y_ref.max, @y.max].max
				y_names = ["reference", methodname]
				extra_cols.each{|series|
					y_names << series[0]
					y_cols << series[1]
					y_min = [y_min, series[1].min].min
					y_max = [y_max, series[1].max].max
				}
				plot = Plot.new(@x, y_cols, y_names)
				plot.title = @name
				plot.x_label = xlabel if xlabel != ""
				plot.y_label = ylabel if ylabel != ""
				if force_zero
					miny = [y_min, 0].min
					maxy = [y_max, 0].max
					miny < 0 ? miny *= 1.05 : miny *= 0.95
					maxy < 0 ? maxy *= 1.05 : maxy *= 0.95
					plot.y_range = [miny, maxy]
				end
				plot.gnuplot_png(@filename)
			when :gnuplot_script
				y_cols = [@y_ref, @y]
				y_min = [@y_ref.min, @y.min].min
				y_max = [@y_ref.max, @y.max].max
				y_names = ["reference", methodname]
				extra_cols.each{|series|
					y_names << series[0]
					y_cols << series[1]
					y_min = [y_min, series[1].min].min
					y_max = [y_max, series[1].max].max
				}
				plot = Plot.new(@x, y_cols, y_names)
				plot.x_label = xlabel if xlabel != ""
				plot.y_label = ylabel if ylabel != ""
				plot.title = @name
				if force_zero
					miny = [y_min, 0].min
					maxy = [y_max, 0].max
					miny < 0 ? miny *= 1.05 : miny *= 0.95
					maxy < 0 ? maxy *= 1.05 : maxy *= 0.95
					plot.y_range = [miny, maxy]
				end
				plot.gnuplot_script_and_data(@filename)
			when :yaml
				one_file.puts(({@name => @y}).to_yaml.gsub(/^---\s*$/,""))
			else
				raise "Unknown value of plot format"
			end
		end

		def is_complete?(dataset)
			@x_values.each_index{|i|
				item_i = @first + i
				return false unless dataset.items.index(dataset.original_items[item_i])
			}
			return true
		end
	end

	class DataSetReference
		# Behaves like array but provides metadata
		# Array of values
		attr_accessor :values
		# List of references - hash DOI => text
		attr_accessor :references
		# Additional metadata
		attr_accessor :notes
		attr_accessor :method_geometry
		attr_accessor :method_energy
		attr_accessor :errors


		def initialize
			@values = []
			@references = {}
		end

		def [](i)
			return @values[i]
		end
	end

	class DataSet
		attr_reader :description
		attr_accessor :items
		attr_accessor :alternative_reference # Hash of Arrays / DataSetReference objects
		attr_accessor :best_reference
		attr_accessor :plots
		attr_reader :original_items # Preserves the whole set before a selection is made

		def initialize
			# Sample only, real data are read from yaml
			@description = DataSetDescription.new
			@items = [DataSetItem.new]
			@alternative_reference = nil
		end

		def save_orig_items
			@original_items = []
			@items.each_index{|i|
				@original_items[i] = @items[i]
			}
			return nil
		end

		def save_plots(jobs, settings, job_results)
			return unless @plots
			format = settings[:dataset_save_plots]
			format = :gnuplot if settings[:dataset_save_plots] == :gnuplot_tiled

			# Get plots to be built
			active_plots = []
			@plots.each{|plot|
				# Check if the plot is complete in this selection
				active_plots << plot if plot.is_complete?(self)
			}
			if active_plots.empty?
				Cuby::warning("The selection applied does not yield complete curves for plotting")
				return
			end

			# Add extra cols from input
			extra_cols = {}
			active_plots.each{|plot| extra_cols[plot.name] = []}

			settings[:dataset_plot_extra].each_pair{|name, filename|
				f = File.open(filename)
				begin
					plotdata = YAML.unsafe_load(f)
				rescue => exc
					Cuby::error("Error parsing YAML file #{filename}:\n" + exc.message)
				end
				f.close

				active_plots.each{|plot|
					extra_cols[plot.name] << [name, plotdata[plot.name]]
				}
			}

			# Writting all into one file - prepare the file
			if settings[:dataset_save_plots] == :yaml
				one_file = File.open(settings[:dataset_plots_fn] + ".yaml", "w+")
			else
				one_file = nil
			end

		
			# Save plots
			active_plots.each_index{|i|
				plot = active_plots[i]
				plot.collect_results(self, jobs, job_results)
				plot.save(format, one_file, extra_cols[plot.name], settings[:dataset_plot_methodname], settings[:dataset_plot_scale_zero], xlabel=settings[:dataset_plot_x], ylabel=settings[:dataset_plot_y])
			}

			# Writting all into one file - close the file
			one_file.close if one_file

			# Tile images using imagemagick
			if settings[:dataset_save_plots] == :gnuplot_tiled
				filelist = ""
				active_plots.each{|plot| filelist += '"' + plot.filename + '.png" ' }

				begin
					# Tile images
					`montage #{filelist} -geometry +2+2 -tile #{settings[:dataset_plot_columns]}x #{settings[:dataset_plots_fn]}.png`
					# delete original files
					`rm -f #{filelist}`
				rescue
					Cuby::error "Calling imagemagick for composing the plots failed"
				end
			end
		end
	end
end
