require "protocols/dataset/classes/data_set_errors.rb"
require "protocols/dataset/classes/data_set_curve.rb"
require "classes/math/predictive_index.rb"

module ProtocolDataset
	#=======================================================================
	# Classes used for statistical analysis and printing of results
	#=======================================================================
	
	class DataSetStatistics
		attr_reader :errors
		attr_reader :rel_errors_pct
		attr_reader :curve_error_rmse
		attr_reader :curve_error_rmse_abs
		attr_reader :avg_ref_abs
		attr_accessor :unit_name

		def initialize(set_items, job_results, longnums = false, shortnames = false)
			# Default values
			@unit_name = "kcal/mol"

			# Store data
			@set_items = set_items
			@job_results = job_results
			
			# Output format
			@statformat = "%10.3f"
			@statformat = "%15.8f" if longnums
			@shortnames = shortnames

			# Calculate errors
			@errors = DataSetErrors.new
			@rel_errors_pct = DataSetErrors.new

			# Average values in the set
			@avg_abs = 0.0
			@avg_ref_abs = 0.0

			@job_results.each_index{|i|
				@errors[i] = @job_results[i] - @set_items[i].reference_value

				# Relative error - there's special case when both values are 0
				if @errors[i] == 0.0 && @set_items[i].reference_value == 0.0
					@rel_errors_pct[i] = 0.0
				else
					@rel_errors_pct[i] = @errors[i] / @set_items[i].reference_value.abs * 100.0
				end


				# Weighting of errors
				if @set_items[i].error_weight
					@errors[i] *= @set_items[i].error_weight
					@rel_errors_pct[i] *= @set_items[i].error_weight
				end

				@avg_abs += @job_results[i].abs
				@avg_ref_abs +=  @set_items[i].reference_value.abs
			}

			@avg_abs /= @job_results.size
			@avg_ref_abs /= @job_results.size

			# Correlation measures
			require "classes/math/statistics_xy"
			x = []
			y = []
			@job_results.each_index{|i|
				x[i] = @set_items[i].reference_value
				y[i] = @job_results[i]
			}
			stats_xy = StatisticsXY.new(x,y)
			@r_squared = stats_xy.pearson_r_squared
			@predictive_index = stats_xy.predictive_index
			@kendall_tau = stats_xy.kendall_tau


			# If dataset contains curves, evaluate extra data
			if @set_items.first.curve
				curve_statistics_eval
			end
		end

		def print_table(group_separator = "")
			# header
			puts "=========================================================================================="
			printf("%-40s",'name')
			printf("%10s",'E')
			printf("%10s",'Eref')
			printf("%10s",'error')
			printf("%10s",'error(%)')
			puts
			puts "------------------------------------------------------------------------------------------"

			# Contents
			@set_items.each_index{|i|
				# group separator
				if group_separator != "" && i > 0 && @set_items[i].group != @set_items[i-1].group
					puts group_separator
				end

				# Name
				if @shortnames
					name = @set_items[i].shortname
				else
					name = @set_items[i].name
				end
				# Shorten name to 40 characters
				if name.size > 40
					name = name[0..36] + "..."
				end
				printf("%-40s",name)

				printf("%10.3f", @job_results[i])
				printf("%10.3f", @set_items[i].reference_value)
				printf("%10.3f", @errors[i])
				printf("%10.3f", @rel_errors_pct[i])
				puts
			}
		end

		def print_stats
			puts "=========================================================================================="
			printf("%-15s#{@statformat}   %s\n", "RMSE", @errors.rmse, @unit_name)
			printf("%-15s#{@statformat}   %s\n", "MUE", @errors.mue, @unit_name)
			puts "------------------------------------------------------------------------------------------"
			printf("%-15s#{@statformat}   %s\n", "MSE", @errors.mse, @unit_name)
			printf("%-15s#{@statformat}   %s\n", "min", @errors.min, @unit_name)
			printf("%-15s#{@statformat}   %s\n", "max", @errors.max, @unit_name)
			printf("%-15s#{@statformat}   %s\n", "range", @errors.max - @errors.min, @unit_name)
			printf("%-15s#{@statformat}   %s\n", "min abs", @errors.min_abs, @unit_name)
			printf("%-15s#{@statformat}   %s\n", "max abs", @errors.max_abs, @unit_name)
		end

		def print_stats_rel
			puts "=========================================================================================="
			printf("%-15s#{@statformat}   %s\n", "RMSE", @rel_errors_pct.rmse, "%")
			printf("%-15s#{@statformat}   %s\n", "MUE", @rel_errors_pct.mue, "%")
			printf("%-15s#{@statformat}   %s\n", "MSE", @rel_errors_pct.mse, "%")
			printf("%-15s#{@statformat}   %s\n", "min", @rel_errors_pct.min, "%")
			printf("%-15s#{@statformat}   %s\n", "max", @rel_errors_pct.max, "%")
			printf("%-15s#{@statformat}   %s\n", "range", @rel_errors_pct.max - @rel_errors_pct.min, "%")
			printf("%-15s#{@statformat}   %s\n", "min abs", @rel_errors_pct.min_abs, "%")
			printf("%-15s#{@statformat}   %s\n", "max abs", @rel_errors_pct.max_abs, "%")
			puts "------------------------------------------------------------------------------------------"
			printf("%-15s#{@statformat}   %s\n", "RMSE/|avg|", @errors.rmse / @avg_ref_abs * 100, "%")
			printf("%-15s#{@statformat}   %s\n", "MUE/|avg|", @errors.mue / @avg_ref_abs * 100, "%")
		end

		def print_stats_correlation
			puts "=========================================================================================="
			#printf("%-15s#{@statformat}   %s\n", "Covariance", @covariance, "")
			printf("%-15s#{@statformat}   %s\n", "R^2", @r_squared, "")
			printf("%-15s#{@statformat}   %s\n", "Predictive ind.", @predictive_index, "")
			printf("%-15s#{@statformat}   %s\n", "Kendall tau", @kendall_tau, "")

		end

		def print_stats_groups(groups)
			puts "=========================================================================================="
			puts "Statistics for groups:"
			rel_rmse = {}
			groups.each{|group|
				# Build error set
				gerrors = DataSetErrors.new
				avg_ref_abs = 0.0
				count = 0.0
				@job_results.each_index{|i|
					if @set_items[i].group && @set_items[i].group.downcase == group.downcase
						# Averaging
						avg_ref_abs += @set_items[i].reference_value.abs
						count += 1
						# Save error
						gerrors << @job_results[i] - @set_items[i].reference_value
					end
				}
				# Average in the group
				avg_ref_abs = avg_ref_abs / count
				rel_rmse[group] = gerrors.rmse / avg_ref_abs * 100.0
				# Print
				printf("%-15s%-5s%-5s#{@statformat}   %-5s#{@statformat}   %s\n", group, "(#{gerrors.size})", "RMSE", gerrors.rmse, "MSE", gerrors.mse, @unit_name)

			}
			# Print relative RMSE
			puts "--------------------------------------------------------------------------------"
			groups.each{|group|
				printf("%-15s%-15s%.2f %\n", group, "RMSE/|avg|", rel_rmse[group])
			}

		end

		def print_stats_multigroup(groups, multi_group_averages)
			puts "-" * 90
			puts "Averages over multiple groups:"

			# Calculate errors in groups
			err_in_group = {}
			groups.each{|group|
				# Build error set
				gerrors = DataSetErrors.new
				@job_results.each_index{|i|
					if @set_items[i].group && @set_items[i].group.downcase == group.downcase
						gerrors << @job_results[i] - @set_items[i].reference_value
					end
				}
				err_in_group[group] = gerrors
			}

			multi_group_averages.each_pair{|metagroup, members|
				avg = 0.0
				members.each{|groupname|
					avg += err_in_group[groupname].rmse
				}
				avg /= members.size
				printf("%-15s |RMSE| #{@statformat} %s\n", metagroup, avg, @unit_name )
			}
		end

		def print_array
			str_array = []
			@set_items.each_index{|i|
				str_array << sprintf("%.3f", @job_results[i])
			}
			puts "[ " + str_array.join(", ") + " ]"
		end

		#---------------------------------------------------------------
		# Curves
		#---------------------------------------------------------------

		def curve_statistics_eval
			# Check if curve onformation is present
			unless @set_items.first.curve
				@curve_stats_available = false
				return
			end

			# Try to build curve info - may fail when only some point are selected
			begin
				@curves = {}
				# Construct and fill curve objects
				@job_results.each_index{|i|
					# Create new curve if it does not exist
					unless @curves[@set_items[i].curve]
						@curves[@set_items[i].curve] = DataSetCurve.new(@set_items[i].curve)
					end
					# Add a point to curve
					@curves[@set_items[i].curve].add_point(
						@set_items[i].curve_x, 
						@set_items[i].reference_value, 
						@job_results[i],
						@set_items[i].curve_x == 1.0
					)
				}
				# Sort points in curves by x values, get max and min size
				sizes = []
				@curves.each_value{|curve|
					curve.sort_points!
					sizes << curve.size
				}
				@min_curve_size = sizes.min
				@max_curve_size = sizes.max

				# Common x axis taken from first curve
				@curves_x = []
				@curves.values.first.points.each{|p|
					@curves_x << p.x
				}

				# Error at each distance, respective to minimum
				curves_sume = {}
				curves_sumabse = {}
				curves_sumsqe = {}
				@curves_x.each{|x|
					curves_sume[x] = 0.0
					curves_sumabse[x] = 0.0
					curves_sumsqe[x] = 0.0
				}

				# Accumulate sums in each point
				@curves.each_value{|curve|
					curve.points.each{|p|
						d_ref = p.y_ref - curve.refpoint.y_ref
						d_calc = p.y - curve.refpoint.y
						err = d_calc - d_ref
						curves_sume[p.x] += err
						curves_sumabse[p.x] += err.abs
						curves_sumsqe[p.x] += err**2
					}
				}

				# Build errors
				@curve_error_rmse = {}
				@curve_error_mue = {}
				@curve_error_mse = {}
				@curves_x.each{|x|
					@curve_error_rmse[x] = (curves_sumsqe[x] / @curves.size)**0.5
					@curve_error_mue[x] = (curves_sumabse[x] / @curves.size)
					@curve_error_mse[x] = (curves_sume[x] / @curves.size)
				}

				# REPEAT FOR ABS. ERRORS

				# Error at each distance
				curves_sume = {}
				curves_sumabse = {}
				curves_sumsqe = {}
				@curves_x.each{|x|
					curves_sume[x] = 0.0
					curves_sumabse[x] = 0.0
					curves_sumsqe[x] = 0.0
				}

				# Accumulate sums in each point
				@curves.each_value{|curve|
					curve.points.each{|p|
						err = p.y - p.y_ref
						curves_sume[p.x] += err
						curves_sumabse[p.x] += err.abs
						curves_sumsqe[p.x] += err**2
					}
				}

				# Build errors
				@curve_error_rmse_abs = {}
				@curve_error_mue_abs = {}
				@curve_error_mse_abs = {}
				@curves_x.each{|x|
					@curve_error_rmse_abs[x] = (curves_sumsqe[x] / @curves.size)**0.5
					@curve_error_mue_abs[x] = (curves_sumabse[x] / @curves.size)
					@curve_error_mse_abs[x] = (curves_sume[x] / @curves.size)
				}

				@curve_stats_available = true
			rescue
				@curve_stats_available = false
			end

		end

		def curve_statistics_print
			return unless @curve_stats_available
			puts "=========================================================================================="
			#--------------------------------------------------
			# puts "Errors along the curve, relative to reference point"
			# printf("%8s%10s%10s%10s\n", "x", "RMSE", "MUE", "MSE")
			# @curves_x.each{|x|
			# 	printf("%8.2f%10.3f%10.3f%10.3f\n", 
			# 	       x, 
			# 	       @curve_error_rmse[x],
			# 	       @curve_error_mue[x],
			# 	       @curve_error_mse[x]
			# 	)
			# }
			#-------------------------------------------------- 
			puts "Errors along the curve, total"
			printf("%8s%10s%10s%10s\n", "x", "RMSE", "MUE", "MSE")
			@curves_x.each{|x|
				printf("%8.2f%10.3f%10.3f%10.3f\n", 
				       x, 
				       @curve_error_rmse_abs[x],
				       @curve_error_mue_abs[x],
				       @curve_error_mse_abs[x]
				)
			}
		end

		#===============================================================
		# Private
		#===============================================================
		
		def sgn(x)
			return  0.0 if x == 0.0
			return -1.0 if x < 0.0
			return  1.0 if x > 0.0
			raise "Wrong argument #{x} passed to sgn(x)"
		end
	end
end
