# TBD:
# verbosity handling
# item selection by name, tag and group
# weighting (name-based, group-based etc)
# spreadsheet-friendly output

require "classes/misc/search_for_file.rb"
require "protocols/dataset/classes/data_set.rb"
require "protocols/dataset/classes/data_set_statistics.rb"
require "classes/compatibility/yaml_loader.rb"

module ProtocolDataset
	#=======================================================================
	# Protocol header
	#=======================================================================
	PROTOCOL_TYPE = :batch
	PROTOCOL_DESCRIPTION = "Calculation over a predefined data set and processing of the results"

	INPUT_BLOCKS = [
		InputBlock[:calculation_overwrite, :optional, "Calculation setup that does not fit at the root level"]
	]
	#=======================================================================

	DATASETS_PATH = "data/datasets"
	DATASETS_PATH_DEV = "data/datasets_devel"

	def print_header_protocol
		# Print header
		Cuby.log.print_cuby_logo(:comment => false, :text => 'Dataset calculation')
	end

	def prepare_protocol
		# Messages to be printed
		@print_messages = []

		# Load data set
		if @settings[:dataset].downcase == "from_files"
			# Ad-hoc dataset from geometry files
			init_dataset_from_geometry_files
		else
			# Read dataset from YAML file
			init_dataset_from_file
		end

		# Create child jobs
		@item_jobs = []

		# Check whether there is a job definition
		settings_global = Settings.from_hash(@set.description.global_setup, @settings)
		Cuby::error("Job type for dataset items must be set in description:global_setup in the yaml file") unless settings_global.set?(:job)

		@set.items.each_index{|i|
			item = @set.items[i]
			# Build Settings for the item
			# Copy from root level, exclude the calculation blocks already made and all dataset-specific keywords
			# not needed in the calculations
			settings = @settings.child_block_as_copy(("set_item_" + item.shortname).to_sym, [/^set_item_/, /^dataset.*/, /^calculation_overwrite$/])
			settings.copy_from!(settings_global, :overwrite) # Add global dataset setup from dataset file
			settings.copy_from!(@settings.block(:calculation_overwrite), :overwrite) if @settings.has_block?(:calculation_overwrite) # Add global setup from input
			settings.copy_from!(Settings.from_hash(item.setup, @settings), :overwrite) if item.setup # Add item-specific setup if exist
			# Set the geometry
			if File.exist?(@dataset_dir + "/" + item.geometry)
				settings[:geometry] = @dataset_dir + "/" + item.geometry
			else
				settings[:geometry] = item.geometry
			end

			# Create the jobs
			@item_jobs[i] = Job.new(@name + "_" + item.shortname, settings)
		}

		# Prepare the jobs
		@item_jobs.each{|job| job.prepare }
	end

	def run_protocol(queue)
		# Before calculation, update the dataset_add_ref_to_result
		# expressions - used in parameterization
		update_add_ref
		# Submit jobs to queue
		@item_jobs.each{|job| job.run}
	end

	def results
		# Results are collected only first time they are needed
		unless @results
			# Get results
			@job_results = []
			@set.items.each_index{|i|
				@job_results[i] = @item_jobs[i].single_number_result
			}

			# Modify the results
			if @settings.set?(:dataset_add_ref_to_result)
				@set.items.each_index{|i|
					if @set.items[i].add_to_result.kind_of? Numeric
						@job_results[i] += @set.items[i].add_to_result
					else
						Cuby::error("Failed calculating dataset_add_ref_to_result for system:" +
							"\n   #{@set.items[i].name}" +
							"\nPossible causes: the reference used is not available for this system,"+
							"\nor evaluation of the expression failed:" +
							"\n" +  @set.items[i].add_to_result.message
						)
					end
				}
			end
			# Processing of results
			case @settings[:dataset_processing]
			when :relative_to_average
				average = 0.0
				@set.items.each_index{|i|
					average += @job_results[i]
				}
				average = average / @item_jobs.size
				@set.items.each_index{|i|
					@job_results[i] -= average
				}
				# Now the same for reference values
				average = 0.0
				@set.items.each_index{|i|
					average += @set.items[i].reference_value
				}
				average = average / @item_jobs.size
				@set.items.each_index{|i|
					@set.items[i].reference_value -= average
				}
			when :relative_to_group_average
				averages = {}
				averages.default = 0.0
				sizes = {}
				sizes.default = 0.0
				# Colect sums and sizes
				@set.items.each_index{|i|
					averages[@set.items[i].group] += @job_results[i]
					sizes[@set.items[i].group] += 1
				}
				# Calculate averages
				averages.each_key{|key|
					averages[key] = averages[key] / sizes[key]
				}
				# Subtract averages
				@set.items.each_index{|i|
					@job_results[i] -= averages[@set.items[i].group]
				}
				# Now the same for reference values
				averages = {}
				averages.default = 0.0
				sizes = {}
				sizes.default = 0.0
				# Colect sums and sizes
				@set.items.each_index{|i|
					averages[@set.items[i].group] += @set.items[i].reference_value
					sizes[@set.items[i].group] += 1
				}
				# Calculate averages
				averages.each_key{|key|
					averages[key] = averages[key] / sizes[key]
				}
				# Subtract averages
				@set.items.each_index{|i|
					@set.items[i].reference_value -= averages[@set.items[i].group]
				}
			when :relative_to_min
				# Find minimum
				minimum = 9.9e99
				min_calc = 9.9e99
				@set.items.each_index{|i|
					if @set.items[i].reference_value < minimum
						minimum = @set.items[i].reference_value
						min_calc = @job_results[i]
					end
				}
				# Subtract
				@set.items.each_index{|i|
					@set.items[i].reference_value -= minimum
					@job_results[i] -= min_calc
				}
			when :relative_to_group_min
				# Find minimum benchmark value in each group
				minima = {}
				minima.default = 9.9e99
				minima_calc = {}
				@set.items.each_index{|i|
					if @set.items[i].reference_value < minima[@set.items[i].group]
						minima[@set.items[i].group] = @set.items[i].reference_value
						minima_calc[@set.items[i].group] = @job_results[i]
					end
				}
				# Subtract minimum from benchmark and from calculated values
				@set.items.each_index{|i|
					@set.items[i].reference_value -= minima[@set.items[i].group]
					@job_results[i] -= minima_calc[@set.items[i].group]
				}
			end
			# Statistics
			@results = DataSetStatistics.new(@set.items, @job_results, @settings[:dataset_long_numbers], @settings[:dataset_short_names])
			# Unit for results, default is "kcal/mol"
			@results.unit_name = @set.description.unit if @set.description.unit
		end
		return @results
	end

	def print_protocol
		# print dataset info
		# #!#
		#--------------------------------------------------
		# if @set.description.references
		# 	puts "Cite"
		# 	@set.description.references.each_pair{|doi, citation|
		# 		puts "   #{citation} DOI: #{doi}"
		# 	}
		# 	puts
		# end
		#-------------------------------------------------- 
		# Print messages
		@print_messages.each{|message|
			Cuby::log.puts_v(:normal, message)
		}

		# Print results
		results.print_table(@settings[:dataset_group_separator]) if @settings[:dataset_print].include?(:table)
		results.print_stats if @settings[:dataset_print].include?(:stats)
		results.print_stats_rel if @settings[:dataset_print].include?(:stats_rel)
		results.print_stats_groups(@set.description.groups) if @settings[:dataset_print].include?(:stats_groups) && @set.description.groups
		results.print_stats_multigroup(@set.description.groups, @set.description.multi_group_averages) if @settings[:dataset_print].include?(:stats_groups) && @set.description.multi_group_averages
		results.curve_statistics_print if @settings[:dataset_print].include?(:stats_curves)
		results.print_stats_correlation if @settings[:dataset_print].include?(:correlation)
		puts "==========================================================================================" unless @settings[:dataset_print].empty?
		# Save plots
		if @set.plots && @settings[:dataset_save_plots] != :none
			Cuby::log.puts_debug("Saving plots")
			@set.save_plots(@item_jobs, settings, @job_results)
		end
		# Print array for use in datasets
		results.print_array if @settings[:dataset_print].include?(:results_array)
	end

	def cleanup_protocol
		@item_jobs.each{|job| job.cleanup}
	end

	#=======================================================================
	# Optional protocol functionality
	#=======================================================================
	
	def single_number_result
		if @settings[:development][:dataset_snr_expression]
			return eval(@settings[:development][:dataset_snr_expression])
		else
			return results.errors.rmse
		end
	end

	#=======================================================================
	# Private methods
	#=======================================================================
	
	def init_dataset_from_file
		unless fn = SearchForFile.search(:filename => @settings[:dataset], :extensions => ['', '.yaml'], :raise_error => false)
			# The dataset keyword does not point to a file, look for predefined datasets
			dataset_name = @settings[:dataset].strip.gsub(" ","_").downcase
			# Firstly, a yaml file
			# Secondly, file set.yaml in a directory named after the dataset
			unless fn = Cuby.install_dir + "/" + DATASETS_PATH + "/#{dataset_name}.yaml"
				fn = Cuby.install_dir + "/" + DATASETS_PATH + "/#{dataset_name}/set.yaml"
			end

			# Try another location - datsets in development
			unless FileTest.exist?(fn)
				unless fn = Cuby.install_dir + "/" + DATASETS_PATH_DEV + "/#{dataset_name}.yaml"
					fn = Cuby.install_dir + "/" + DATASETS_PATH_DEV + "/#{dataset_name}/set.yaml"
				end
				Cuby::warning("Using a data set under development") if FileTest.exist?(fn)
			end

			Cuby::error("Dataset not found (neither file nor predefined dataset)") unless FileTest.exist?(fn)
		end

		Cuby::error("Dataset keyword does not points to a valid file") unless FileTest.file?(fn)
		@dataset_dir = File.dirname(fn) # Path, later used for loading geometries

		# Load set data from YAML
		File.open(fn, "r") {|f|
			begin
				@set = YAMLLoader.load(f, [DataSet, DataSetDescription, DataSetItem, DataSetPlot, DataSetReference, Symbol])
			rescue
				Cuby::error("Error reading dataset file")
			end
		}
		# Check type
		Cuby::error("Invalid data set file, does not contain DataSet object\n(file #{fn})") unless @set.class == DataSet

		# Sort data set items (used only in preparation of data sets)
		#--------------------------------------------------
		# @set.items.sort!{|a,b|
		# 	a.shortname <=> b.shortname
		# }
		# puts @set.to_yaml
		# exit
		#-------------------------------------------------- 
		
		# Fix alternative reference format
		if @set.alternative_reference
			@set.alternative_reference.each_pair{|key, value|
				if value.class == Array && value.first.class == Array
					value.flatten!
				end
			}
		end

		# Modify job setup
		if @set.description.dataset_setup
			settings = Settings.from_hash(@set.description.dataset_setup, @settings)
			@settings.copy_from!(settings, :keep)
		end

		# Save original items
		@set.save_orig_items
		@set.items.each_index{|i|
			@set.items[i].full_set_index = i
		}

		# Load items to be modified
		all_items = @set.items

		# Best alternative reference
		if @settings[:dataset_reference].downcase == "best"
			if @set.best_reference
				@settings[:dataset_reference] = @set.best_reference
				@print_messages << "Using 'best' reference: #{@set.best_reference}"
				if @set.alternative_reference[@set.best_reference].class == DataSetReference && @settings[:dataset_print].include?(:references)
					@print_messages << "Cite the reference as:"
					citations = @set.alternative_reference[@set.best_reference].references
					citations.each_pair{|doi, text|
						@print_messages << "   " + text 
						@print_messages << "   http://dx.doi.org/" + doi
					}
				end
			else
				if @set.alternative_reference
					# More references are present but none is labeled as best
				Cuby::warning("Dataset #{@set.description.name} doesn't have best reference defined, default reference will be used")
				end
				@settings[:dataset_reference] = "default"
			end
		end

		# Load alternative reference if applicable
		if @settings[:dataset_reference].downcase == "default"
			# Default reference, nothing to be done
		elsif @settings[:dataset_reference].downcase == "zero"
			# Set all references to zero
			@set.items.each_index{|i|
				@set.items[i].reference_value = 0.0
			}
		else
			if @set.alternative_reference
				if @set.alternative_reference.has_key?(@settings[:dataset_reference])
					@set.items.each_index{|i|
						# The referene may be incomplete (with nil value) but this is checked later
						@set.items[i].reference_value = @set.alternative_reference[@settings[:dataset_reference]][i]
					}
				else
					Cuby::error("Alternative reference data requested but not found in the data set file")
				end
			else
				Cuby::error("Alternative reference data requested but not found in the data set file")
			end
		end

		# Add to results data
		update_add_ref

		# Get geometries for all items
		all_geos = []
		# Loading the geometry must be aware of actual settings
		geo_settings = Settings.new
		geo_settings.copy_from!(@settings, :overwrite)
		all_items.each{|item|
			# Get the geometry
			if File.exist?(@dataset_dir + "/" + item.geometry)
				geo_settings[:geometry] = @dataset_dir + "/" + item.geometry
			else
				geo_settings[:geometry] = item.geometry
			end
			all_geos << Geometry.load_from_settings(geo_settings)
		}

		# Apply selection filters
		name_filter = Regexp.new(@settings[:dataset_select_name]) if @settings.set?(:dataset_select_name)
		tag_filter = Regexp.new(@settings[:dataset_select_tag]) if @settings.set?(:dataset_select_tag)
		group_filter = Regexp.new(@settings[:dataset_select_group]) if @settings.set?(:dataset_select_group)
		name_skip = Regexp.new(@settings[:dataset_skip_name]) if @settings.set?(:dataset_skip_name)
		tag_skip = Regexp.new(@settings[:dataset_skip_tag]) if @settings.set?(:dataset_skip_tag)
		group_skip = Regexp.new(@settings[:dataset_skip_group]) if @settings.set?(:dataset_skip_group)
		@set.items = []
		all_items.each_index{|i|
			item = all_items[i]
			if (!name_filter || name_filter.match(item.name)) && 
			   (!tag_filter || tag_filter.match(item.tags)) && 
			   (!group_filter || group_filter.match(item.group)) &&
			   (!name_skip || !name_skip.match(item.name)) &&
			   (!tag_skip || !tag_skip.match(item.tags)) &&
			   (!group_skip || !group_skip.match(item.group)) &&
			   all_geos[i].element_match(@settings[:dataset_select_elements])
				@set.items << item
			end
		}

		if @set.items.size != all_items.size

			@print_messages << "Selected #{@set.items.size} out of #{all_items.size} data set entries"
		end

		if @set.items.size == 0
			Cuby::error("There are no items in the set matching the selection criteria")
		end

		# Check reference values
		@set.items.each{|item|
			unless item.reference_value
				Cuby::error("Selected reference value not available for entry:\n   #{item.name}")
			end
		}
	end

	def init_dataset_from_geometry_files
		# Initialize data set
		@set = DataSet.new
		@dataset_dir = "."

		# Data set description
		@set.description.name = "Ad-hoc data set"
		@set.description.groups = nil

		@set.description.global_setup = {"job" => @settings[:dataset_from_files_job]}

		# List of files to be used
		filelist = Dir.glob(@settings[:dataset_from_files])
		if filelist.size == 0
			Cuby::error("No files matching selection '#{@settings[:dataset_from_files]}'")
		end
		filelist.sort!

		# Build data set entries
		@set.items = []
		filelist.each{|fn|
			filefn = File.basename(fn);

			item = DataSetItem.new
			item.name = filefn
			item.shortname = filefn
			item.geometry = fn
			item.reference_value = 0.0
			item.group = nil
			item.tags = ""

			@set.items << item
		}

		# Reference energies from table
		if @settings.set?(:dataset_from_files_reference)
			# Reset references
			@set.items.each{|item| item.reference_value = nil}
			# Load table
			table = IO.readlines(@settings[:dataset_from_files_reference])
			table.each{|line|
				name, value = line.strip.split
				@set.items.each{|item|
					if item.name == name
						item.reference_value = value.to_f
						break
					end
				}
			}
			# Check missing
			@set.items.each{|item|
				unless item.reference_value
					Cuby::error "Reference value not found in the table for item '#{item.name}'"
				end
			}
		end
	end

	def update_add_ref
		if @settings.set?(:dataset_add_ref_to_result)
			if @set.alternative_reference
				# Get all used references and check if they exist
				refs_used = []
				vars_used = []
				# split the string on math operators +-*/
				#(" "+@settings[:dataset_add_ref_to_result]+" ").split(/\s*[+\-\*\/\(\)]\s/).each{|ref|
				(" "+@settings[:dataset_add_ref_to_result]+" ").split(/\s*[+\-\*\/]\s/).each{|ref|
					ref.strip!
					next if ref == "" # in case the expression starts with -
					if @set.alternative_reference.has_key?(ref)
						refs_used << ref
					elsif ref =~ /x[0-9]/
						vars_used << ref
					else
						unless ref =~ /[0-9]+\.[0-9]+/
							# Raise error when it is not a number
							Cuby::error("Alternative reference data '#{ref}' requested but not found in the data set file (add_ref_to_result)")
						end
					end
				}

				# Make expression to eval
				expression = @settings[:dataset_add_ref_to_result].dup

				# Replace them with full names - 2-step, starting with longest,
				# to avoid submatches
				refs_used.sort!{|a,b| a.size <=> b.size}
				refs_used.reverse!
				refs_used.each_index{|i|
					ref = refs_used[i]
					expression.gsub!(ref, "%!#{i}!%")
				}
				refs_used.each_index{|i|
					ref = refs_used[i]
					expression.gsub!("%!#{i}!%", "@set.alternative_reference['#{ref}'][index]")
				}

				# Set up the environment for the evaluation
				x0 = @settings[:x0] if @settings.set?(:x0)
				x1 = @settings[:x1] if @settings.set?(:x1)
				x2 = @settings[:x2] if @settings.set?(:x2)
				x3 = @settings[:x3] if @settings.set?(:x3)
				x4 = @settings[:x4] if @settings.set?(:x4)
				x5 = @settings[:x5] if @settings.set?(:x5)
				x6 = @settings[:x6] if @settings.set?(:x6)
				x7 = @settings[:x7] if @settings.set?(:x7)
				x8 = @settings[:x8] if @settings.set?(:x8)
				x9 = @settings[:x9] if @settings.set?(:x9)
				# Evaluate for each item
				@set.items.each_index{|i|
					# Calculate actual index in filtered set
					index = @set.items[i].full_set_index
					# Evaluate
					begin
						@set.items[i].add_to_result = eval(expression)
					rescue => e
						# If eval fails, set to nil, will be caught later
						@set.items[i].add_to_result = e
					end
				}
			else
				Cuby::error("Alternative reference data requested but not found in the data set file (add_ref_to_result)")
			end
		end
	end

	
end
