module ActionsConnectors

	#=======================================================================
	# Actions - adding connectors
	#=======================================================================

	def add_connector
		# Add a connector dummy atom. It's position is defined by connector_orientation keyword
		case @settings[:connector_type]
		when :apex
			sel_no = 1
		else
			sel_no = 3
		end

		# Separate selections
		selections = @settings[:connector_orientation].strip.split(";")
		unless selections.size == sel_no
			Cuby::error "Keyword connector_orientation must define #{sel_no} selections"
		end

		# Convert them to atomlists
		atomlists = selections.map{|s|
			list = @geometry.atomlist_from_selection(s)
			Cuby::error "Selection in connector_orientation selects no atoms" unless list.size > 0
			list
		}


		# Minimum distance
		conn_dist = 0.0
		# Or largest vdw radius in the selection
		atomlists[0].each{|i|
			case @settings[:connector_distance].downcase
			when "vdw"
				r = PeriodicTable.vdw_radius(@geometry[i].element)
			when "covalent"
				r = PeriodicTable.covalent_radius(@geometry[i].element)
			when /[0-9][0-9]*\.[0-9][0-9]*/
				r = @settings[:connector_distance].to_f
			else
				Cuby::error "Invalid value of keyword 'connector_distance'"
			end
			if r > conn_dist
				conn_dist = r
			end
		}

		if conn_dist == 0.0
			Cuby::error("Connector distance set to 0, there is some problem")
		elsif conn_dist < 0.6
			Cuby::warning("Connector distance very small ( < 0.6 A )")
			r_minus = conn_dist / 2
		else
			r_minus = 0.5
		end



		case @settings[:connector_type]
		when :linear
			@geometry.orient!(atomlists[0],atomlists[1],atomlists[2])
			@geometry << Atom.new(:X, conn_dist - r_minus, 0, 0)
			@geometry << Atom.new(:X, conn_dist, 0, 0)
		when :perpendicular
			@geometry.orient!(atomlists[0],atomlists[1],atomlists[2])
			@geometry << Atom.new(:X, 0, 0, conn_dist - r_minus)
			@geometry << Atom.new(:X, 0, 0, conn_dist)
		when :out_of_plane
			@geometry.orient!(atomlists[0],atomlists[1],atomlists[2])
			dx = Math::cos(@settings[:connector_angle] / 180 * Math::PI)
			dz = Math::sin(@settings[:connector_angle] / 180 * Math::PI)
			@geometry << Atom.new(:X, dx*(conn_dist - r_minus), 0, dz*(conn_dist - r_minus))
			@geometry << Atom.new(:X, dx*conn_dist, 0, dz*conn_dist)
		when :in_plane
			@geometry.orient!(atomlists[0],atomlists[1],atomlists[2])
			dx = Math::cos(@settings[:connector_angle] / 180 * Math::PI)
			dy = Math::sin(@settings[:connector_angle] / 180 * Math::PI)
			@geometry << Atom.new(:X, dx*(conn_dist - r_minus), dy*(conn_dist - r_minus), 0)
			@geometry << Atom.new(:X, dx*conn_dist, dy*conn_dist, 0)
		when :apex
			unless atomlists[0].size == 1
				Cuby::error "Connctor of type 'apex' is defined only for a single atom"
			end
			atom_i = atomlists[0].first
			conn = Connectivity.new(@geometry, @settings[:connectivity_tolerance])
			vec = Coordinate[0,0,0]
			conn.bound_atoms(atom_i).each{|atom_j|
				bond = @geometry[atom_i] - @geometry[atom_j]
				vec += bond / bond.abs
			}
			# Normalize the vector
			vec = vec / vec.abs
			@geometry << Atom.from_coordinate(:X, @geometry[atom_i] + vec * (conn_dist-r_minus))
			@geometry << Atom.from_coordinate(:X, @geometry[atom_i] + vec * conn_dist)
		end

		# The connector is the last atom - write it to header
		@geometry.info[:protocol_comment] = "connector=#{@geometry.size}"

	end

	def connector_set_distance
		conn = @geometry.atomlist_from_selection(@settings[:geometry_connector])
		Cuby::error("Keyword geometry_connector should select one atom (#{conn.size} atoms selected)") unless conn.size == 1
		c_atom = @geometry[conn[0]]
		# Find closest atom
		a_atom = connect_connectors_get_anchor(@geometry, c_atom)
		# Calculate new position
		norm = (c_atom - a_atom) / (c_atom - a_atom).abs
		case @settings[:connector_distance].downcase
		when "vdw"
			dist = PeriodicTable.vdw_radius(a_atom.element)
		when "covalent"
			dist = PeriodicTable.covalent_radius(a_atom.element)
		when /[0-9][0-9]*\.[0-9][0-9]*/
			dist = @settings[:connector_distance].to_f
		else
			Cuby::error "Invalid value of keyword 'connector_distance'"
		end
		newpos = a_atom + norm * dist
		c_atom.set_coord(newpos)
	end

	#=======================================================================
	# Actions - structure building using connectors
	#=======================================================================
	
	def connect_connectors
		geometry2 = Geometry.from_file(@settings[:geometry2], @settings)

		mod_radii = @settings.elements_hash(:geometry_custom_vdw_radii)

		# Get connectors
		if @settings[:geometry_connector].downcase == "auto"
			if @geometry.info[:xyz_remark] && (matchdata = /connector *= *(\S+)/.match(@geometry.info[:xyz_remark]))
				puts "Connector read from geometry file: " + matchdata[1]
				@settings[:geometry_connector] = matchdata[1]
			else
				Cuby::error "Connector information not found in geometry"
			end
		end
		conn1 = @geometry.atomlist_from_selection(@settings[:geometry_connector])
		if @settings[:geometry_connector2].downcase == "auto"
			if geometry2.info[:xyz_remark] && (matchdata = /connector *= *(\S+)/.match(geometry2.info[:xyz_remark]))
				puts "Connector read from geometry file: " + matchdata[1]
				@settings[:geometry_connector2] = matchdata[1]
			else
				Cuby::error "Connector information not found in geometry2"
			end
		end
		conn2 = geometry2.atomlist_from_selection(@settings[:geometry_connector2])
		Cuby::error("Keyword geometry_connector should select one atom (#{conn1.size} atoms selected)") unless conn1.size == 1
		Cuby::error("Keyword geometry_connector2 should select one atom (#{conn2.size} atoms selected)") unless conn2.size == 1
		c1_atom = @geometry[conn1[0]]
		c2_atom = geometry2[conn2[0]]

		puts "Connectors:"
		puts c1_atom
		puts c2_atom

		# Get anchors - closest atom to connector
		a1_atom = connect_connectors_get_anchor(@geometry, c1_atom)
		a2_atom = connect_connectors_get_anchor(geometry2, c2_atom)

		puts "Anchors:"
		puts a1_atom
		puts a2_atom
		puts

		# Try to get charges
		charge1 = @geometry.info[:geofile_charge] || 0
		charge2 = geometry2.info[:geofile_charge] || 0
		puts "Charges:"
		puts charge1
		puts charge2
		puts

		# Move molecule 1 connector to origin
		@geometry.translate!(-(c1_atom.to_coordinate))

		# Move molecule 2
		geometry2.translate!(-(c2_atom.to_coordinate))

		# Annotate
		@geometry.each{|atom|
			atom.properties[:pdb_res_no] = 1
			atom.properties[:pdb_res_name] = "fr1"
		}
		geometry2.each{|atom|
			atom.properties[:pdb_res_no] = 2
			atom.properties[:pdb_res_name] = "fr2"
		}
		a1_atom.properties[:mol2_atomtype] = "L1"
		a2_atom.properties[:mol2_atomtype] = "L2"

		# Align a1 c1=c2 a2 to line
		while Coordinate.angle(a1_atom, c1_atom, a2_atom) < 0.1
			# break linearity
			geometry2.rotate_angle_axis!(0.3, Coordinate[rand,rand,rand])
		end
		normal = (a1_atom - c1_atom).cross_product(a2_atom - c2_atom)
		normal = normal / normal.abs
		angle = Coordinate.angle(a1_atom, c1_atom, a2_atom)
		geometry2.rotate_angle_axis!((Math::PI - angle), normal)
		angle = Coordinate.angle(a1_atom, c1_atom, a2_atom)

		# Rotate molecules around the axis
		com1 = @geometry.center
		com2 = geometry2.center
		torsion = Coordinate.torsion(com1,a1_atom,a2_atom,com2)
		# Maximize overlap - COMs must be aligned
		geometry2.rotate_angle_axis!(torsion, (a1_atom - a2_atom))

		if @settings[:geometry_connect_orientation] == :max_center_dist
			# Minimize overlap - COMs at opposite sides
			geometry2.rotate_angle_axis!(Math::PI, (a1_atom - a2_atom))
		end

		if @settings[:geometry_connect_break_symm]
			# Rotate by 10 deg around axis
			geometry2.rotate_angle_axis!(10.0 * Math::PI / 180, (a1_atom - a2_atom))
			# Bend out of axis
			geometry2.rotate_angle_axis!(10.0 * Math::PI / 180, normal)
			n2 = (a1_atom - a2_atom).cross_product(normal)
			n2 = n2 / n2.abs
			geometry2.rotate_angle_axis!(10.0 * Math::PI / 180, n2)
		end


		# Perform distance scan if requested
		if @settings[:geometry_connect_scan].size > 0
			scan_factors = @settings[:geometry_connect_scan]

			# Intramolecular axis - positive direction = increase distance
			intermolecular_v = (a2_atom - a1_atom)
			intermolecular_v /= intermolecular_v.abs

			# Geometries without dummy atoms
			g1 = Geometry.new
			@geometry.each{|atom|
				g1 << atom unless atom.element == :X
			}
			g2 = Geometry.new # Use deep copy of atoms here
			geometry2.each{|atom|
				g2 << atom.deep_copy unless atom.element == :X
			}

			# Distance between the connected atoms
			real_distance = (a2_atom - a1_atom).abs

			# Translate g2 to vdw distance	
			if @settings[:geometry_connect_scan_ref_dist] == :move_to_vdw
				# Read covalent radii, for dummy atoms the radius of
				# an adjacent atom is stored in .x0
				if a1_atom.element == :X
					r1 = a1_atom.properties[:x0]
				else
					r1 = PeriodicTable.vdw_radius(a1_atom.element, :default, mod_radii)
				end
				if a2_atom.element == :X
					r2 = a2_atom.properties[:x0]
				else
					r2 = PeriodicTable.vdw_radius(a2_atom.element, :default, mod_radii)
				end
				vdw_distance =  r1 + r2

				printf("Scan: moving to vdW distance %.3f, original distance was %.3f\n",
				       vdw_distance, real_distance)

				g2.translate!(intermolecular_v * (vdw_distance - real_distance))
				# Update the distance
				real_distance = (a2_atom - a1_atom).abs
			end

			# Find the actual closest contact that may involve other atoms
			closest_contact = 9.9e99
			ai = nil; aj = nil
			g1.each_index{|i|
				next if g1[i].element == :X
				g2.each_index{|j|
					next if g2[j].element == :X
					if g1[i].distance(g2[j]) < closest_contact
						closest_contact = g1[i].distance(g2[j])
						ai = i
						aj = j
					end
				}
			}
			if ai && aj
				vdw_closest =  PeriodicTable.vdw_radius(g1[ai].element, :default, mod_radii) + PeriodicTable.vdw_radius(g2[aj].element, :default, mod_radii)
				# Warn the user if there is another closest contact
				if (closest_contact < real_distance) && (closest_contact < vdw_closest)
					@messages << "WARNING: Prior to the scan,
		 other atoms form a close contact at #{sprintf('%.2f',closest_contact/vdw_closest*100)} % of their vdW distance"
				end
			end
		
			# Perform the scan
			g2.each{|atom|
				g1 << atom
			}
			scan_factors.each{|scale|
				translate = -(1.0 - scale) * real_distance
				g2.translate!(intermolecular_v * translate)
				g1.write_xyz(:file => "scan.xyz", :append => (scale != scan_factors.first))
				printf("Scan: setting distance to %.3f\n", real_distance * scale)
				#g1.write_xyz(:file => "point_#{'%.2f' % scale}.xyz")
				g2.translate!(intermolecular_v * -translate)
			}
			puts if scan_factors.size > 0
		end

		# Add geometry2 to geometry
		geometry2.each{|atom|
			@geometry << atom
		}

		# Save total charge
		@geometry.info[:protocol_comment] = "charge=#{charge1 + charge2} charge_a=#{charge1} charge_b=#{charge2}"
	end

	# Not an action, used in the method above
	def connect_connectors_get_anchor(geometry, atom_i) 

		# Closest non-X atom
		a = nil
		mindist = 9.9e99
		geometry.each_index{|i|
			next if geometry[i] == atom_i
			next if geometry[i].element == :X
			if (d = atom_i.distance(geometry[i])) < mindist
				mindist = d
				a = i
			end
		}

		# If there are multiple atoms close, make an average
		atoms = []
		allowance = 1.1
		geometry.each_index{|i|
			next if geometry[i] == atom_i
			next if geometry[i].element == :X
			d = atom_i.distance(geometry[i])
			if d < mindist * allowance
				atoms << geometry[i]
			end
		}

		# Only one atom found - return it
		return atoms[0] if atoms.size == 1

		# Make average, find max. vdW radius
		mod_radii = @settings.elements_hash(:geometry_custom_vdw_radii)
		crd = Coordinate[0,0,0]
		maxrad = 0.0
		atoms.each{|x|
			crd += x / atoms.size
			if PeriodicTable.vdw_radius(x.element, :default, mod_radii) > maxrad
				maxrad = PeriodicTable.vdw_radius(x.element, :default, mod_radii)
			end
		}
		a = Atom.from_coordinate(:X, crd)
		a.properties[:x0] = maxrad
		geometry << a
		return a
	end
end
