require 'tempfile'
require "classes/tools/output_parser.rb"

module ActionsSymmetry

	def print_symmetry
		# No geometry is written
		@write_geo = false

		symmol = Symmol.new(@geometry, @settings)

		# Print it according to verbosity level
		case @settings[:verbosity] 
		when :minimal
			Cuby::log.puts("#{symmol.symmetry}\tRMSD = #{symmol.rmsd} A")
		else
			Cuby::log.puts('Symmetry: ' + symmol.symmetry + "\tRMSD = #{symmol.rmsd} A")
		end

	end

	def symmetrize
		symmol = Symmol.new(@geometry, @settings)

		# Print it according to verbosity level
		case @settings[:verbosity] 
		when :minimal
			Cuby::log.puts("#{symmol.symmetry}\tRMSD = #{symmol.rmsd} A")
		else
			Cuby::log.puts('Symmetry: ' + symmol.symmetry + "\tRMSD = #{symmol.rmsd} A")
		end

		# Save geometry
		@geometry = symmol.symmetrized_geo
	end

	#=======================================================================
	# Symmol class
	#=======================================================================

	class Symmol
		attr_reader :symmetry
		attr_reader :rmsd
		attr_reader :symmetrized_geo

		def initialize(geometry, settings)
			# Check if symmol is configured
			unless FileTest.exist?(settings[:symmol_exe])
				Cuby::error "Path to symmol not configured, set up keyword 'symmol_exe'"
			end

			# Generate unique temporary dir
			symmol_dir= Cuby.make_tmpname('symmol_')
			Dir::mkdir(symmol_dir)

			# Tolerance
			tolerance = settings[:symmetry_tolerance]

			# Number atoms by elements
			atom_numbers = geometry.atom_numbering_by_element

			# Write input for symmol
			f = File.open(symmol_dir+'/symmol.in', "w+")
			f.puts "# Symmol input generated by Cuby"
			f.puts " 1.00000  1.00000  1.00000 90.00000 90.00000 90.00000"
			f.puts " 1 1 #{tolerance} #{tolerance}"
			geometry.each_with_index{|atom,i|
				f.printf("%-6s%2d%9.6f%9.6f%9.6f\n", "#{atom.element.to_s}#{atom_numbers[i]}", 1, atom.x, atom.y, atom.z)
			}
			f.close

			# Run symmol
			unless system("cd #{symmol_dir}; #{settings[:symmol_exe]} < symmol.in > symmol.stdout 2> symmol.stderr")
				Cuby::error "Symmol run failed, see directory '#{symmol_dir}' for details"
			end

			# Parse the output
			# Schoenflies symbol = D6h      CSM =  0.0000     Molecular RMS =  0.0000
			parser = OutputParser.new(symmol_dir + '/symmol.out')
			parser.add_pattern(:group, /^\s*Schoenflies symbol = *(\S+)/, :get => :submatch, :required => false, :type => :as_is)
			parser.add_pattern(:rmsd, / Molecular RMS = *(\S+)/, :get => :submatch, :required => false, :type => :float)
			parser.add_pattern(:nosymm, /(NO SYMMETRY EXISTS WITHIN GIVEN TOLERANCE)/, :get => :submatch, :required => false, :type => :as_is)
			parser.add_pattern(:few_atoms, /(GROUP WITH LESS THAN THREE ATOMS)/, :get => :submatch, :required => false, :type => :as_is)

			parser.add_block(
				:coords,
				/SYMMETRIZED ORTHOGONAL COORDINATES/,
				/^[A-Z]/,
				/^\s*$/,
				{:get => :line, :required => false}
			)

			symmgeo = Geometry.new
			parser.execute{|name, result, count|
				case name
				when :coords
					atom, no, x,y,z, dx, dy, dz = result.strip.split
					symmgeo << Atom.new(atom.gsub(/[0-9]+/,""), x.to_f, y.to_f, z.to_f)
				end
			}

			if parser[:group]
				@symmetry = parser[:group]
				@rmsd = parser[:rmsd]
				@symmetrized_geo = symmgeo
				unless symmgeo.size == geometry.size
					Cuby::error "Symmol geometry size does not match input geometry"
				end
				`rm -rf #{symmol_dir}`
			elsif parser[:nosymm]
				@symmetry = 'C1'
				@rmsd = 0.0
				@symmetrized_geo = geometry
				`rm -rf #{symmol_dir}`
			elsif parser[:few_atoms]
				# less than 3 atoms
				if geometry.size == 1
					@symmetry = 'Atom'
				else
					if geometry[0].element == geometry[1].element
						@symmetry = 'D(inf)h'
					else
						@symmetry = 'C(inf)v'
					end
				end
				@rmsd = 0.0
				@symmetrized_geo = geometry
			else
				Cuby::error "Error reading symmol output in '#{symmol_dir}'"
			end
		end

	end

end
