module DimerSettings
	#: Automated setup for dimers

	def auto_dimer_selections(settings, geometry, system_name='?')
		# Check if settings are there
		if settings.block(:molecule_a) && settings.block(:molecule_b)
			# Both set - nothing will be done
		elsif !settings.block(:molecule_a) && !settings.block(:molecule_b)
			# None set - automated setup
			Cuby::log.puts_debug("No monomers found in the input, automatic assignment will be attempted")
			settings[:molecule_a, :selection] = "auto"
			settings[:molecule_b, :selection] = "auto"
			# Multiplicity
			if settings[:multiplicity] != 1
				Cuby::error("Multiplicity of monomers can not be automatically assigned (#{system_name})")
			end
		else
			# Only one set - not allowed
			Cuby::error("Two monomers must be defined in the input. (#{system_name})")
		end

		# Auto selections
		info =  geometry.get_extended_info
		if settings[:molecule_a, :selection] == "auto"
			# Read selection from geometry
			begin
				settings[:molecule_a, :selection] = info.fragment_selection["a"]
				Cuby::log.puts_debug("Monomer A selection loaded from geometry file")
				# Read charge (optional)
				if info.fragment_charge && info.fragment_charge["a"]
					settings[:molecule_a, :charge] = info.fragment_charge["a"]
					Cuby::log.puts_debug("Monomer A charge loaded from geometry file")
				end
			rescue
				Cuby::log.puts_debug("Monomer A not defined in geometry, first molecule will be used")
				settings[:molecule_a, :selection] = "%molecule(1;2)" 
			end
		end
		if settings[:molecule_b, :selection] == "auto"
			# Read selection from geometry
			begin
				settings[:molecule_b, :selection] = info.fragment_selection["b"]
				Cuby::log.puts_debug("Monomer B selection loaded from geometry file")
				# Read charge (optional)
				if info.fragment_charge && info.fragment_charge["b"]
					settings[:molecule_b, :charge] = info.fragment_charge["b"]
					Cuby::log.puts_debug("Monomer B charge loaded from geometry file")
				end
			rescue
				Cuby::log.puts_debug("Monomer B not defined in geometry, 2nd molecule will be used")
				settings[:molecule_b, :selection] = "%molecule(2;2)" 
			end
		end
	end

	def fix_monomer_charges_settings(settings, system_name='?')
		# Set charges from files first
		if settings.set?(:molecule_a, :charge_from_file)
			settings[:molecule_a, :charge] = MolecularCharge.from_file(settings[:molecule_a, :charge_from_file])
			Cuby::log.puts_v(:normal, "Molecule A charge #{settings[:molecule_a, :charge]} read from file #{settings[:molecule_a, :charge_from_file]}")
		end
		if settings.set?(:molecule_b, :charge_from_file)
			settings[:molecule_b, :charge] = MolecularCharge.from_file(settings[:molecule_b, :charge_from_file])
			Cuby::log.puts_v(:normal, "Molecule B charge #{settings[:molecule_b, :charge]} read from file #{settings[:molecule_b, :charge_from_file]}")
		end

		# Fix charges
		if settings.set?(:molecule_a, :charge) && settings.set?(:molecule_b, :charge)
			# Both charges are set - keep as is
		elsif  !settings.set?(:molecule_a, :charge) && !settings.set?(:molecule_b, :charge)
			# No charge is set
			if !settings.set?(:charge) || settings[:charge] == 0
				# If total charge is 0, assume neutral monomers
				Cuby::warning("Charges not defined for monomers, using default value \"0\"")
				settings[:molecule_a, :charge] = 0
				settings[:molecule_b, :charge] = 0
			else
				# Otherwise, alert the user
				Cuby::error("Charge of monomers can not be automatically assigned (#{system_name})")
			end
		else
			# Only one charge is set - not allowed
			Cuby::error("Charge is defined only for one monomer.")
		end
	end
end
