require "protocols/scan/classes/scan_generator"

class ScanGeneratorZmatrix < ScanGenerator

	SETUP_KEYS = ["variables", "variables_scale"]

	def initialize(settings, geometry)
		# Check if the geometry has z-matrix
		unless geometry.info[:z_matrix]
			Cuby::error "Scan in z-matrix coordinates requires input geometry in z-matrix format"
		end
		@z_matrix = geometry.info[:z_matrix]

		# Check for keys
		check_setup_at_least_one(settings, SETUP_KEYS)

		# Load all the scans
		@absolute = {}
		@relative = {}
		@relative_orig = {}

		# Parse data on absolute variables
		if settings[:scan_generator_setup]["variables"]
			unless settings[:scan_generator_setup]["variables"].class == Hash
				Cuby::error("Keyword 'scan_generator_setup', key 'variables': Hash (variable name => series of values) expected")
			end
			settings[:scan_generator_setup]["variables"].each_pair{|name, vals|
				if @z_matrix.coord_by_name(name)
					@absolute[name] = range_from_value(vals, "variables:#{name}")
				else
					Cuby::error("Coordinate '#{name}' not found in the z-matrix")
				end
			}
		end

		# Parse data on relative scaling
		if settings[:scan_generator_setup]["variables_scale"]
			unless settings[:scan_generator_setup]["variables_scale"].class == Hash
				Cuby::error("Keyword 'scan_generator_setup', key 'variables_scale': Hash (variable name => series of values) expected")
			end
			settings[:scan_generator_setup]["variables_scale"].each_pair{|name, vals|
				if @z_matrix.coord_by_name(name)
					@relative[name] = range_from_value(vals, "variables_scale:#{name}")
					@relative_orig[name] = @z_matrix.coord_by_name(name).value
				else
					Cuby::error("Coordinate '#{name}' not found in the z-matrix")
				end
			}
		end

		# At least one has to be set
		if @absolute.size == 0 && @relative.size == 0
			Cuby::error("At least one z-matrix variable must be set in 'variables' or 'variables_scale'")
		end

		# keys should not overlap
		if (@absolute.keys & @relative.keys).size > 0
			Cuby::error("One z-matrix variable is set in both 'variables' and 'variables_scale'")
		end

		# All series must be of the same length
		all = (@absolute.merge @relative)
		ref = all.keys.first
		@series_length = all[ref].size
		all.each_pair{|name, series|
			if series.size != @series_length
				Cuby::error("All scanned series must be of the same length, '#{name}' difefrs from '#{ref}'")
			end
		}
	end

	def iterate(geometry)
		@series_length.times{|i|
			# Absolute vars
			@absolute.each_pair{|name, series|
				value = series[i]
				# Convert input in degrees to rad
				value = value / 180.0 * Math::PI unless @z_matrix.coord_by_name(name).type == :distance
				@z_matrix.coord_by_name(name).value = value
			}
			# Relative vars
			@relative.each_pair{|name, series|
				value = series[i] * @relative_orig[name]
				@z_matrix.coord_by_name(name).value = value
			}
			@z_matrix.update_geometry!(geometry)
			str = sprintf("scan: %d", i)
			yield geometry, str
		}
		return @series_length
	end

end
